/*
 * sp0718 Camera Driver
 *
 * Copyright (C) 2011 Actions Semiconductor Co.,LTD
 * Wang Xin <wangxin@actions-semi.com>
 *
 * Based on ov227x driver
 *
 * Copyright (C) 2008 Renesas Solutions Corp.
 * Kuninori Morimoto <morimoto.kuninori@renesas.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */


#include <linux/init.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/videodev2.h>
#include <media/v4l2-chip-ident.h>
#include <media/v4l2-subdev.h>
#include <media/soc_camera.h>
#include <media/soc_mediabus.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/regulator/consumer.h>
#include <mach/actions_reg_gl5202.h>
#include <mach/gl5202_cmu.h>
#include <mach/hardware.h>
#include <mach/gl5202_gpio.h>
#include <linux/clk.h>
#include <linux/pinctrl/consumer.h>
#include <mach/clock.h>
#include <mach/gpio.h>
#include <asm/prom.h>
#include "sp0718.h"
#include "../isp/gl5202_camera.h"

#define MODULE_NAME "camera_sp0718"


#define sp0718_DEBUG

#ifndef sp0718_DEBUG
#  define assert(expr) do {} while (0)
#else
#  define assert(expr) \
		if (unlikely(!(expr))) {				\
		pr_err("Assertion failed! %s,%s,%s,line=%d\n",	\
			   #expr, __FILE__, __func__, __LINE__);	\
		}
	#define debug_print(fmt,args...) printk(KERN_ALERT fmt, ##args)

#endif




#define PID					0x02 /* Product ID Number  */

/*
 * ID
 */
#define sp0718  0x71

#define OUTTO_SENSO_CLOCK 24000000


#if defined LOCAL_GL5202_DEMO
#define POWERDOWN_NUM              (23)
#else
#define POWERDOWN_NUM              (24)
#endif




/*
 * struct
 */
struct regval_list {
	unsigned char reg_num;
	unsigned char value;
};

struct sp0718_win_size {
	char								*name;
	__u32								width;
	__u32								height;
	__u32								exposure_line_width;
//	__u32								max_exposure_lines;
	__u32								capture_maximum_shutter;
	const struct regval_list			*win_regs;
	const struct regval_list			*lsc_regs;
	unsigned int        *frame_rate_array;
};

typedef struct
{
    unsigned int max_shutter;
    unsigned int shutter;
    unsigned int gain;
    unsigned int dummy_line;
    unsigned int dummy_pixel;
    unsigned int extra_line;
}exposure_param_t;

struct sp0718_priv {
	struct v4l2_subdev					subdev;
	struct v4l2_ctrl_handler hdl;
	struct module_info			*info;
	const struct sp0718_color_format	*cfmt;
	const struct sp0718_win_size		*win;
	int									model;

	unsigned short						brightness;
	unsigned short						contrast;
	unsigned short						saturation;
	unsigned short						auto_white_balance;
	unsigned short						exposure;
	unsigned short						autogain;
	unsigned short						gain;
	unsigned short						flag_vflip:1;
	unsigned short						flag_hflip:1;
	unsigned short						power_line_frequency;

	unsigned short						white_balance_temperature;
	unsigned short						sharpness;
	unsigned short						colorfx;

	unsigned short						band_stop_filter;

	unsigned short						exposure_auto;
	unsigned short						scene_exposure;


	struct v4l2_rect rect; /* Sensor window */
	struct v4l2_fract timeperframe;
	enum prev_capt prev_capt_mode;
	exposure_param_t preview_exposure_param;
	exposure_param_t capture_exposure_param;
};

//static DEFINE_MUTEX(camera_lock);

static struct regulator *sensor_avdd = NULL;
static struct regulator *sensor_dvdd = NULL;
static struct clk *sensor_clk;

static int front_hmirror, back_hmirror;
static int front_vflip, back_vflip;
static unsigned int is_two_same_module;
static unsigned int back_front_flag = 98;  //ĬΪ b = 98  'f' = 102

static struct gpio_pre_cfg gpio_sensor_pwdn;
static unsigned int gpio_sensor_pwdn_pin;
static struct gpio_pre_cfg gpio_sensor_reset;
static unsigned int gpio_sensor_reset_pin;



static struct gpio_pre_cfg gpio_sensor_avdd;
static unsigned int gpio_sensor_avdd_pin;

//static int avdd_use_ragulator, avdd_use_ldo, dvdd_use_ragulator, dvdd_use_ldo, dovdd_use_ragulator, dovdd_use_ldo;
static struct gpio_pre_cfg gpio_sensor_pwdn_another;
static unsigned int gpio_sensor_pwdn_pin_another;

#if 0
static struct pinctrl *p_pinctrl;
#endif 



//ǰõǸģΪpower downģʽ
static int camera_gpio_init_another(struct device *dev)
{
	struct soc_camera_link *sp0718_link = dev->platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	int err = 0;
	char *powerdown_name_another;
	
	if (sp0718_info->flags & MODULE_FLAG_FRONT)
	{
		powerdown_name_another = GPIO_NAME_SENSOR_BACK_PWDN;
		debug_print("another gpio config powerdown_name %s \n",powerdown_name_another);
	}
	else if (sp0718_info->flags & MODULE_FLAG_BACK)
	{
		powerdown_name_another = GPIO_NAME_SENSOR_FRONT_PWDN;
		debug_print("another gpio config powerdown_name %s \n",powerdown_name_another);
	}
	else
	{
		dev_err(dev, "No matched module in %s\n", __FUNCTION__);
		return -1;
	}

	/*
	 * 1) _init gpio_sensor_pwdn, gpio_sensor_reset
	 */
	 
	memset(&gpio_sensor_pwdn_another, 0, sizeof(gpio_sensor_pwdn_another));

	/*
	 * 2) get gpio_pre_cfg of sensor_pwdn & sensor_reset
	 */
	err = gpio_get_pre_cfg(powerdown_name_another, &gpio_sensor_pwdn_another);
	if (err != 0) {
		dev_err(dev, "gpio_get_pre_cfg gpio_sensor_pwdn_another error, err = %d\n", err);
		return -1;
	}

  debug_print("gpio_sensor_pwdn: group %d, pin %d\n", gpio_sensor_pwdn_another.iogroup, gpio_sensor_pwdn_another.pin_num);
	/*
	 * 3) get pin_no of gpio_sensor_pwdn_pin from gpio_sensor_pwdn
	 *    get pin_no of gpio_sensor_reset_pin from gpio_sensor_reset
	 */
	gpio_sensor_pwdn_pin_another = ASOC_GPIO_PORT(gpio_sensor_pwdn_another.iogroup, gpio_sensor_pwdn_another.pin_num);

	/*
	 * 4) get gpio of sensor_pwdn
	 *    get gpio of sensor_reset
	 */
	err = gpio_request(gpio_sensor_pwdn_pin_another, powerdown_name_another);
	if (err != 0) {
		dev_err(dev, "gpio_request sensor_pwdn_another error, err = %d\n", err);
		return -1;
	}

	/*
	 * 5) input/outut/set/get operation
	 */
	if (gpio_sensor_pwdn_another.gpio_dir == GPIOF_DIR_OUT) {
		err = gpio_direction_output(gpio_sensor_pwdn_pin_another, gpio_sensor_pwdn_another.active_level);
		if (err != 0) {
			gpio_free(gpio_sensor_pwdn_pin_another);
			dev_err(dev, "gpio_direction_output gpio_sensor_pwdn_another error, err = %d\n", err);
			return -1;
		}
	} else {
		err = gpio_direction_input(gpio_sensor_pwdn_pin_another);
		if (err != 0) {
			gpio_free(gpio_sensor_pwdn_pin_another);
			dev_err(dev, "gpio_direction_input gpio_sensor_pwdn_another error, err = %d\n", err);
			return -1;
		}
	}

	mdelay(10);
	return 0;
}
static int camera_gpio_init(struct device *dev);
static void camera_avdd_enable(void)
{
	int err = 0;
	char *avdd_ctl_name = GPIO_NAME_SENSOR_AVDD_CTL;
	debug_print("camera_avdd_enable\n");
	
	
	memset(&gpio_sensor_avdd, 0, sizeof(gpio_sensor_avdd));

	err = gpio_get_pre_cfg(avdd_ctl_name, &gpio_sensor_avdd);
	if (err != 0) {
		printk(KERN_ERR"get avdd control gpio error!!\n");
		return;
	}
	
	printk("gpio_sensor_avdd: group %d, pin %d\n", gpio_sensor_avdd.iogroup, gpio_sensor_avdd.pin_num);
	
	gpio_sensor_avdd_pin = ASOC_GPIO_PORT(gpio_sensor_avdd.iogroup, gpio_sensor_avdd.pin_num);
	
	err = gpio_request(gpio_sensor_avdd_pin, avdd_ctl_name);
	if (err) {
		printk(KERN_ERR"request avdd control gpio error!!\n");
		return;
	}
	

	/*
	 * 5) input/outut/set/get operation
	 */
	if (gpio_sensor_avdd.gpio_dir == GPIOF_DIR_OUT) {
		err = gpio_direction_output(gpio_sensor_avdd_pin, gpio_sensor_avdd.active_level);
		if (err != 0) {
		printk(KERN_ERR" set avdd  gpio output error!!\n");
		return;
		}
	} else {
		err = gpio_direction_input(gpio_sensor_avdd_pin);
		if (err != 0) {
		printk(KERN_ERR" set avdd  gpio input error!!\n");
		return;
		}
	}

	mdelay(1);
	return;
	
}
static void camera_avdd_disable(void)
{
	int err;
	if (gpio_sensor_avdd.gpio_dir == GPIOF_DIR_OUT) {
		err = gpio_direction_output(gpio_sensor_avdd_pin, gpio_sensor_avdd.unactive_level);
		if (err != 0) {
		printk(KERN_ERR" set avdd gpio output error in %s!!\n",__FUNCTION__);
		}
	} else {
		err = gpio_direction_input(gpio_sensor_avdd_pin);
		if (err != 0) {
		printk(KERN_ERR" set avdd  gpio input error in %s!!\n",__FUNCTION__);
		}
	}
	gpio_free(gpio_sensor_avdd_pin);
	
}



static int camera_regulator_init(struct device *dev)
{
	struct soc_camera_link *sp0718_link = dev->platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	int err = 0;
	char *avdd_name;
	char *dvdd_name;
		
	//if(!(avdd_use_ragulator || avdd_use_ldo || dvdd_use_ragulator || dvdd_use_ldo || dovdd_use_ragulator ||dovdd_use_ldo)) 
	//return 0;
	
		
	if (sp0718_info->flags & MODULE_FLAG_FRONT)
	{
		avdd_name = "sensor28";  //sensor avdd, һ2.8v
		dvdd_name = "wifi18";	//sensor dvdd, һ1.8v
	}
	else if (sp0718_info->flags & MODULE_FLAG_BACK)
	{
		avdd_name = "sensor28";
		dvdd_name = "wifi18";
	}
	else
	{
		dev_err(dev, "No matched module in %s\n", __FUNCTION__);
		return -1;
	}
		
	

	#if 0  
		sensor_avdd = regulator_get(NULL, avdd_name);
		if (IS_ERR(sensor_avdd)) {
			dev_err(dev, "failed to get regulator %s\n", avdd_name);
			err = PTR_ERR(sensor_avdd);
			goto eget1;
		}
	
		regulator_enable(sensor_avdd);
		err = regulator_set_voltage(sensor_avdd, 3000000, 3000000);
		if (err != 0) {
			dev_err(dev, "regulator_set_voltage sensor_avdd 3.0V error, err = %d\n", err);
			goto eset1;
		}
	#else //sensor avdd ldo磬ֱӲgpioأregulator
		camera_avdd_enable();
	#endif
	mdelay(10);
	
		sensor_dvdd = regulator_get(NULL, dvdd_name);
		if (IS_ERR(sensor_dvdd)) {
			dev_err(dev, "failed to get regulator %s\n", dvdd_name);
			err = PTR_ERR(sensor_dvdd);
			goto eget2;
		}
	
		regulator_enable(sensor_dvdd);
		err = regulator_set_voltage(sensor_dvdd, 1800000, 1800000);
		if (err != 0) {
			dev_err(dev, "regulator_set_voltage sensor_dvdd 1.8V error, err = %d\n", err);
			goto eset2;
		}
		mdelay(10);

		return 0;
	
	eset2:
		regulator_disable(sensor_dvdd);
		regulator_put(sensor_dvdd);
		sensor_dvdd = NULL;
	eget2:
#if 0
		regulator_disable(sensor_avdd);
	eset1:
			regulator_put(sensor_avdd);
			sensor_avdd = NULL;
#else
	camera_avdd_disable();
#endif
	eget1:
		return err;
}

static void camera_regulator_exit(struct device *dev)
{
		if (sensor_dvdd) {
			regulator_disable(sensor_dvdd);
			regulator_put(sensor_dvdd);
			sensor_dvdd = NULL;
		}

	#if 0
		if (sensor_avdd) {
			regulator_disable(sensor_avdd);
			regulator_put(sensor_avdd);
			sensor_avdd = NULL;
		}
	#else
	camera_avdd_disable();
	#endif
}




static int camera_gpio_init(struct device *dev)
{
	struct soc_camera_link *sp0718_link = dev->platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	int err = 0;
	char *powerdown_name;
	char *reset_name;
	
	if (sp0718_info->flags & MODULE_FLAG_FRONT)
	{
		powerdown_name = GPIO_NAME_SENSOR_FRONT_PWDN;
		reset_name = GPIO_NAME_SENSOR_FRONT_RESET;
		debug_print("front gpio config powerdown_name %s reset_name %s\n",powerdown_name,reset_name);
	}
	else if (sp0718_info->flags & MODULE_FLAG_BACK)
	{
		powerdown_name = GPIO_NAME_SENSOR_BACK_PWDN;
		reset_name = GPIO_NAME_SENSOR_BACK_RESET;
		debug_print("back gpio config\n");
	}
	else
	{
		dev_err(dev, "No matched module in %s\n", __FUNCTION__);
		return -1;
	}

	/*
	 * 1) _init gpio_sensor_pwdn, gpio_sensor_reset
	 */
	 
	memset(&gpio_sensor_pwdn, 0, sizeof(gpio_sensor_pwdn));
	memset(&gpio_sensor_reset, 0, sizeof(gpio_sensor_reset));

	/*
	 * 2) get gpio_pre_cfg of sensor_pwdn & sensor_reset
	 */
	 
	err = gpio_get_pre_cfg(powerdown_name, &gpio_sensor_pwdn);
	if (err != 0) {
		dev_err(dev, "gpio_get_pre_cfg gpio_sensor_pwdn error, err = %d\n", err);
		goto exit;
	}

	err = gpio_get_pre_cfg(reset_name, &gpio_sensor_reset);
	if (err != 0) {
		dev_err(dev, "gpio_get_pre_cfg gpio_sensor_reset error, err = %d\n", err);
		goto exit;
	}

	debug_print("gpio_sensor_pwdn: group %d, pin %d\n", gpio_sensor_pwdn.iogroup, gpio_sensor_pwdn.pin_num);
	debug_print("gpio_sensor_reset: group %d, pin %d\n", gpio_sensor_reset.iogroup, gpio_sensor_reset.pin_num);

	/*
	 * 3) get pin_no of gpio_sensor_pwdn_pin from gpio_sensor_pwdn
	 *    get pin_no of gpio_sensor_reset_pin from gpio_sensor_reset
	 */
	gpio_sensor_pwdn_pin = ASOC_GPIO_PORT(gpio_sensor_pwdn.iogroup, gpio_sensor_pwdn.pin_num);
	gpio_sensor_reset_pin = ASOC_GPIO_PORT(gpio_sensor_reset.iogroup, gpio_sensor_reset.pin_num);


	/*
	 * 4) get gpio of sensor_pwdn
	 *    get gpio of sensor_reset
	 */
	err = gpio_request(gpio_sensor_pwdn_pin, powerdown_name);
	if (err != 0) {
		dev_err(dev, "gpio_request sensor_pwdn error, err = %d\n", err);
		goto ereq1;
	}

	err = gpio_request(gpio_sensor_reset_pin, reset_name);
	if (err != 0) {
		dev_err(dev, "gpio_request sensor_reset error, err = %d\n", err);
		goto ereq2;
	}

	/*
	 * 5) input/outut/set/get operation
	 */
	if (gpio_sensor_pwdn.gpio_dir == GPIOF_DIR_OUT) {
		err = gpio_direction_output(gpio_sensor_pwdn_pin, gpio_sensor_pwdn.init_level);
		if (err != 0) {
			dev_err(dev, "gpio_direction_output gpio_sensor_pwdn error, err = %d\n", err);
			goto exit;
		}
	} else {
		err = gpio_direction_input(gpio_sensor_pwdn_pin);
		if (err != 0) {
			dev_err(dev, "gpio_direction_input gpio_sensor_pwdn error, err = %d\n", err);
			goto exit;
		}
	}

	mdelay(10);

	if (gpio_sensor_reset.gpio_dir == GPIOF_DIR_OUT) {
		err = gpio_direction_output(gpio_sensor_reset_pin, gpio_sensor_reset.init_level);
		if (err != 0) {
			dev_err(dev, "gpio_direction_output gpio_sensor_reset error, err = %d\n", err);
			goto exit;
		}
	} else {
		err = gpio_direction_input(gpio_sensor_reset_pin);
		if (err != 0) {
			dev_err(dev, "gpio_direction_input gpio_sensor_reset error, err = %d\n", err);
			goto exit;
		}
	}

	mdelay(10);

	return 0;

exit:
	gpio_free(gpio_sensor_reset_pin);
ereq2:
	gpio_free(gpio_sensor_pwdn_pin);
ereq1:
	return err;
}

 
static void camera_gpio_exit_another(void)
{
	gpio_free(gpio_sensor_pwdn_pin_another);
} 

static void camera_gpio_exit(void)
{
	gpio_free(gpio_sensor_reset_pin);
	gpio_free(gpio_sensor_pwdn_pin);
}

static void camera_reset(struct device *dev)
{
	 debug_print("%s %d\n",__FUNCTION__,__LINE__);
	/* RESET */
	gpio_direction_output(gpio_sensor_reset_pin, gpio_sensor_reset.active_level);
	mdelay(10);
	gpio_direction_output(gpio_sensor_reset_pin, gpio_sensor_reset.unactive_level);
	mdelay(10);
}

static void change_module_flag(unsigned int back_front, struct device *dev)
{
	struct soc_camera_link *sp0718_link = dev->platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	
	switch(back_front)
	{
		case 102://front
			
			sp0718_info->flags &=  (~MODULE_FLAG_BACK);
			sp0718_info->flags &=  (~MODULE_FLAG_HFLIP);
			sp0718_info->flags &=  (~MODULE_FLAG_VFLIP);

			sp0718_info->flags |= MODULE_FLAG_FRONT;
			
			if (front_hmirror)
				sp0718_info->flags |= MODULE_FLAG_HFLIP;
			if (front_vflip)
				sp0718_info->flags |= MODULE_FLAG_VFLIP;
			break;
		case 98: //back
			sp0718_info->flags &=  (~MODULE_FLAG_FRONT);
			sp0718_info->flags &=  (~MODULE_FLAG_HFLIP);
			sp0718_info->flags &=  (~MODULE_FLAG_VFLIP);

			sp0718_info->flags |= MODULE_FLAG_BACK;
			
			if (back_hmirror)
				sp0718_info->flags |= MODULE_FLAG_HFLIP;
			if (back_vflip)
				sp0718_info->flags |= MODULE_FLAG_VFLIP;	
			break;
	}
}


static int camera_power_on(struct device *dev)
{
	int ret = 0;

   debug_print("%s %d\n",__FUNCTION__,__LINE__);
 
  if (is_two_same_module)
  {
  	change_module_flag(back_front_flag, dev); 	
  	camera_gpio_init_another(dev);
  }
  
  if(camera_gpio_init(dev))
	{
		debug_print("%s gpio init error \n",__FUNCTION__);
		return -1;
	}
   
  #if 0   //sensor1ҪmfpϵͳʱĬʹܣܸãҪmfp
  debug_print(" dev name for pinctrl: %s \n",dev_name(dev));

	p_pinctrl = pinctrl_get_select_default(dev);
	if (IS_ERR(p_pinctrl)) 
	{
	    dev_err(dev, "failed to request pinctrl\n");
	    return PTR_ERR(p_pinctrl);
	}
	#endif

//	mutex_lock(&camera_lock);
	
	if(camera_regulator_init(dev))
	{
		debug_print("%s error1 \n",__FUNCTION__);
		return -1;
	}
	
	ret = gpio_direction_output(gpio_sensor_pwdn_pin, gpio_sensor_pwdn.unactive_level);
	if(ret)
	{
		printk(KERN_ERR "set gpio output err in %s() line %d\n\n",__FUNCTION__,__LINE__);
		return ret;
	}
	
	sensor_clk = clk_get_sys(CLK_NAME_SENSOR_OUT1_CLK, NULL); /*clk_nameȡclkṹ*/
	if (IS_ERR(sensor_clk))
	{
		ret = PTR_ERR(sensor_clk);
		printk(KERN_ERR "get sensor clock error in %s() line %d\n\n",__FUNCTION__,__LINE__);
		return ret;
	}
	clk_enable(sensor_clk);  /*enable clk*/
	ret = clk_set_rate(sensor_clk, OUTTO_SENSO_CLOCK); /*sensorƵ*/
	if(ret)
	{
		printk(KERN_ERR "set isp clock error in %s() line %d\n\n",__FUNCTION__,__LINE__);
		return ret;
	}
	debug_print("now sensor clock is %d\n",(int)(clk_get_rate(sensor_clk) / 1000000));
	mdelay(10);
	
	camera_reset(dev);
	
	return ret;
}

static int camera_power_off(struct device *dev)
{

	debug_print("%s %d\n",__FUNCTION__,__LINE__);
	clk_disable(sensor_clk);
	gpio_direction_output(gpio_sensor_pwdn_pin, gpio_sensor_pwdn.active_level);
	mdelay(10);
	camera_regulator_exit(dev);
	
#if 0  //sensor1ҪmfpϵͳʱĬʹܣܸãҪmfp
	pinctrl_put(p_pinctrl);
#endif
	camera_gpio_exit();
	
	if (is_two_same_module)
  {
		camera_gpio_exit_another();
  }
	
	return 0;
}


static int sp0718_power(struct device *dev, int mode)
{
	if (mode)
		return camera_power_on(dev);
	else
		return camera_power_off(dev);
}

static int sp0718_reset(struct device *dev)
{
	#if 0 //soc_cameraϵͳлpower on֮ǰresetresetʱgpioδ루gpiopower onʱ룩˴
	camera_reset(dev);
	
	#endif 
	return 0;
}

static struct i2c_board_info asoc_i2c_camera = {

		I2C_BOARD_INFO(MODULE_NAME, 0x21),
};

static struct module_info sp0718_info = {
	.flags		= MODULE_FLAG_8BIT,
//	.edgectrl	= sp0718_AUTO_EDGECTRL(0xf, 0),
};

static struct soc_camera_link sp0718_link = {
	.bus_id		= 0,
	.power		= sp0718_power,
	.reset		= sp0718_reset,
	.board_info	= &asoc_i2c_camera,
	.i2c_adapter_id	= 2,  //idŴ0ʼ
	.module_name	= MODULE_NAME,
	.priv		= &sp0718_info,
};

static struct platform_device asoc_camera_device = {
		.name	= "soc-camera-pdrv",
		.id	= SP0718_PLATFORM_ID,
		.dev	= {
			.platform_data = &sp0718_link,
			//.init_name = "sens1",
		},
};

#define ENDMARKER { 0xff, 0xff }

static struct regval_list sp0718_fmt_yuv422_yuyv[] = 
{		
//YCbYCr	
	0xfd,0x01,	
	0x35,0x40,
	ENDMARKER,
};

static struct regval_list sp0718_fmt_yuv422_yvyu[] = 
{		
//YCrYCb	
	0xfd,0x01,	
	0x35,0x41,
	ENDMARKER,
};
static struct regval_list sp0718_fmt_yuv422_vyuy[] = 
{		
//CrYCbY	
	0xfd,0x01,	
	0x35,0x01,
	ENDMARKER,
};
static struct regval_list sp0718_fmt_yuv422_uyvy[] = 
{		
//CbYCrY	
	0xfd,0x01,	
	0x35,0x00,
	ENDMARKER,
};
static struct regval_list sp0718_fmt_raw[] = 
{
	0xfd,0x01,
	ENDMARKER,
};
/*
*AWB
*/
static const struct regval_list sp0718_awb_regs_enable[] = 
{	
	{0xfd,0x02},                      
	{0x26,0xc8},		                  
	{0x27,0xb6},                      
	{0xfd,0x01}, 		
	{0x32,0x15},   
	{0xfd,0x00},  
	ENDMARKER,
};
static const struct regval_list sp0718_awb_regs_diable[] = 
{	
		//0xfd,0x01,	
		//0x28,0xce,	
		//0x29,0x8a,
		0xfd,0x00,
		0x32,0x05,
		ENDMARKER,
};
static struct regval_list sp0718_wb_cloud_regs[] = 
{	
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xdc},		             
	{0x27,0xe5},		             
	{0xfd,0x00}, 
	ENDMARKER,
};

static struct regval_list sp0718_wb_daylight_regs[] = 
{		
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xc8},		             
	{0x27,0x89},		             
	{0xfd,0x00}, 
	ENDMARKER,
};
static struct regval_list sp0718_wb_incandescence_regs[] = 
{	
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x75},		             
	{0x27,0xe2},		             
	{0xfd,0x00}, 
	ENDMARKER,
};
static struct regval_list sp0718_wb_fluorescent_regs[] = 
{		
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x91},		             
	{0x27,0xc8},		             
	{0xfd,0x00}, 
	ENDMARKER,
};
static struct regval_list sp0718_wb_tungsten_regs[] = 
{
	
	ENDMARKER,
};

static struct regval_list sp0718_colorfx_none_regs[] = 
{
	{0xfd, 0x01},
	{0x66, 0x00},
	{0x67, 0x80},
	{0x68, 0x80},
	ENDMARKER,
};

static struct regval_list sp0718_colorfx_bw_regs[] = 
{	
	0xfd,0x01,
	0x62,0x20,	
	0x63,0x80,
	0x64,0x80,
	ENDMARKER,
};

static struct regval_list sp0718_colorfx_sepia_regs[] = 
{	
	0xfd,0x01,
	0x62,0x10,	
	0x63,0xb0,	
	0x64,0x40,
	ENDMARKER,
};

static struct regval_list sp0718_colorfx_negative_regs[] = 
{	
	0xfd,0x01,	
	0x62,0x04,
	0x63,0x80,	
	0x64,0x80,
	ENDMARKER,
};


/*
 * register setting for window size
 */
#define SP0718_NORMAL_Y0ffset  0x20
#define SP0718_LOWLIGHT_Y0ffset  0x25
//AE target
#define  SP0718_P1_0xeb  0x78
#define  SP0718_P1_0xec  0x6c
//HEQ
#define  SP0718_P1_0x10  0x00//outdoor
#define  SP0718_P1_0x14  0x20
#define  SP0718_P1_0x11  0x00//nr
#define  SP0718_P1_0x15  0x18
#define  SP0718_P1_0x12  0x00//dummy
#define  SP0718_P1_0x16  0x10
#define  SP0718_P1_0x13  0x00//low
#define  SP0718_P1_0x17  0x00

/*
 * register setting for window size
 */
static const struct regval_list sp0718_vga_init_regs[] =
{
		{0xfd,0x00},
		{0x1C,0x00},
		{0x31,0x10},
		{0x27,0xb3},//0xb3	//2x gain
		{0x1b,0x17},
		{0x26,0xaa},
		{0x37,0x02},
		{0x28,0x8f},
		{0x1a,0x73},
		{0x1e,0x1b},
		{0x21,0x06},  //blackout voltage
		{0x22,0x2a},  //colbias
		{0x0f,0x3f},
		{0x10,0x3e},
		{0x11,0x00},
		{0x12,0x01},
		{0x13,0x3f},
		{0x14,0x04},
		{0x15,0x30},
		{0x16,0x31},
		{0x17,0x01},
		{0x69,0x31},
		{0x6a,0x2a},
		{0x6b,0x33},
		{0x6c,0x1a},
		{0x6d,0x32},
		{0x6e,0x28},
		{0x6f,0x29},
		{0x70,0x34},
		{0x71,0x18},
		{0x36,0x00},//02 delete badframe
		{0xfd,0x01},
		{0x5d,0x51},//position
		{0xf2,0x19},

		//Blacklevel
		{0x1f,0x10},
		{0x20,0x1f},
		//pregain 
		{0xfd,0x02},
		{0x00,0x88},
		{0x01,0x88},
		//SI15_SP0718 24M 50Hz 15-8fps 
		//ae setting
		{0xfd,0x00},
		{0x03,0x01},
		{0x04,0xce},
		{0x06,0x00},
		{0x09,0x02},
		{0x0a,0xc4},
		{0xfd,0x01},
		{0xef,0x4d},
		{0xf0,0x00},
		{0x02,0x0c},
		{0x03,0x01},
		{0x06,0x47},
		{0x07,0x00},
		{0x08,0x01},
		{0x09,0x00},
		//Status   
		{0xfd,0x02},
		{0xbe,0x9c},
		{0xbf,0x03},
		{0xd0,0x9c},
		{0xd1,0x03},
		{0xfd,0x01},
		{0x5b,0x03},
		{0x5c,0x9c},

		//rpc
		{0xfd,0x01},
		{0xe0,0x40},////24//4c//48//4c//44//4c//3e//3c//3a//38//rpc_1base_max
		{0xe1,0x30},////24//3c//38//3c//36//3c//30//2e//2c//2a//rpc_2base_max
		{0xe2,0x2e},////24//34//30//34//2e//34//2a//28//26//26//rpc_3base_max
		{0xe3,0x2a},////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_4base_max
		{0xe4,0x2a},////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_5base_max
		{0xe5,0x28},////24//2c//2a//2c//28//2c//24//22//20//rpc_6base_max
		{0xe6,0x28},////24//2c//2a//2c//28//2c//24//22//20//rpc_7base_max
		{0xe7,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_8base_max
		{0xe8,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_9base_max
		{0xe9,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_10base_max
		{0xea,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_11base_max
		{0xf3,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_12base_max
		{0xf4,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_13base_max
		//ae gain &status
		{0xfd,0x01},
		{0x04,0xe0},//rpc_max_indr
		{0x05,0x26},//1e//rpc_min_indr 
		{0x0a,0xa0},//rpc_max_outdr
		{0x0b,0x26},//rpc_min_outdr
		{0x5a,0x40},//dp rpc   
		{0xfd,0x02}, 
		{0xbc,0xa0},//rpc_heq_low
		{0xbd,0x80},//rpc_heq_dummy
		{0xb8,0x80},//mean_normal_dummy
		{0xb9,0x90},//mean_dummy_normal

		//ae target
		{0xfd,0x01}, 
		{0xeb,SP0718_P1_0xeb},//78 
		{0xec,SP0718_P1_0xec},//78
		{0xed,0x0a},	
		{0xee,0x10},

		//lsc       
		{0xfd,0x01},
		{0x26,0x30},
		{0x27,0x2c},
		{0x28,0x07},
		{0x29,0x08},
		{0x2a,0x40},
		{0x2b,0x03},
		{0x2c,0x00},
		{0x2d,0x00},
		  
		{0xa1,0x24},
		{0xa2,0x27},
		{0xa3,0x27},
		{0xa4,0x2b},
		{0xa5,0x1c},
		{0xa6,0x1a},
		{0xa7,0x1a},
		{0xa8,0x1a},
		{0xa9,0x18},
		{0xaa,0x1c},
		{0xab,0x17},
		{0xac,0x17},
		{0xad,0x08},
		{0xae,0x08},
		{0xaf,0x08},
		{0xb0,0x00},
		{0xb1,0x00},
		{0xb2,0x00},
		{0xb3,0x00},
		{0xb4,0x00},
		{0xb5,0x02},
		{0xb6,0x06},
		{0xb7,0x00},
		{0xb8,0x00},
		   
		   
		//DP       
		{0xfd,0x01},
		{0x48,0x09},
		{0x49,0x99},
		    
		//awb       
		{0xfd,0x01},
		{0x32,0x05},
		{0xfd,0x00},
		{0xe7,0x03},
		{0xfd,0x02},
		{0x26,0xc8},
		{0x27,0xb6},
		{0xfd,0x00},
		{0xe7,0x00},
		{0xfd,0x02},
		{0x1b,0x80},
		{0x1a,0x80},
		{0x18,0x26},
		{0x19,0x28},
		{0xfd,0x02},
		{0x2a,0x00},
		{0x2b,0x08},
		{0x28,0xef},//0xa0//f8
		{0x29,0x20},

		//d65 90  e2 93
		{0x66,0x42},//0x59//0x60////0x58//4e//0x48
		{0x67,0x62},//0x74//0x70//0x78//6b//0x69
		{0x68,0xee},//0xd6//0xe3//0xd5//cb//0xaa
		{0x69,0x18},//0xf4//0xf3//0xf8//ed
		{0x6a,0xa6},//0xa5
		//indoor 91
		{0x7c,0x3b},//0x45//30//41//0x2f//0x44
		{0x7d,0x5b},//0x70//60//55//0x4b//0x6f
		{0x7e,0x15},//0a//0xed
		{0x7f,0x39},//23//0x28
		{0x80,0xaa},//0xa6
		//cwf   92 
		{0x70,0x3e},//0x38//41//0x3b
		{0x71,0x59},//0x5b//5f//0x55
		{0x72,0x31},//0x30//22//0x28
		{0x73,0x4f},//0x54//44//0x45
		{0x74,0xaa},
		//tl84  93 
		{0x6b,0x1b},//0x18//11
		{0x6c,0x3a},//0x3c//25//0x2f
		{0x6d,0x3e},//0x3a//35
		{0x6e,0x59},//0x5c//46//0x52
		{0x6f,0xaa},
		//f    94
		{0x61,0xea},//0x03//0x00//f4//0xed
		{0x62,0x03},//0x1a//0x25//0f//0f
		{0x63,0x6a},//0x62//0x60//52//0x5d
		{0x64,0x8a},//0x7d//0x85//70//0x75//0x8f
		{0x65,0x6a},//0xaa//6a
		  
		{0x75,0x80},
		{0x76,0x20},
		{0x77,0x00},
		{0x24,0x25},

		//ڵƫʹ//ƫ
		{0x20,0xd8},
		{0x21,0xa3},//82//a8ƫնȻеƫ
		{0x22,0xd0},//e3//bc
		{0x23,0x86},

		//outdoor r\b range
		{0x78,0xc3},//d8
		{0x79,0xba},//82
		{0x7a,0xa6},//e3
		{0x7b,0x99},//86


		//skin 
		{0x08,0x15},//
		{0x09,0x04},//
		{0x0a,0x20},//
		{0x0b,0x12},//
		{0x0c,0x27},//
		{0x0d,0x06},//
		{0x0e,0x63},//

		//wt th
		{0x3b,0x10},
		//gw
		{0x31,0x60},
		{0x32,0x60},
		{0x33,0xc0},
		{0x35,0x6f},

		// sharp
		{0xfd,0x02},
		{0xde,0x0f},
		{0xd2,0x02},//6//ƺڰױߣ0-ߴ֣f-ϸ
		{0xd3,0x06},
		{0xd4,0x06},
		{0xd5,0x06},
		{0xd7,0x20},//10//2xжֵ
		{0xd8,0x30},//24//1A//4x
		{0xd9,0x38},//28//8x
		{0xda,0x38},//16x
		{0xdb,0x08},//
		{0xe8,0x58},//48//ǿ
		{0xe9,0x48},
		{0xea,0x30},
		{0xeb,0x20},
		{0xec,0x48},//60//80
		{0xed,0x48},//50//60
		{0xee,0x30},
		{0xef,0x20},
		//ƽ̹
		{0xf3,0x50},
		{0xf4,0x10},
		{0xf5,0x10},
		{0xf6,0x10},
		//dns       
		{0xfd,0x01},
		{0x64,0x44},//طԵƽ  //0-ǿ8-
		{0x65,0x22},
		{0x6d,0x04},//8//ǿƽƽ̹ƽֵ
		{0x6e,0x06},//8
		{0x6f,0x10},
		{0x70,0x10},
		{0x71,0x08},//0d//ƽƽ̹ƽֵ	
		{0x72,0x12},//1b
		{0x73,0x1c},//20
		{0x74,0x24},
		{0x75,0x44},//[7:4]ƽ̹ǿȣ[3:0]ƽ̹ǿȣ0-ǿ8-
		{0x76,0x02},//46
		{0x77,0x02},//33
		{0x78,0x02},
		{0x81,0x10},//18//2x//жֵǿƽƽ
		{0x82,0x20},//30//4x
		{0x83,0x30},//40//8x
		{0x84,0x48},//50//16x
		{0x85,0x0c},//12/8+reg0x81 ڶֵƽ̹ͷƽ̹
		{0xfd,0x02},
		{0xdc,0x0f},
		   
		//gamma    
		{0xfd,0x01},
		{0x8b,0x00},//00//00     
		{0x8c,0x0a},//0c//09     
		{0x8d,0x16},//19//17     
		{0x8e,0x1f},//25//24     
		{0x8f,0x2a},//30//33     
		{0x90,0x3c},//44//47     
		{0x91,0x4e},//54//58     
		{0x92,0x5f},//61//64     
		{0x93,0x6c},//6d//70     
		{0x94,0x82},//80//81     
		{0x95,0x94},//92//8f     
		{0x96,0xa6},//a1//9b     
		{0x97,0xb2},//ad//a5     
		{0x98,0xbf},//ba//b0     
		{0x99,0xc9},//c4//ba     
		{0x9a,0xd1},//cf//c4     
		{0x9b,0xd8},//d7//ce     
		{0x9c,0xe0},//e0//d7     
		{0x9d,0xe8},//e8//e1     
		{0x9e,0xef},//ef//ea     
		{0x9f,0xf8},//f7//f5     
		{0xa0,0xff},//ff//ff     
		//CCM      
		{0xfd,0x02},
		{0x15,0xd0},//b>th
		{0x16,0x95},//r<th
		//gcͷƫ
		//!F        
		{0xa0,0x80},//80
		{0xa1,0x00},//00
		{0xa2,0x00},//00
		{0xa3,0x00},//06
		{0xa4,0x8c},//8c
		{0xa5,0xf4},//ed
		{0xa6,0x0c},//0c
		{0xa7,0xf4},//f4
		{0xa8,0x80},//80
		{0xa9,0x00},//00
		{0xaa,0x30},//30
		{0xab,0x0c},//0c 
		//F        
		{0xac,0x8c},
		{0xad,0xf4},
		{0xae,0x00},
		{0xaf,0xed},
		{0xb0,0x8c},
		{0xb1,0x06},
		{0xb2,0xf4},
		{0xb3,0xf4},
		{0xb4,0x99},
		{0xb5,0x0c},
		{0xb6,0x03},
		{0xb7,0x0f},
		    
		//sat u     
		{0xfd,0x01},
		{0xd3,0x9c},//0x88//50
		{0xd4,0x98},//0x88//50
		{0xd5,0x8c},//50
		{0xd6,0x84},//50
		//sat v   
		{0xd7,0x9c},//0x88//50
		{0xd8,0x98},//0x88//50
		{0xd9,0x8c},//50
		{0xda,0x84},//50
		//auto_sat  
		{0xdd,0x30},
		{0xde,0x10},
		{0xd2,0x01},//autosa_en
		{0xdf,0xff},//a0//y_mean_th
		    
		//uv_th     
		{0xfd,0x01},
		{0xc2,0xaa},
		{0xc3,0xaa},
		{0xc4,0x66},
		{0xc5,0x66}, 

		//heq
		{0xfd,0x01},
		{0x0f,0xff},
		{0x10,SP0718_P1_0x10}, //out
		{0x14,SP0718_P1_0x14}, 
		{0x11,SP0718_P1_0x11}, //nr
		{0x15,SP0718_P1_0x15},  
		{0x12,SP0718_P1_0x12}, //dummy
		{0x16,SP0718_P1_0x16}, 
		{0x13,SP0718_P1_0x13}, //low 	
		{0x17,SP0718_P1_0x17},   	

		{0xfd,0x01},
		{0xcd,0x20},
		{0xce,0x1f},
		{0xcf,0x20},
		{0xd0,0x55},  
		//auto 
		{0xfd,0x01},
		{0xfb,0x33},
		{0x32,0x15},
		{0x33,0xff},
		{0x34,0xe7},
		{0x35,0x40},	
		//vga
		{0xfd, 0x01},
		{0x4a, 0x00},
		{0x4b, 0x01},
		{0x4c, 0xe0},
		{0x4d, 0x00},
		{0x4e, 0x02},
		{0x4f, 0x80},

		{0xfd, 0x02},
		{0x0f, 0x00},	// 1/4 subsample disable

		{0xfd, 0x00},
		{0x30, 0x00},

		ENDMARKER,
};             


static const struct regval_list sp0718_qvga_regs[] = {
	{0xfd, 0x01},
	{0x4a, 0x00},
	{0x4b, 0x01},
	{0x4c, 0xe0},
	{0x4d, 0x00},
	{0x4e, 0x02},
	{0x4f, 0x80},

	{0xfd, 0x02},
	{0x0f, 0x01},	// 1/4 subsample enable	
	
	{0xfd, 0x00},
	{0x30, 0x10},
	{0xff, 0xff}
};

static const struct regval_list sp0718_vga_regs[] = {
	{0xfd, 0x01},
	{0x4a, 0x00},
	{0x4b, 0x01},
	{0x4c, 0xe0},
	{0x4d, 0x00},
	{0x4e, 0x02},
	{0x4f, 0x80},

	{0xfd, 0x02},
	{0x0f, 0x00},	// 1/4 subsample disable
		
	{0xfd, 0x00},
	{0x30, 0x00},
	{0xff, 0xff}
};


struct sp0718_color_format {
	enum v4l2_mbus_pixelcode code;
	enum v4l2_colorspace colorspace;
};



/*
 * supported color format list.
 * see definition in
 *     http://thread.gmane.org/gmane.linux.drivers.video-input-infrastructure/12830/focus=13394
 * YUYV8_2X8_LE == YUYV with LE packing
 * YUYV8_2X8_BE == UYVY with LE packing
 * YVYU8_2X8_LE == YVYU with LE packing
 * YVYU8_2X8_BE == VYUY with LE packing
 */
static const struct sp0718_color_format sp0718_cfmts[] = {
	{
		.code		= V4L2_MBUS_FMT_YUYV8_2X8,
		.colorspace	= V4L2_COLORSPACE_JPEG,
//		.sys		= (2U << 2), /* YUV422 */
//		.dvp1		= 0,
//		.dvp2		= (0U << 5), /* YUYV */
	},
	{
		.code		= V4L2_MBUS_FMT_UYVY8_2X8,
		.colorspace	= V4L2_COLORSPACE_JPEG,
//		.sys		= (2U << 2), /* YUV422 */
//		.dvp1		= 0,
//		.dvp2		= (1U << 5), /* UYVY */
	},
	{
		.code		= V4L2_MBUS_FMT_YVYU8_2X8,
		.colorspace	= V4L2_COLORSPACE_JPEG,
//		.sys		= (2U << 2), /* YUV422 */
//		.dvp1		= 0,
//		.dvp2		= (2U << 5), /* YVYU */
	},
	{
		.code		= V4L2_MBUS_FMT_VYUY8_2X8,
		.colorspace	= V4L2_COLORSPACE_JPEG,
//		.sys		= (2U << 2), /* YUV422 */
//		.dvp1		= 0,
//		.dvp2		= (3U << 5), /* VYUY */
	},
};


/*
 * window size list
 */
#define QVGA_WIDTH			320
#define QVGA_HEIGHT			240
#define VGA_WIDTH		    640
#define VGA_HEIGHT		  480
#define sp0718_MAX_WIDTH	VGA_WIDTH
#define sp0718_MAX_HEIGHT	VGA_HEIGHT
//#define sp0718_MIN_WIDTH	QVGA_WIDTH
//#define sp0718_MIN_HEIGHT	QVGA_HEIGHT
//#define SENSOR_PREVIEW_720P_LINE_WIDTH 1600
//#define SENSOR_PREVIEW_SVGA_LINE_WIDTH 1230
//#define SENSOR_CAPTURE_UXGA_LINE_WIDTH 1950 
//#define SENSOR_CAPTURE_UXGA_MAXIMUM_SHUTTER 1228 
//#define MAX_CAPTURE_16X_GAIN  (4 * 16U)  /* MUST >0 && <32*16 */
#define AHEAD_LINE_NUM		15    //10 = 50ѭ(sp0718)

/*sp0718 պԤʱ֡ͬ*/
#define DROP_NUM_CAPTURE			4  
#define DROP_NUM_PREVIEW			8  


static unsigned int frame_rate_qvga[] = {12,};
static unsigned int frame_rate_vga[] = {12,};


/* 320X240 */
static const struct sp0718_win_size sp0718_win_qvga = {
	.name     = "QVGA",
	.width    = QVGA_WIDTH,
	.height   = QVGA_HEIGHT,
	//.exposure_line_width = 1230,
	//.capture_maximum_shutter = 618,
	.win_regs = sp0718_qvga_regs,
	//.lsc_regs = sp0718_non_hd720p_lenc_regs,
	.frame_rate_array = frame_rate_qvga,
};

/* 640X480 */
static const struct sp0718_win_size sp0718_win_vga = {
	.name     = "VGA",
	.width    = VGA_WIDTH,
	.height   = VGA_HEIGHT,
	//.exposure_line_width = 1600,
	//.capture_maximum_shutter = 0, //?
	.win_regs = sp0718_vga_regs,
	//.lsc_regs = sp0718_hd720p_lenc_regs,
	.frame_rate_array = frame_rate_vga,
};


static const struct sp0718_win_size *sp0718_win[] = {
	&sp0718_win_qvga,
	&sp0718_win_vga,
};



static struct platform_device  two_module_flag_device = {
		.name	= "camera_flag_device",
		.id	= 0,
		.dev	= {
//			.platform_data = &sp0718_link,
			//.init_name = "sens1",
		},
};



static ssize_t show_front_back(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	*buf = back_front_flag;
	*(buf+1) = '\n';
	return 2;
}


static ssize_t store_front_back(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count)
{
	
	switch (*buf)
	{
		case 102:  //'f'
			back_front_flag = 102;
			printk("now switch to front camera\n");
			break;
		case 98:   //'b'
			back_front_flag = 98;
			printk("now switch to back camera\n");
			break;
		default:
			printk("set front_back param error\n");
			return -EINVAL;
	}
	
	return count;
}

static DEVICE_ATTR(front_back, S_IRUGO | S_IWUGO,
		   show_front_back, store_front_back);


/*
 * general function
 */

static struct sp0718_priv *to_sp0718(const struct i2c_client *client)
{
	return container_of(i2c_get_clientdata(client), struct sp0718_priv,
				subdev);
}

static int sp0718_write_array(struct i2c_client        *client,
				  const struct regval_list *vals)
{
	while (vals->reg_num != 0xff) {
		int ret = i2c_smbus_write_byte_data(client,
							vals->reg_num,
							vals->value);
		if (ret < 0)
			return ret;
		vals++;
	}
	return 0;
}

static int sp0718_mask_set(struct i2c_client *client,
					  u8  command,
					  u8  mask,
					  u8  set)
{
	s32 val = i2c_smbus_read_byte_data(client, command);
	if (val < 0)
		return val;

	val &= ~mask;
	val |= set & mask;

	return i2c_smbus_write_byte_data(client, command, val);
}

static int sp0718_soft_reset(struct i2c_client *client)
{
#if 0
	int reg_0xfe = i2c_smbus_read_byte_data(client, 0xfe);
	reg_0xfe |= (0x1<<7);
	int ret = i2c_smbus_write_byte_data(client, 0xfe, reg_0xfe);
	msleep(10);
	dev_err(&client->dev, "%s, %d\n", __func__, __LINE__);
#endif
	return 0;
}

/*
 * soc_camera_ops function
 */
static const struct sp0718_win_size *sp0718_select_win(u32 width, u32 height)
{
	__u32 diff;
	const struct sp0718_win_size *win;

	/* default is VGA */
	diff = abs(width - sp0718_win_vga.width) +
		abs(height - sp0718_win_vga.height);
	win = &sp0718_win_vga;

	/* QVGA */
	if (diff >
		abs(width  - sp0718_win_qvga.width) +
		abs(height - sp0718_win_qvga.height)) {
		win = &sp0718_win_qvga;
	}
	return win;
}

static int sp0718_set_mbusformat(struct i2c_client *client, const struct sp0718_color_format *cfmt)
{
	enum v4l2_mbus_pixelcode code;
	code = cfmt->code;
	int ret = -1;
	switch (code)
	{
		case V4L2_MBUS_FMT_YUYV8_2X8:
		
			ret  = sp0718_write_array(client, sp0718_fmt_yuv422_yuyv);
			break;		
		case V4L2_MBUS_FMT_UYVY8_2X8:
			ret  = sp0718_write_array(client, sp0718_fmt_yuv422_uyvy);
		
			break;
		case V4L2_MBUS_FMT_YVYU8_2X8:
	
			ret  = sp0718_write_array(client, sp0718_fmt_yuv422_yvyu);
			break;
		case V4L2_MBUS_FMT_VYUY8_2X8:
			ret  = sp0718_write_array(client, sp0718_fmt_yuv422_vyuy);
	
			break;
		default:
			printk(KERN_ERR "mbus code error in %s() line %d\n",__FUNCTION__, __LINE__);
		
	}
}

static int sp0718_s_mirror_flip(struct i2c_client *client, unsigned int mirror, unsigned int flip)
{
return 0;
}


static int sp0718_set_params(struct i2c_client *client, u32 *width, u32 *height,
				 enum v4l2_mbus_pixelcode code)
{
	struct sp0718_priv *priv = to_sp0718(client);
	struct v4l2_subdev *sd = i2c_get_clientdata(client);
	struct soc_camera_device *icd = v4l2_get_subdev_hostdata(sd);
	struct soc_camera_host *ici = to_soc_camera_host(icd->parent);
	struct gl5202_camera_dev *cam_dev = ici->priv;
	const struct sp0718_win_size *old_win, *new_win;
	struct soc_camera_link *icl = soc_camera_i2c_to_link(client);
	struct module_info *mod_info = icl->priv;
	
	int ret = 0;
	int i;
	/*
	 * select format
	 */
  debug_enable = &(cam_dev->debug_print_switch); 
	 
	debug_print("width %d height %d\n",*width, *height);
	priv->cfmt = NULL;
	for (i = 0; i < ARRAY_SIZE(sp0718_cfmts); i++) {
		if (code == sp0718_cfmts[i].code) {
			priv->cfmt = sp0718_cfmts + i;
			break;
		}
	}
	if (!priv->cfmt) {
		dev_err(&client->dev, "Unsupported sensor format.\n");
		goto sp0718_set_fmt_error;
	}
	//dev_err(&client->dev, "%s, %d\n", __func__, __LINE__);
	/*
	 * select win
	 */
	old_win = priv->win;
	new_win = sp0718_select_win(*width, *height);
	priv->win = new_win;

  if ((old_win != NULL) && (new_win != NULL))
  {
		debug_print("old win (w*h): %d*%d ,new win (w*h):%d*%d\n",old_win->width, old_win->height, new_win->width, new_win->height);
	}
	else
	{
		printk(KERN_ERR "win select error in %s() line %d\n",__FUNCTION__, __LINE__);
	}
	
	

	/*
	 * set size format
	 */
//	ret = sp0718_write_array(client, sp0718_hd720p_init_regs);
  if ((cam_dev->started == DEV_CLOSE) || (cam_dev->started == DEV_RESUME))
  {
		ret = sp0718_write_array(client, sp0718_vga_init_regs);
		//ret |= sp0718_s_mirror_flip(client, (mod_info->flags) & MODULE_FLAG_HFLIP, (mod_info->flags) & MODULE_FLAG_VFLIP);
		debug_print("write moduel sp0718 init code (many regs should be written.).\n");
	}
	if (ret < 0) {
//		dev_err(&client->dev, "sp0718_hd720p_init error.\n");
		dev_err(&client->dev, "sp0718_svga_init2 error.\n");
		goto sp0718_set_fmt_error;
	}

	//dev_err(&client->dev, "win is %s, %s, %d\n", priv->win->name, __func__, __LINE__);
	ret = sp0718_write_array(client, priv->win->win_regs);
	debug_print("write win %s regs\n",priv->win->name);
	if (ret < 0) {
		dev_err(&client->dev, "sp0718 set windows & lenc parameters error.\n");
		goto sp0718_set_fmt_error;
	}

	ret = sp0718_set_mbusformat(client, priv->cfmt);

	
	if (ret < 0) {
		dev_err(&client->dev, "sp0718 set mbus format error.\n");
		goto sp0718_set_fmt_error;
	}

	//dev_err(&client->dev, "%s, %d\n", __func__, __LINE__);
	priv->rect.left = 0;
	priv->rect.top = 0;
	priv->rect.width = priv->win->width;
	priv->rect.height = priv->win->height;

	*width = priv->win->width;
	*height = priv->win->height;
	
	return ret;

sp0718_set_fmt_error:
	dev_err(&client->dev, "%s, %d\n", __func__, __LINE__);
	sp0718_soft_reset(client);
	priv->win = NULL;
	priv->cfmt = NULL;

	return ret;
}



static int sp0718_g_mbus_config(struct v4l2_subdev *sd,
				struct v4l2_mbus_config *cfg)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = i2c_get_clientdata(client);
	struct soc_camera_link *icl = soc_camera_i2c_to_link(client);
	
	cfg->flags = V4L2_MBUS_PCLK_SAMPLE_RISING | V4L2_MBUS_MASTER |
		V4L2_MBUS_HSYNC_ACTIVE_HIGH | V4L2_MBUS_VSYNC_ACTIVE_HIGH |/////LOW
		V4L2_MBUS_DATA_ACTIVE_HIGH;
		
	if (priv->info->flags & MODULE_FLAG_PALL)
	{
		cfg->type = V4L2_MBUS_PARALLEL;
	}
	else
	{
		debug_print("line %d in %s() ERROR\n",__LINE__,__FUNCTION__);
	}

	if (priv->info->flags & MODULE_FLAG_8BIT)
		cfg->flags |= SOCAM_DATAWIDTH_8;

	cfg->flags = soc_camera_apply_board_flags(icl, cfg);
	return 0;
}

static int sp0718_s_mbus_config(struct v4l2_subdev *sd,
				const struct v4l2_mbus_config *cfg)
{
	return 0;
}

static int sp0718_g_chip_ident(struct v4l2_subdev *sd,
				   struct v4l2_dbg_chip_ident *id)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);

	id->ident    = priv->model;
	id->revision = 0;

	return 0;
}

//static int sp0718_log_status(struct v4l2_subdev *sd)
//{
//
//	return 0;
//}


/* #####   USER CONTROLS DEFINITIONS  -  LOCAL TO THIS SOURCE FILE   ######## */
static int sp0718_set_brightness(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//	int brightness = ctrl->val;
//
//	unsigned int reg_0xb5, reg_0xd3;
//	int ret;
//
//	if (brightness < 0 || brightness > 6) {
//		dev_err(&client->dev, "set brightness over range, brightness = %d\n", brightness);
//		return -ERANGE;
//	}
//
//	
//	switch(brightness)
//  {
//		case 0:
//		reg_0xb5 = 0xd0;
//		reg_0xd3 = 0x68;	
//		break;      
//		case 1:
//		reg_0xb5 = 0xe0;
//		reg_0xd3 = 0x70;
//		break;              
//		case 2:
//		reg_0xb5 = 0xf0;
//		reg_0xd3 = 0x78;
//		break;              
//		case 3:
//		reg_0xb5 = 0x10;
//		reg_0xd3 = 0x88;//80
//		break;              
//		case 4:
//		reg_0xb5 = 0x20;
//		reg_0xd3 = 0x88;
//		break;              
//		case 5:
//		reg_0xb5 = 0x30;
//		reg_0xd3 = 0x90;	
//		break;              
//		case 6:
//		reg_0xb5 = 0x40;
//		reg_0xd3 = 0x98;
//		break;              		
//  }
//	
//	ret = i2c_smbus_write_byte_data(client, 0xb5, reg_0xb5);
//	ret |= i2c_smbus_write_byte_data(client, 0xd3, reg_0xd3);
//	assert(ret == 0);
//
//	priv->brightness = brightness;
//	ctrl->cur.val = brightness;
//	
	return 0;
}

static int sp0718_set_contrast(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
return 0;
}


static  struct regval_list sp0718_whitebance_auto[]=
{
	{0xfd,0x02},                      
	{0x26,0xc8},		                  
	{0x27,0xb6},                      
	{0xfd,0x01}, 		
	{0x32,0x15},   
	{0xfd,0x00},  
	ENDMARKER,
};
static  struct regval_list sp0718_whitebance_cloudy[]=
{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xdc},		             
	{0x27,0xe5},		             
	{0xfd,0x00}, 	
	ENDMARKER,
};
static  struct regval_list sp0718_whitebance_sunny[]=
{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xc8},		             
	{0x27,0x89},		             
	{0xfd,0x00},
	ENDMARKER,
};
/* Office Colour Temperature : 3500K - 5000K ,ӫ */
static  struct regval_list sp0718_whitebance_fluorescent[]=
{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x91},		             
	{0x27,0xc8},		             
	{0xfd,0x00},
	ENDMARKER,

};
/* Home Colour Temperature : 2500K - 3500K ׳ */
static  struct regval_list sp0718_whitebance_incandescent[]=
{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x75},		             
	{0x27,0xe2},		             
	{0xfd,0x00}, 
	ENDMARKER,
};


/*ģʽ*/
static  struct regval_list sp0718_effect_normal[] =
{
	{0xfd, 0x01},
	{0x66, 0x00},
	{0x67, 0x80},
	{0x68, 0x80},
	ENDMARKER,
};
/*ɫڰƬ*/
static  struct regval_list sp0718_effect_white_black[] =
{
	{0xfd, 0x01},
	{0x66, 0x20},
	{0x67, 0x80},
	{0x68, 0x80},
	ENDMARKER,
};
/*ƬЧ*/
static  struct regval_list sp0718_effect_negative[] =
{
	//Negative
	{0xfd, 0x01},
	{0x66, 0x04},
	{0x67, 0x80},
	{0x68, 0x80},
	ENDMARKER,
};
/*Ч*/
static  struct regval_list sp0718_effect_antique[] =
{
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0xc0},
	{0x68, 0x20},
	ENDMARKER,
};


static  struct regval_list sp0718_scene_auto[] =
{
	{0xfd,0x01},
	{0xcd,0x00},
	{0xce,0x1f},
	ENDMARKER,
};

static  struct regval_list sp0718_scene_night[] =
{
	{0xfd,0x01},
	{0xcd,0x15},
	{0xce,0x1f},
	ENDMARKER,
};

static int sp0718_set_auto_white_balance(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	int auto_white_balance = ctrl->val;
	int ret;

	if (auto_white_balance < 0 || auto_white_balance > 1) {
		dev_err(&client->dev, "set auto_white_balance over range, auto_white_balance = %d\n", auto_white_balance);
		return -ERANGE;
	}

	switch(auto_white_balance)
	{
		case 0:
		
			ret = sp0718_write_array(client, sp0718_awb_regs_diable);
			break;
		case 1:
			
			ret = sp0718_write_array(client, sp0718_awb_regs_enable);
			break;
	}

	assert(ret == 0);

	priv->auto_white_balance = auto_white_balance;
	ctrl->cur.val = auto_white_balance;

	return 0;
}

//static int sp0718_set_exposure(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
///*
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//	int exposure = ctrl->val;
//
//	unsigned int reg_0x13, reg_0x02, reg_0x03;
//	int ret;
//
//	if (exposure < 0 || exposure > 0xFFFFU) {
//		dev_err(&client->dev, "set exposure over range, exposure = %d\n", exposure);
//		return -ERANGE;
//	}
//
//	reg_0x13 = i2c_smbus_read_byte_data(client, 0x13);
//	assert(reg_0x13 >= 0);
//	reg_0x13 &= ~(1U << 1); // AUTO1[1]: AEC_en
//
//	reg_0x03 = exposure & 0xFFU;
//	reg_0x02 = (exposure >> 8) & 0xFFU;
//
//	ret = i2c_smbus_write_byte_data(client, 0x13, reg_0x13);
//	ret |= i2c_smbus_write_byte_data(client, 0x02, reg_0x02);
//	ret |= i2c_smbus_write_byte_data(client, 0x03, reg_0x03);
//	assert(ret == 0);
//
//	priv->exposure_auto = 0;
//	priv->exposure = exposure;
//	ctrl->cur.val = exposure;
//*/
//
//	return 0;
//}

//static int sp0718_set_autogain(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//	int autogain = ctrl->val;
//
//	unsigned int reg_0x13;
//	int ret;
//
//	if (autogain < 0 || autogain > 1) {
//		dev_err(&client->dev, "set autogain over range, autogain = %d\n", autogain);
//		return -ERANGE;
//	}
//
//	reg_0x13 = i2c_smbus_read_byte_data(client, 0x13);
//	assert(reg_0x13 >= 0);
//	reg_0x13 &= ~(1U << 0);
//	reg_0x13 |= ((autogain & 0x1) << 0); // AUTO1[0]: AGC_en
//
//	ret = i2c_smbus_write_byte_data(client, 0x13, reg_0x13);
//	assert(ret == 0);
//
//	priv->autogain = autogain;
//	ctrl->cur.val = autogain;
//
//	return 0;
//}

static int sp0718_set_gain(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	return 0;
}

static int sp0718_set_hflip(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	
	return 0;
}

//static int sp0718_set_vflip(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//	int vflip;
//	vflip = ctrl->val ? VFLIP_IMG : 0x00;
//		if (priv->info->flags & MODULE_FLAG_VFLIP)
//			vflip ^= VFLIP_IMG;
//	int ret;
//
//	if (vflip < 0 || vflip > 1) {
//		dev_err(&client->dev, "set vflip over range, vflip = %d\n", vflip);
//		return -ERANGE;
//	}
//
//	ret = sp0718_mask_set(client, SYS, VFLIP_IMG, vflip);
//	assert(ret == 0);
//
//	priv->flag_vflip = vflip;
//	ctrl->cur.val = vflip;
//
//	return 0;
//}

//static int sp0718_set_power_line_frequency(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//  int power_line_frequency = ctrl->val;
//
//	int ret;
//	unsigned int banding_filter, max_banding, banding_step;
//	unsigned int pixel_clk, HTS, light_freq, f_rate;
//	unsigned int reg_0x13, reg_0x14, reg_0x1d, reg_0x1e, reg_0x1f, reg_0x29, reg_0x2a;
//
//	reg_0x13 = i2c_smbus_read_byte_data(client, 0x13);
//	assert(reg_0x13 >= 0);
//	reg_0x14 = i2c_smbus_read_byte_data(client, 0x14);
//	assert(reg_0x14 >= 0);
//	reg_0x1d = i2c_smbus_read_byte_data(client, 0x1d);
//	assert(reg_0x1d >= 0);
//	reg_0x29 = i2c_smbus_read_byte_data(client, 0x29);
//	assert(reg_0x29 >= 0);
//	reg_0x2a = i2c_smbus_read_byte_data(client, 0x2a);
//	assert(reg_0x2a >= 0);
//
//	switch (power_line_frequency) {
//		case V4L2_CID_POWER_LINE_FREQUENCY_DISABLED:
//		{
//			reg_0x13 &= ~(1 << 5);
//			ret = i2c_smbus_write_byte_data(client, 0x13, reg_0x13);
//			assert(ret == 0);
//			return 0;
//		}
//		case V4L2_CID_POWER_LINE_FREQUENCY_50HZ:
//		{
//			light_freq = 50 * 2; // 100
//			banding_filter = 1; // 50HZ
//			break;
//		}
//		case V4L2_CID_POWER_LINE_FREQUENCY_60HZ:
//		{
//			light_freq = 60 * 2; // 120
//			banding_filter = 0; // 60HZ
//			break;
//		}
//		default:
//		{
//			dev_err(&client->dev, "set power_line_frequency over range, power_line_frequency = %d\n", power_line_frequency);
//			return -ERANGE;
//		}
//	}
//
//	HTS = (reg_0x29 << 8) | reg_0x2a;
//	f_rate = priv->timeperframe.denominator / priv->timeperframe.numerator;
//
//	if (priv->win == &sp0718_win_svga) {
//		pixel_clk = 48; // MHZ
//		/* bit[5:0]:max_banding, (1/f_rate) / (1/light_freq) - 1 */
//		max_banding = (light_freq / f_rate) - 1;
//	} else if (priv->win == &sp0718_win_hd720p) {
//		pixel_clk = 72; // MHZ
//		/* bit[5:0]:max_banding, (1/f_rate) / (1/light_freq) - 1 */
//		max_banding = (light_freq / f_rate) - 1;
//	} else if (priv->win == &sp0718_win_uxga) {
//		pixel_clk = 36; // MHZ
//		/* bit[5:0]:max_banding, (1/f_rate) / (1/light_freq) - 1 */
//		max_banding = (light_freq * 100 / 750) - 1; //7.5fps
//	} else {
//		dev_err(&client->dev, "sensor resolution is not correct.\n");
//		return -EINVAL;
//	}
//
//	reg_0x14 &= ~(1U << 7);
//	reg_0x14 |= ((banding_filter & 0x1) << 7);
//	reg_0x14 |= (1 << 5);
//
//	reg_0x13 |= (3U << 4);
//
//	/* bit[5:0]:max_banding, (1/f_rate) / (1/light_freq) - 1 */
//	if (max_banding >= 0x3FU) {
//		max_banding = 0x3FU;
//	} else if (max_banding <= 0x00) {
//		max_banding = 0x01;
//	} else {
//		max_banding &= 0x3FU;
//	}
//
//	reg_0x1d = (reg_0x1d & (~0x3FU)) | max_banding;
//
//	/*
//	 * Banding filter value = Frame rate * Maximum Exposure Line / light_freq
//	 * Banding_Filter value = PCLK_Frequency / (light_freq * default_line_width * 2)
//	 */
//	banding_step = pixel_clk * 1000000U / (HTS * light_freq * 2);
//	if (banding_step >= 0x3FFU) {
//		banding_step = 0x3FFU;
//	} else if (banding_step <= 0x0U) {
//		banding_step = 0x1U;
//	} else {
//		banding_step &= 0x3FFU;
//	}
//	reg_0x1e = (banding_step >> 8) & 0x3U;
//	reg_0x1f = banding_step & 0xFFU;
//
//	ret =  i2c_smbus_write_byte_data(client, 0x14, reg_0x14);
//	ret |= i2c_smbus_write_byte_data(client, 0x13, reg_0x13);
//	ret |= i2c_smbus_write_byte_data(client, 0x1d, reg_0x1d);
//	ret |= i2c_smbus_write_byte_data(client, 0x1e, reg_0x1e);
//	ret |= i2c_smbus_write_byte_data(client, 0x1f, reg_0x1f);
//	assert(ret == 0);
//
//	priv->power_line_frequency = power_line_frequency;
//	ctrl->cur.val = power_line_frequency;
//
//	return 0;
//}

static int sp0718_set_white_balance_temperature(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
  int white_balance_temperature = ctrl->val;

	//unsigned int reg_0xfd, reg_0x32, reg_0x26, reg_0x27;
	int ret;

	switch(white_balance_temperature)
	{
		case V4L2_WHITE_BALANCE_INCANDESCENT: /* ׳ */
		{
	
			ret = sp0718_write_array(client, sp0718_wb_incandescence_regs);
			break;
		}
		case V4L2_WHITE_BALANCE_FLUORESCENT: /* ӫ */
		{
	
			ret = sp0718_write_array(client,sp0718_wb_fluorescent_regs);
			break;
		}
		case V4L2_WHITE_BALANCE_DAYLIGHT: /* չ ()*/
		{
	
			ret = sp0718_write_array(client,sp0718_wb_daylight_regs);
			break;
		}
		case V4L2_WHITE_BALANCE_CLOUDY_DAYLIGHT: /*  죩*/
		{
	
			ret = sp0718_write_array(client,sp0718_wb_cloud_regs);
			break;
		}
		default:
		{
			dev_err(&client->dev, "set white_balance_temperature over range, white_balance_temperature = %d\n", white_balance_temperature);
			return -ERANGE;
		}
	}
	

	assert(ret == 0);
	
	priv->auto_white_balance = 0;
	priv->white_balance_temperature = white_balance_temperature;
	ctrl->cur.val = white_balance_temperature;
	
	return 0;
}

//static int sp0718_set_sharpness(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//  int sharpness = ctrl->val;
//
//	unsigned int reg_0x9f, reg_0xa0, reg_0xa7, reg_0xa8, reg_0xa1, reg_0xa2;
//	int ret;
//
//	if (sharpness < -2 || sharpness > 2) {
//		dev_err(&client->dev, "set sharpness over range, sharpness = %d\n", sharpness);
//		return -ERANGE;
//	}
//
//	reg_0x9f = 0xa;
//	reg_0xa0 = 0xf;
//	reg_0xa7 = 0xa;
//	reg_0xa8 = 0xf;
//
//	if (sharpness == 0) {
//		reg_0xa1 = 0x10;
//		reg_0xa2 = 0x04;
//	} else {
//		assert(abs(sharpness) * 0x2 <= 0x04);
//		assert(abs(sharpness) * 0x4 <= 0x10);
//		reg_0xa1 = 0x18 + (sharpness * 0x4);
//		reg_0xa2 = 0x04 + (sharpness * 0x2);
//	}
//
//	ret = i2c_smbus_write_byte_data(client, 0x9f, reg_0x9f);
//	ret |= i2c_smbus_write_byte_data(client, 0xa0, reg_0xa0);
//	ret |= i2c_smbus_write_byte_data(client, 0xa7, reg_0xa7);
//	ret |= i2c_smbus_write_byte_data(client, 0xa8, reg_0xa8);
//	ret |= i2c_smbus_write_byte_data(client, 0xa1, reg_0xa1);
//	ret |= i2c_smbus_write_byte_data(client, 0xa2, reg_0xa2);
//	assert(ret == 0);
//
//	priv->sharpness = sharpness;
//	ctrl->cur.val = sharpness;
//	
//	return 0;
//}

static int sp0718_set_colorfx(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
  int colorfx = ctrl->val;

	int ret;

	switch (colorfx) {
		case V4L2_COLORFX_NONE: /* normal */
		{
			ret = sp0718_write_array(client, sp0718_colorfx_none_regs);
			break;
		}
		case V4L2_COLORFX_BW: /* black and white */
		{	
			ret = sp0718_write_array(client, sp0718_colorfx_bw_regs);
			break;
		}
		case V4L2_COLORFX_SEPIA: /* antique ,*/
		{
			ret = sp0718_write_array(client, sp0718_colorfx_sepia_regs);

			break;
		}
		case V4L2_COLORFX_NEGATIVE: /* negativeƬ */
		{
			ret = sp0718_write_array(client, sp0718_colorfx_negative_regs);
			break;
		}
		default:
		{
			dev_err(&client->dev, "set colorfx over range, colorfx = %d\n", colorfx);
			return -ERANGE;
		}
	}

	assert(ret == 0);

	priv->colorfx = colorfx;
	ctrl->cur.val = colorfx;

	return 0;
}

//static int sp0718_set_band_stop_filter(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//  int band_stop_filter = ctrl->val;
//
//	unsigned int reg_0x1d;
//	int ret;
//
//	if (band_stop_filter < 0 || band_stop_filter > 63) {
//		dev_err(&client->dev, "set band_stop_filter over range, band_stop_filter = %d\n", band_stop_filter);
//		return -ERANGE;
//	}
//
//	reg_0x1d = band_stop_filter & 0x3f;
//
//	ret = i2c_smbus_write_byte_data(client, 0x1d, reg_0x1d);
//	assert(ret == 0);
//
//	priv->band_stop_filter = band_stop_filter;
//	ctrl->cur.val = band_stop_filter;
//	
//	return 0;
//}

static int sp0718_set_exposure_auto(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
  int exposure_auto = ctrl->val;

	
	int ret;




	if (exposure_auto < 0 || exposure_auto > 1) {
		dev_err(&client->dev, "set exposure_auto over range, exposure_auto = %d\n", exposure_auto);
		return -ERANGE;
	}
  switch (exposure_auto)
  {
   	case V4L2_EXPOSURE_AUTO: //auto
	  {       	
	
      debug_print("set exposure non auto in %s() line %d\n",__FUNCTION__,__LINE__);
		  break;
	  }
	
	  case V4L2_EXPOSURE_MANUAL:/*  auto */
	  {
	
      debug_print("set exposure  auto in %s() line %d\n",__FUNCTION__,__LINE__);
      goto change_val;
		  break;
	  }
  }


	assert(ret == 0);
change_val:
	priv->exposure_auto = exposure_auto;
	ctrl->cur.val = exposure_auto;

	return 0;
}

static int set_exposure_auto(struct v4l2_subdev *sd)
{
	return 0;
}


static int sp0718_set_scene_exposure(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
  int scene_exposure = ctrl->val;
#if 0 //not suppot now caichsh added
	int ret;




	switch(scene_exposure)
	{
		case V4L2_SCENE_MODE_HOUSE:  //
		{
	
      debug_print("set exposure  V4L2_SCENE_MODE_HOUSE in %s() line %d\n",__FUNCTION__,__LINE__);
			break;
		}
		case V4L2_SCENE_MODE_SUNLIGHT:  //
		{
		
			debug_print("set exposure  V4L2_SCENE_MODE_SUNLIGHT in %s() line %d\n",__FUNCTION__,__LINE__);
			break;
		}
		default:
		{
			dev_err(&client->dev, "set scene_exposure over range, scene_exposure = %d\n", scene_exposure);
			return -ERANGE;
		}
	}


	assert(ret == 0);
#endif
	priv->scene_exposure = scene_exposure;
	ctrl->cur.val = scene_exposure;

	return 0;
}

static int sp0718_set_prev_capt_mode(struct v4l2_subdev *sd, struct v4l2_ctrl *ctrl)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
  int mode = ctrl->val;

	switch(mode)
	{
		case PREVIEW_MODE:
		{
			priv->prev_capt_mode = mode;
			break;
		}
		case CAPTURE_MODE:
		{
      priv->prev_capt_mode = mode;
			break;
		}
		default:
		{
			dev_err(&client->dev, "set_prev_capt_mode over range, prev_capt_mode = %d\n", mode);
			return -ERANGE;
		}
	}
	ctrl->cur.val = mode;

	return 0;
}

//static int sp0718_g_volatile_ctrl(struct v4l2_ctrl *ctrl)
//{
//	
//	return 0;
//}

static int sp0718_s_ctrl(struct v4l2_ctrl *ctrl)
{
	struct sp0718_priv *priv = container_of(ctrl->handler,
					       struct sp0718_priv , hdl);
	struct v4l2_subdev *sd = &priv->subdev;
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	int ret = 0;
debug_print("%s line %d id %x\n",__FUNCTION__,__LINE__, ctrl->id);
	switch (ctrl->id) {
//	case V4L2_CID_BRIGHTNESS:
//		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_brightness(sd, ctrl);
//		
//		break;
//	case V4L2_CID_CONTRAST:
//		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//		sp0718_set_contrast(sd, ctrl);
//		
//	case V4L2_CID_SATURATION:
//		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_saturation(sd, ctrl);
//		
//		break;

	case V4L2_CID_AUTO_WHITE_BALANCE:
		sp0718_set_auto_white_balance(sd, ctrl);
		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;

//	case V4L2_CID_EXPOSURE:
//		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_exposure(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//	case V4L2_CID_AUTOGAIN: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_autogain(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//	case V4L2_CID_GAIN: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_gain(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//	case V4L2_CID_HFLIP: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_hflip(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//	case V4L2_CID_VFLIP: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_vflip(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;
//	case V4L2_CID_POWER_LINE_FREQUENCY: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_power_line_frequency(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;

	case V4L2_CID_WHITE_BALANCE_TEMPERATURE: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		sp0718_set_white_balance_temperature(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;
//	case V4L2_CID_SHARPNESS: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_sharpness(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;

	case V4L2_CID_COLORFX: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		sp0718_set_colorfx(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;

//	case V4L2_CID_BAND_STOP_FILTER: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		sp0718_set_band_stop_filter(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
//		break;

	case V4L2_CID_EXPOSURE_AUTO: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		//sp0718_set_exposure_auto(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;

	case V4L2_CID_SCENE_EXPOSURE: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		//sp0718_set_scene_exposure(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;
	case V4L2_CID_PRIVATE_PREV_CAPT: debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		//sp0718_set_prev_capt_mode(sd, ctrl);debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		break;
		
	default:
		dev_err(&client->dev, "getctrl invalid control id %d\n",ctrl->id);
		return -EINVAL;
	}
debug_print("%s line %d\n",__FUNCTION__,__LINE__);		
	return ret;
}


static const struct v4l2_ctrl_ops sp0718_ctrl_ops = {
//	.g_volatile_ctrl = sp0718_g_volatile_ctrl,
	.s_ctrl = sp0718_s_ctrl,
};

//#ifdef CONFIG_VIDEO_ADV_DEBUG
//static int sp0718_g_register(struct v4l2_subdev *sd,
//				 struct v4l2_dbg_register *reg)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	int ret;
//
//	reg->size = 1;
//	if (reg->reg > 0xff)
//		return -EINVAL;
//
//	ret = i2c_smbus_read_byte_data(client, reg->reg);
//	if (ret < 0)
//		return ret;
//
//	reg->val = (__u64)ret;
//
//	return 0;
//}

//static int sp0718_s_register(struct v4l2_subdev *sd,
//				 struct v4l2_dbg_register *reg)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//
//	if (reg->reg > 0xff ||
//		reg->val > 0xff)
//		return -EINVAL;
//
//	return i2c_smbus_write_byte_data(client, reg->reg, reg->val);
//}
//#endif



static struct v4l2_subdev_core_ops sp0718_subdev_core_ops = {
	.g_chip_ident	= sp0718_g_chip_ident,
////	.log_status		= sp0718_log_status,

	//.queryctrl		= sp0718_queryctrl,
//	.g_ctrl			= sp0718_g_ctrl,
//	.s_ctrl			= sp0718_s_ctrl,

//	.g_ext_ctrls	= sp0718_g_ext_ctrls,
//	.s_ext_ctrls	= sp0718_s_ext_ctrls,
//	.try_ext_ctrls	= sp0718_try_ext_ctrls,
////	.s_std			= sp0718_s_std,
//#ifdef CONFIG_VIDEO_ADV_DEBUG
//	.g_register		= sp0718_g_register,
//	.s_register		= sp0718_s_register,
//#endif
};

/*
*softstandby:ͨüĴʹģ͹ģʽ״̬pin衣ĴֵҼĴд
*ͨpwdn pin͹ģʽʱpin裬Ĵֵ䣬Ĵд
*/
int sp0718_module_softstandby(struct i2c_client *client)
{
	#if 0
	int ret;
	ret = i2c_smbus_write_byte_data(client, 0xfd, 0x00);
	ret |= i2c_smbus_write_byte_data(client, 0x0e, 0x3e);
	debug_print("sensor work state standby\n");
	#endif
	return 0;
	
}
/*
*softstandby״ָ̬normal״̬
*/
int sp0718_module_normal(struct i2c_client *client)
{
	#if 0
	int ret;
	ret = i2c_smbus_write_byte_data(client, 0xfd, 0x00);
	ret |= i2c_smbus_write_byte_data(client, 0x0e, 0x00);
	debug_print("sensor work state nornal\n");
	#endif
	return 0;
}


static int sp0718_s_stream(struct v4l2_subdev *sd, int enable)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	struct soc_camera_device *icd = v4l2_get_subdev_hostdata(sd);
	struct gl5202_camera_param *cam_param = icd->host_priv;
	int ret;
	debug_print("%s line %d\n",__FUNCTION__,__LINE__);
	if (!enable) {
		sp0718_module_softstandby(client);
		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		return 0;
	}
	debug_print("%s line %d\n",__FUNCTION__,__LINE__);
	if (!priv->win || !priv->cfmt) {
		dev_err(&client->dev, "norm or win select error\n");
		return -EPERM;
	}
	debug_print("%s line %d\n",__FUNCTION__,__LINE__);
	if ( priv->prev_capt_mode == PREVIEW_MODE)
	{ 

		cam_param->drop_frame_num = DROP_NUM_PREVIEW;
		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
		
	}
  else if ( priv->prev_capt_mode == CAPTURE_MODE)
	{	  
		cam_param->drop_frame_num = DROP_NUM_CAPTURE;
		//debug_print("cameracapture mode\n");
		debug_print("%s line %d\n",__FUNCTION__,__LINE__);
	}
	ret |= sp0718_module_normal(client);
	
  if(ret)
  {
  	debug_print(" %s line %d error\n",__FUNCTION__,__LINE__);
  }
	// wait 550ms (wait for the 3rd frame)
	/* ΪúܿѾһм,
	 * ĵ2֡һʱ,
	 * ʵֻҪȵڶvsyncͿDMA, DMA
	 * ȵvsync,	ץľǵ.
	 * delay = (1/7.5)	* 4	= 533.3mS
	 */
	//msleep(550);  //˴msleep֡Ϊprelineжϼ֡
	debug_print("%s line %d\n",__FUNCTION__,__LINE__);
	dev_dbg(&client->dev, "format %d, win %s\n",
		priv->cfmt->code, priv->win->name);

	return ret;
}

static int sp0718_cropcap(struct v4l2_subdev *sd, struct v4l2_cropcap *a)
{
	a->bounds.left				= 0;
	a->bounds.top				= 0;
	a->bounds.width				= sp0718_MAX_WIDTH;
	a->bounds.height			= sp0718_MAX_HEIGHT;
	a->defrect					= a->bounds;
	a->type						= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	a->pixelaspect.numerator	= 1;
	a->pixelaspect.denominator	= 30;

	return 0;
}

static int sp0718_g_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);

	a->c	= priv->rect;
	a->type	= V4L2_BUF_TYPE_VIDEO_CAPTURE;

	return 0;
}

//static int sp0718_s_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
//{
//	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	struct sp0718_priv *priv = to_sp0718(client);
//	struct v4l2_rect rect = a->c;
//	int ret;
//
//	/* Bayer format - even size lengths */
//	if (V4L2_MBUS_FMT_SBGGR10_2X8_PADHI_LE == priv->cfmt.code) {
//		rect.width	= ALIGN(rect.width, 2);
//		rect.height	= ALIGN(rect.height, 2);
//		/* Let the user play with the starting pixel */
//	}
//
//	soc_camera_limit_side(&rect.left, &rect.width, 0, 2, 1600);
//
//	soc_camera_limit_side(&rect.top, &rect.height, 0, 1, 1200);
//
//
//	dev_dbg(&client->dev, "Frame %dx%d pixel\n", rect.width, rect.height);
//
//	mt9v022->rect = rect;
//
//	return 0;
//}

static int sp0718_g_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	struct v4l2_captureparm *cp = &parms->parm.capture;

	if (parms->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	memset(cp, 0, sizeof(struct v4l2_captureparm));
	cp->capability = V4L2_CAP_TIMEPERFRAME;
	cp->timeperframe.numerator = priv->timeperframe.numerator;
	cp->timeperframe.denominator = priv->timeperframe.denominator;

	return 0;
}

static int sp0718_set_framerate(struct v4l2_subdev *sd, __u32 numerator, __u32 denominator)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	unsigned int frame_rate = denominator / numerator;
	unsigned int *frame_rate_array = priv->win->frame_rate_array;
	unsigned int array_size = sizeof(frame_rate_array)/sizeof(unsigned int);
	int i = 0;
	
	debug_print("frame rate array size %d in %s()\n",array_size, __FUNCTION__);

	do{
		if (frame_rate_array[i] == frame_rate)
			break;
		i++;
	}while(i < array_size);
	
	if (i == array_size)
	{
		dev_err(&client->dev, "frame rate denominator(%d) numerator(%d) no match.\n", denominator, numerator);
		return -EINVAL;
	}
	
	priv->timeperframe.numerator = numerator;
	priv->timeperframe.denominator = denominator;

	return 0;
}

static int sp0718_s_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct v4l2_captureparm *cp = &parms->parm.capture;
	struct v4l2_fract *tpf = &cp->timeperframe;
	int ret = 0;

	if (parms->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (cp->extendedmode != 0)
		return -EINVAL;

	ret = sp0718_set_framerate(sd, tpf->numerator, tpf->denominator);
	if (ret != 0) {
		dev_err(&client->dev, "sp0718_set_framerate error\n");
		return ret;
	}

	return 0;
}

static int sp0718_enum_framesizes(struct v4l2_subdev *sd, struct v4l2_frmsizeenum *fsize)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	if (fsize->index >= ARRAY_SIZE(sp0718_win)) {
		//dev_err(&client->dev, "index(%d) is over range %d\n", fsize->index, ARRAY_SIZE(sp0718_win));
		return -EINVAL;
	}

	switch (fsize->pixel_format) {
		case V4L2_PIX_FMT_YUV420:
		case V4L2_PIX_FMT_YUV422P:
		case V4L2_PIX_FMT_NV12:
		case V4L2_PIX_FMT_YUYV:
		{
			fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
			fsize->discrete.width = sp0718_win[fsize->index]->width;
			fsize->discrete.height = sp0718_win[fsize->index]->height;
			break;
		}
		default:
		{
			dev_err(&client->dev, "pixel_format(%d) is Unsupported\n", fsize->pixel_format);
			return -EINVAL;
		}
	}

	return 0;
}

static int sp0718_enum_frameintervals(struct v4l2_subdev *sd, struct v4l2_frmivalenum *fival)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	const struct sp0718_win_size *win_size = sp0718_select_win(fival->width, fival->height);
	unsigned int array_size = sizeof(win_size->frame_rate_array)/sizeof(unsigned int);
	
	debug_print("frame rate array size %d\n",array_size);
	if (fival->index >= array_size)
		return -EINVAL;
	if ((win_size->width != fival->width) || (win_size->height != fival->height))
	{
		dev_err(&client->dev, "width(%d) height(%d) is over range.\n", fival->width, fival->height);
		return -EINVAL;
	}
	
	fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
	fival->discrete.numerator = 1;
	fival->discrete.denominator = (win_size->frame_rate_array)[fival->index];
	return 0;
}

static int sp0718_enum_fmt(struct v4l2_subdev *sd, unsigned int index,
			   enum v4l2_mbus_pixelcode *code)
{
	if (index >= ARRAY_SIZE(sp0718_cfmts))
		return -EINVAL;

	*code = sp0718_cfmts[index].code;
	return 0;
}

static int sp0718_g_fmt(struct v4l2_subdev *sd,
			struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
debug_print("%s() line %d\n",__FUNCTION__,__LINE__);
	if (!priv->win || !priv->cfmt) {
		u32 width = VGA_WIDTH, height = VGA_HEIGHT;
		int ret = sp0718_set_params(client, &width, &height,
						V4L2_MBUS_FMT_UYVY8_2X8);
		if (ret < 0) {
			return ret;
		}
	}

	mf->width	= priv->win->width;
	mf->height	= priv->win->height;
	mf->code	= priv->cfmt->code;
	mf->colorspace	= priv->cfmt->colorspace;
	mf->field	= V4L2_FIELD_NONE;
	return 0;
}

static int sp0718_try_fmt(struct v4l2_subdev *sd,
			  struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	const struct sp0718_win_size *win;
	int i;

	/*
	 * select suitable win
	 */
	win = sp0718_select_win(mf->width, mf->height);

	mf->width	= win->width;
	mf->height	= win->height;
	mf->field	= V4L2_FIELD_NONE;

	for (i = 0; i < ARRAY_SIZE(sp0718_cfmts); i++)
		if (mf->code == sp0718_cfmts[i].code)
			break;

	if (i == ARRAY_SIZE(sp0718_cfmts)) {
		/* Unsupported format requested. Propose either */
		if (priv->cfmt) {
			/* the current one or */
			mf->colorspace = priv->cfmt->colorspace;
			mf->code = priv->cfmt->code;
		} else {
			/* the default one */
			mf->colorspace = sp0718_cfmts[0].colorspace;
			mf->code = sp0718_cfmts[0].code;
		}
	} else {
		/* Also return the colorspace */
		mf->colorspace	= sp0718_cfmts[i].colorspace;
	}

	return 0;
}

static int sp0718_s_fmt(struct v4l2_subdev *sd,
			struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct sp0718_priv *priv = to_sp0718(client);
	struct soc_camera_device *icd = v4l2_get_subdev_hostdata(sd);
	struct gl5202_camera_param *cam_param = icd->host_priv;
	int ret = sp0718_set_params(client, &mf->width, &mf->height,
					mf->code);
	debug_print("%s() line %d\n",__FUNCTION__,__LINE__);
	
	if (!ret)
		mf->colorspace = priv->cfmt->colorspace;
  cam_param->ahead_line_num = AHEAD_LINE_NUM; 
  
	return ret;
}

static struct v4l2_subdev_video_ops sp0718_subdev_video_ops = {
//	.querystd				= sp0718_querystd,
	.s_stream				= sp0718_s_stream,
	.cropcap				= sp0718_cropcap,
	.g_crop					= sp0718_g_crop,
//	.s_crop					= sp0718_s_crop,
	.g_parm					= sp0718_g_parm,
	.s_parm					= sp0718_s_parm,
	.enum_framesizes		= sp0718_enum_framesizes,
	.enum_frameintervals	= sp0718_enum_frameintervals,
	.enum_mbus_fmt			= sp0718_enum_fmt,
	.g_mbus_fmt				= sp0718_g_fmt,
	.try_mbus_fmt			= sp0718_try_fmt,
	.s_mbus_fmt				= sp0718_s_fmt,
	.g_mbus_config	= sp0718_g_mbus_config,
	.s_mbus_config	= sp0718_s_mbus_config,

};

static struct v4l2_subdev_ops sp0718_subdev_ops = {
	.core	= &sp0718_subdev_core_ops,
	.video	= &sp0718_subdev_video_ops,
};


/*
 * i2c_driver function
 */

static int sp0718_video_probe(struct i2c_client *client)
{
	struct sp0718_priv *priv = to_sp0718(client);
	u8                  pid;
	const char         *devname;

	/*
	 * check and show product ID and manufacturer ID
	 */
	pid = i2c_smbus_read_byte_data(client, PID);
	
	switch (pid) {
	case sp0718:
		devname     = MODULE_NAME;
		priv->model = V4L2_IDENT_SP0718;
		printk("sp0718 get product id ok\n");
		
		break;
	default:
		dev_err(&client->dev,
			"Product ID error %x:\n", pid);
		return -ENODEV;
	}

	//dev_info(&client->dev, "%s Product ID %x\n", devname, pid);

	return v4l2_ctrl_handler_setup(&priv->hdl);
}
static void sp0718_priv_init(struct sp0718_priv * priv)
{
	
	priv->prev_capt_mode = PREVIEW_MODE;	
	priv->autogain = 1;  
	priv->exposure_auto = 1;
	priv->auto_white_balance = 1;
	priv->power_line_frequency = V4L2_CID_POWER_LINE_FREQUENCY_50HZ;
	priv->timeperframe.denominator = 12;
	priv->timeperframe.numerator = 1;
	priv->win = &sp0718_win_vga;
	
	
}

static struct v4l2_ctl_cmd_info v4l2_ctl_array[] =
{
//	{ .id = V4L2_CID_BRIGHTNESS, 
//		.min = 0, 
//		.max = 6, 
//		.step = 1, 
//		.def = 0, },
//	{	.id = V4L2_CID_CONTRAST, 
//		.min = -6, 
//		.max = 6, 
//		.step = 1, 
//		.def = 0, },
//	{ .id = V4L2_CID_SATURATION, 
//		.min = -6, 
//		.max = 6, 
//		.step = 1, 
//		.def = 0, },
	{	.id = V4L2_CID_AUTO_WHITE_BALANCE, 
		.min = 0,
		.max = 1, 
		.step = 1, 
		.def = 1,},
//	{	.id = V4L2_CID_EXPOSURE, 
//		.min = 0, 
//		.max = 0xFFFFU, 
//		.step = 1, 
//		.def = 500,},
//	{	.id = V4L2_CID_AUTOGAIN, 
//		.min = 0, 
//		.max = 1, 
//		.step = 1, 
//		.def = 1,},
//	{	.id = V4L2_CID_GAIN, 
//		.min = 0, .
//		max = 0xFFU, 
//		.step = 1, 
//		.def = 128,},
//	{	.id = V4L2_CID_HFLIP, 
//		.min = 0, 
//		.max = 1, 
//		.step = 1, 
//		.def = 0,},
//	{	.id = V4L2_CID_VFLIP, 
//		.min = 0, 
//		.max = 1, 
//		.step = 1, 
//		.def = 0,},
	{	.id = V4L2_CID_WHITE_BALANCE_TEMPERATURE, 
		.min = 0, 
		.max = 3, 
		.step = 1, 
		.def = 1,},
//	{	.id = V4L2_CID_SHARPNESS, 
//		.min = -2, 
//		.max = 2, 
//		.step = 1, 
//		.def = 0,},
//	{	.id = V4L2_CID_BAND_STOP_FILTER, 
//		.min = 0, 
//		.max = 63, 
//		.step = 1, 
//		.def = 2,},
	{	.id = V4L2_CID_SCENE_EXPOSURE, 
		.min = 0, 
		.max = 1, 
		.step = 1, 
		.def = 0,},//3.4ںûж
	{	.id = V4L2_CID_PRIVATE_PREV_CAPT, 
		.min = 0, 
		.max = 1, 
		.step = 1, 
		.def = PREVIEW_MODE,},//3.4ںûж

};

static struct v4l2_ctl_cmd_info_menu v4l2_ctl_array_menu[] =
{
//	{	.id = V4L2_CID_POWER_LINE_FREQUENCY, 
//		.max = 2, 
//		.mask = 0x0, 
//		.def = 1,},
	{	.id = V4L2_CID_COLORFX, 
		.max = 3, 
		.mask = 0x0, 
		.def = 0,},
	{	.id = V4L2_CID_EXPOSURE_AUTO, 
		.max = 1, 
		.mask = 0x0, 
		.def = 1,},

};




static void sp0718_v4l2_ctrl_handler_init(struct v4l2_ctrl_handler *hdl,
			const struct v4l2_ctrl_ops *ops)
{
	unsigned int cmd_array_size = sizeof(v4l2_ctl_array)/sizeof(struct v4l2_ctl_cmd_info);
	unsigned int cmd_menu_array_size = sizeof(v4l2_ctl_array_menu)/sizeof(struct v4l2_ctl_cmd_info_menu);
	unsigned int i = 0;
	
	debug_print("cmd_array_size %d cmd_menu_array_size %d",cmd_array_size,cmd_menu_array_size);
	
	v4l2_ctrl_handler_init(hdl, cmd_array_size + cmd_menu_array_size);
	
	while(i < cmd_array_size)
	{
		v4l2_ctrl_new_std(hdl, ops,
			(v4l2_ctl_array+i)->id, (v4l2_ctl_array+i)->min, (v4l2_ctl_array+i)->max, (v4l2_ctl_array+i)->step, (v4l2_ctl_array+i)->def);
		i++;
	};
	i = 0;
	while(i < cmd_menu_array_size)
	{
		v4l2_ctrl_new_std_menu(hdl, ops,
			(v4l2_ctl_array_menu+i)->id, (v4l2_ctl_array_menu+i)->max, (v4l2_ctl_array_menu+i)->mask, (v4l2_ctl_array_menu+i)->def);
		i++;
	};	
}


static int sp0718_probe(struct i2c_client *client,
			const struct i2c_device_id *did)
{
	struct sp0718_priv        *priv;
	struct soc_camera_link *icl = soc_camera_i2c_to_link(client);
	struct i2c_adapter        *adapter = to_i2c_adapter(client->dev.parent);
	int                        ret;
  debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	if (!icl) {
		dev_err(&client->dev, "sp0718: Missing platform_data for driver!\n");
		return -EINVAL;
	}
	debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	if (!icl->priv)
		return -EINVAL;
	debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA)) {
		dev_err(&adapter->dev,
			"I2C-Adapter doesn't support "
			"I2C_FUNC_SMBUS_BYTE_DATA\n");
		return -EIO;
	}
debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;
		debug_print("%s line %d\n",__FUNCTION__, __LINE__);
  sp0718_priv_init(priv);
  debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	priv->info = icl->priv;
debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	v4l2_i2c_subdev_init(&priv->subdev, client, &sp0718_subdev_ops);
	sp0718_v4l2_ctrl_handler_init(&priv->hdl, &sp0718_ctrl_ops);
	priv->subdev.ctrl_handler = &priv->hdl;
	debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	if (priv->hdl.error) {
		ret = priv->hdl.error;
		kfree(priv);
		debug_print("%s line %d\n",__FUNCTION__, __LINE__);
		return ret;
	}
debug_print("%s line %d\n",__FUNCTION__, __LINE__);
	ret = sp0718_video_probe(client);
	if (ret) {
		v4l2_ctrl_handler_free(&priv->hdl);
		kfree(priv);
	}
	
  debug_print("i2c driver probe ok\n");
	return ret;
}

static int sp0718_remove(struct i2c_client *client)
{
	struct sp0718_priv *priv = to_sp0718(client);
	
	v4l2_device_unregister_subdev(&priv->subdev);
	v4l2_ctrl_handler_free(&priv->hdl);
	kfree(priv);
	return 0;
}

static const struct i2c_device_id sp0718_id[] = {
	{ MODULE_NAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, sp0718_id);

static struct i2c_driver sp0718_i2c_driver = {
	.driver = {
		.name = MODULE_NAME,
	},
	.probe    = sp0718_probe,
	.remove   = sp0718_remove,
	.id_table = sp0718_id,
};

void get_camera_front_config(struct platform_device *camera_device)
{
	struct soc_camera_link *sp0718_link = camera_device->dev.platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	int h_mirror, v_flip;
	char interface_type[10];
	int bit_num, channel_num, is_3d, host_num,i2c_num;

	
	sp0718_info->flags |= MODULE_FLAG_FRONT;
	sp0718_info->video_devnum = 3; //ǰͷ豸ڵΪvideo3
	get_config("cameraf.h_mirror", &h_mirror, sizeof(h_mirror));
	get_config("cameraf.v_flip", &v_flip,sizeof(v_flip));
//	get_config("cameraf.interface", interface_type, sizeof(interface_type));
//	get_config("cameraf.bit_num", &bit_num, sizeof(bit_num));
//	get_config("cameraf.channel_num", &channel_num, sizeof(channel_num));
//	get_config("cameraf.is_3d", &is_3d, sizeof(is_3d));
//	get_config("cameraf.i2c_num", &i2c_num, sizeof(i2c_num));
//	get_config("cameraf.host_num", &host_num, sizeof(host_num));

//	if (h_mirror != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_HFLIP;
		
//	}	
//	if (v_flip != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_VFLIP;	
//	}	
//	if (!strcmp(interface_type, INTERFACE_PALL))
//	{
		sp0718_info->flags |= MODULE_FLAG_PALL;
//	}
//  else if (!strcmp(interface_type, MODULE_FLAG_MIPI))
//  {
//  	sp0718_info->flags |= MODULE_FLAG_MIPI;
//  }
//  else
//  {
//  	printk(KERN_ERR "no camera module interface match!\n");
//  }
//  switch(bit_num)
// 	{
// 		case 8:
 			sp0718_info->flags |= MODULE_FLAG_8BIT;
// 			break;
// 		case 10:
// 			sp0718_info->flags |= MODULE_FLAG_10BIT;
// 			break;
// 		case 12:
// 			sp0718_info->flags |= MODULE_FLAG_12BIT;
// 			break;
// 		default:
// 			printk(KERN_ERR "camera module interface bit_width error!\n");
// 	}
// 	switch(channel_num)
// 	{
// 		case 0:
// 			sp0718_info->flags |= MODULE_FLAG_CHANNEL1;
// 			break;
// 		case 1:
 			sp0718_info->flags |= MODULE_FLAG_CHANNEL2;
// 			break;
// 		case 2:
// 			sp0718_info->flags |= (MODULE_FLAG_CHANNEL1 | MODULE_FLAG_CHANNEL2);
// 			break;
// 		default:
// 			printk(KERN_ERR "camera module channel error\n");
// 	}
// 	if (is_3d != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_3D;
//	}	
// 	sp0718_link->bus_id = host_num;
// 	sp0718_link->i2c_adapter_id = i2c_num;	
}

void get_camera_back_config(struct platform_device *camera_device)
{
	struct soc_camera_link *sp0718_link = camera_device->dev.platform_data;
	struct module_info *sp0718_info = sp0718_link->priv;
	int h_mirror, v_flip;
	char interface_type[10];
	int bit_num, channel_num, is_3d, host_num,i2c_num;

	sp0718_info->flags |= MODULE_FLAG_BACK;
	sp0718_info->video_devnum = 0; //ͷ豸ڵΪvideo0
	
	get_config("camerab.h_mirror", &h_mirror, sizeof(h_mirror));
	get_config("camerab.v_flip", &v_flip,sizeof(v_flip));
//	get_config("camerab.interface", interface_type, sizeof(interface_type));
//	get_config("camerab.bit_num", &bit_num, sizeof(bit_num));
//	get_config("camerab.channel_num", &channel_num, sizeof(channel_num));
//	get_config("camerab.is_3d", &is_3d, sizeof(is_3d));
//	get_config("camerab.i2c_num", &i2c_num, sizeof(i2c_num));
//	get_config("camerab.host_num", &host_num, sizeof(host_num));

//	if (h_mirror != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_HFLIP;
//	}	
//	if (v_flip != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_VFLIP;
//	}	
//	if (!strcmp(interface_type, INTERFACE_PALL))
//	{
		sp0718_info->flags |= MODULE_FLAG_PALL;
//	}
//  else if (!strcmp(interface_type, MODULE_FLAG_MIPI))
//  {
//  	sp0718_info->flags |= MODULE_FLAG_MIPI;
//  }
//  else
//  {
//  	printk(KERN_ERR "no camera module interface match!\n");
//  }
//  switch(bit_num)
// 	{
// 		case 8:
 			sp0718_info->flags |= MODULE_FLAG_8BIT;
// 			break;
// 		case 10:
// 			sp0718_info->flags |= MODULE_FLAG_10BIT;
// 			break;
// 		case 12:
// 			sp0718_info->flags |= MODULE_FLAG_12BIT;
// 			break;
// 		default:
// 			printk(KERN_ERR "camera module interface bit_width error!\n");
// 	}
// 	switch(channel_num)
// 	{
// 		case 0:
//sp0718_info->flags |= MODULE_FLAG_CHANNEL1;
// 			break;
// 		case 1:
			sp0718_info->flags |= MODULE_FLAG_CHANNEL2;
// 			break;
// 		case 2:
// 			sp0718_info->flags |= (MODULE_FLAG_CHANNEL1 | MODULE_FLAG_CHANNEL2);
// 			break;
// 		default:
// 			printk(KERN_ERR "camera module channel error\n");
// 	}
// 	if (is_3d != 0)
//	{
//		sp0718_info->flags |= MODULE_FLAG_3D;
//	}	
// 	sp0718_link->bus_id = 0;
// 	sp0718_link->i2c_adapter_id = 2;

}

static void save_modules_param(void)
{
	get_config("camerab.h_mirror", (char *)(&back_hmirror), sizeof(back_hmirror));
	get_config("camerab.v_flip", (char *)(&back_vflip),sizeof(back_vflip));
	get_config("cameraf.h_mirror", (char *)(&front_hmirror), sizeof(front_hmirror));
	get_config("cameraf.v_flip", (char *)(&front_vflip),sizeof(front_vflip));
	back_front_flag = 98; //ǰģͺͬʱֻһ豸ڵ㣬̶Ϊvideo0back_front_flagΪ98ʾĿǰΪ
	
	printk("front mirror %d flip %d; back mirror %d flip %d\n",front_hmirror, front_vflip, back_hmirror, back_vflip);
	

}

void get_camera_config(struct platform_device *camera_device)
{
	char module_name[20];
	get_config("cameraf.module_name", module_name, sizeof(module_name));
	if (!strcmp(module_name, MODULE_NAME))
	{
		memset(&module_name, 0, sizeof(module_name));  //ʼmodule_name
		get_config("camerab.module_name", module_name, sizeof(module_name));
		if (strcmp(module_name, MODULE_NAME))
		{	
			get_camera_front_config(camera_device);
			printk(KERN_ERR "sp0718 get front config\n");
	  }
	  else
	  {
	  	get_camera_back_config(camera_device);
			printk(KERN_ERR "sp0718 get Back config1\n");
			is_two_same_module = 1;
			save_modules_param();
	  }
	}
	else
	{
		memset(&module_name, 0, sizeof(module_name));  //ʼmodule_name
		get_config("camerab.module_name", module_name, sizeof(module_name));
		if (!strcmp(module_name, MODULE_NAME))
		{
			get_camera_back_config(camera_device);
			printk(KERN_ERR "sp0718 get Back config\n");
		}
		else
		{
			printk(KERN_ERR "no module name match config name\n");
		}
		
	}
}

/*
 * module function
 */
static int __init sp0718_module_init(void)
{
	int ret;
	get_camera_config(&asoc_camera_device);
	ret = platform_device_register(&asoc_camera_device);
	
	if (is_two_same_module)
	{
		ret |= platform_device_register(&two_module_flag_device);
		device_create_file(&(two_module_flag_device.dev), &dev_attr_front_back);
	}
	
	return ret |= i2c_add_driver(&sp0718_i2c_driver);
}

static void __exit sp0718_module_exit(void)
{
	i2c_del_driver(&sp0718_i2c_driver);
	
	if (is_two_same_module)
	{
		device_remove_file(&(two_module_flag_device.dev), &dev_attr_front_back);
		platform_device_unregister(&two_module_flag_device);
	}
	
	platform_device_unregister(&asoc_camera_device);
}

module_init(sp0718_module_init);
module_exit(sp0718_module_exit);

MODULE_DESCRIPTION("SoC Camera driver for sp0718");
MODULE_AUTHOR("Wang Xin (wangxin@actions-semi.com)");
MODULE_LICENSE("GPL v2");
