/*****************************************************************************
*  Copyright Statement:
*  --------------------
*  This software is protected by Copyright and the information contained
*  herein is confidential. The software may not be copied and the information
*  contained herein may not be used or disclosed except with the written
*  permission of MediaTek Inc. (C) 2010
*
*  BY OPENING THIS FILE, BUYER HEREBY UNEQUIVOCALLY ACKNOWLEDGES AND AGREES
*  THAT THE SOFTWARE/FIRMWARE AND ITS DOCUMENTATIONS ("MEDIATEK SOFTWARE")
*  RECEIVED FROM MEDIATEK AND/OR ITS REPRESENTATIVES ARE PROVIDED TO BUYER ON
*  AN "AS-IS" BASIS ONLY. MEDIATEK EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES,
*  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
*  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NONINFRINGEMENT.
*  NEITHER DOES MEDIATEK PROVIDE ANY WARRANTY WHATSOEVER WITH RESPECT TO THE
*  SOFTWARE OF ANY THIRD PARTY WHICH MAY BE USED BY, INCORPORATED IN, OR
*  SUPPLIED WITH THE MEDIATEK SOFTWARE, AND BUYER AGREES TO LOOK ONLY TO SUCH
*  THIRD PARTY FOR ANY WARRANTY CLAIM RELATING THERETO. MEDIATEK SHALL ALSO
*  NOT BE RESPONSIBLE FOR ANY MEDIATEK SOFTWARE RELEASES MADE TO BUYER'S
*  SPECIFICATION OR TO CONFORM TO A PARTICULAR STANDARD OR OPEN FORUM.
*
*  BUYER'S SOLE AND EXCLUSIVE REMEDY AND MEDIATEK'S ENTIRE AND CUMULATIVE
*  LIABILITY WITH RESPECT TO THE MEDIATEK SOFTWARE RELEASED HEREUNDER WILL BE,
*  AT MEDIATEK'S OPTION, TO REVISE OR REPLACE THE MEDIATEK SOFTWARE AT ISSUE,
*  OR REFUND ANY SOFTWARE LICENSE FEES OR SERVICE CHARGE PAID BY BUYER TO
*  MEDIATEK FOR SUCH MEDIATEK SOFTWARE AT ISSUE.
*
*  THE TRANSACTION CONTEMPLATED HEREUNDER SHALL BE CONSTRUED IN ACCORDANCE
*  WITH THE LAWS OF THE STATE OF CALIFORNIA, USA, EXCLUDING ITS CONFLICT OF
*  LAWS PRINCIPLES.  ANY DISPUTES, CONTROVERSIES OR CLAIMS ARISING THEREOF AND
*  RELATED THERETO SHALL BE SETTLED BY ARBITRATION IN SAN FRANCISCO, CA, UNDER
*  THE RULES OF THE INTERNATIONAL CHAMBER OF COMMERCE (ICC).
*
*****************************************************************************/

/*****************************************************************************
 *
 * Filename:
 * ---------
 *   image_sensor_SP0718.c
 *
 * Project:
 * --------
 *  
 *
 * Description:
 * ------------
 *  Image sensor driver function
 *  V1.0.0
 *
 * Author:
 * -------
 * Mormo
 *
 *============================================================================
 *             HISTORY
 * Below this line, this part is controlled by PVCS VM. DO NOT MODIFY!!
 *------------------------------------------------------------------------------
 *  * $Log$
 * 2011/06/13 Firstly Released By Mormo;
 *------------------------------------------------------------------------------
 * Upper this line, this part is controlled by PVCS VM. DO NOT MODIFY!!
 *============================================================================
 ****************************************************************************/
#include "drv_comm.h"
#include "isp_comm_if.h"
#include "camera_sccb.h"
#include "image_sensor.h"
#include "cal_comm_def.h"
#include "isp_if.h"

/* #define __SP0718_DEBUG_TRACE__ */
#define SP0718_SET_PAGE0 CamWriteCmosSensor(0xfd , 0x00)
#define SP0718_SET_PAGE1 CamWriteCmosSensor(0xfd , 0x01)
#define SP0718_SET_PAGE2 CamWriteCmosSensor(0xfd , 0x02) 
#define SP0718_NORMAL_Y0ffset  0x20
#define SP0718_LOWLIGHT_Y0ffset  0x25
//AE target
#define  SP0718_P1_0xeb  0x78
#define  SP0718_P1_0xec  0x6c
//HEQ
#define  SP0718_P1_0x10  0x00//outdoor
#define  SP0718_P1_0x14  0x20
#define  SP0718_P1_0x11  0x00//nr
#define  SP0718_P1_0x15  0x18
#define  SP0718_P1_0x12  0x00//dummy
#define  SP0718_P1_0x16  0x10
#define  SP0718_P1_0x13  0x00//low
#define  SP0718_P1_0x17  0x00

kal_bool SP0718_CAM_BANDING_50HZ = KAL_FALSE;
kal_bool SP0718_NIGHT_ENABLE = KAL_FALSE;
kal_bool SP0718_video_mode = KAL_FALSE;
static struct
{
  kal_bool BypassAe;
  kal_bool BypassAwb;
  kal_uint8 Ctrl3A;
  kal_uint8 AecMode3;
  kal_uint16 VideoFps;
  kal_uint32 IntClk; /* internal clock = half of pclk */
  kal_uint16 FrameHeight;
  kal_uint16 LineLength;
  IMAGE_SENSOR_INDEX_ENUM SensorIdx;
  sensor_data_struct *NvramData;
} SP0718Sensor;

/*************************************************************************
* FUNCTION
*    SP0718HalfAdjust
*
* DESCRIPTION
*    This function dividend / divisor and use round-up.
*
* PARAMETERS
*    DividEnd
*    Divisor
*
* RETURNS
*    [DividEnd / Divisor]
*
* LOCAL AFFECTED
*
*************************************************************************/
__inline static kal_uint32 SP0718HalfAdjust(kal_uint32 DividEnd, kal_uint32 Divisor)
{
  return (DividEnd + (Divisor >> 1)) / Divisor; /* that is [DividEnd / Divisor + 0.5]*/
}

/*************************************************************************
* FUNCTION
*    SP0718SetMirror
*
* DESCRIPTION
*    This function set the Mirror to the CMOS sensor
*
* PARAMETERS
*    Mirror
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718SetMirror(kal_uint8 Mirror)
{SP0718_SET_PAGE0;
  switch (Mirror)
  {
  case IMAGE_SENSOR_MIRROR_H:
    CamWriteCmosSensor(0x31, 0x30);
    break;
  case IMAGE_SENSOR_MIRROR_V:
    CamWriteCmosSensor(0x31, 0x50);
    break;
  case IMAGE_SENSOR_MIRROR_HV:
    CamWriteCmosSensor(0x31, 0x70);
    break;
  default:
    CamWriteCmosSensor(0x31, 0x10);
  }
}

/*************************************************************************
* FUNCTION
*    SP0718SetClock
*
* DESCRIPTION
*    This function set sensor vt clock and op clock
*
* PARAMETERS
*    Clk: vt clock
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718SetClock(kal_uint32 Clk)
{
  kal_uint8 ClkDiv;
  
  if (SP0718Sensor.IntClk == Clk)
  {
    return;
  }
  SP0718Sensor.IntClk = Clk;
  switch (SP0718Sensor.IntClk)
  {
  case SP0718_MCLK / 2: ClkDiv = 0x01; break; /* pclk = mclk */
  case SP0718_MCLK / 4: ClkDiv = 0x02; break; /* pclk = mclk / 2 */
  case SP0718_MCLK / 8: ClkDiv = 0x03; break; /* pclk = mclk / 4 */
  default: ASSERT(0);
  }
  SP0718_SET_PAGE0;
  CamWriteCmosSensor(0x30, ClkDiv);
}

/*************************************************************************
* FUNCTION
*    SP0718WriteShutter
*
* DESCRIPTION
*    This function apply Shutter to sensor
*
* PARAMETERS
*    Shutter: integration time
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718WriteShutter(kal_uint32 Shutter)
{SP0718_SET_PAGE0;
  CamWriteCmosSensor(0x03, Shutter >> 8);
  CamWriteCmosSensor(0x04, Shutter);
}

/*************************************************************************
* FUNCTION
*    SP0718ReadShutter
*
* DESCRIPTION
*    This function get shutter from sensor
*
* PARAMETERS
*    None
*
* RETURNS
*    Shutter: integration time
*
* LOCAL AFFECTED
*
*************************************************************************/
static kal_uint16 SP0718ReadShutter(void)
{SP0718_SET_PAGE0;
  return (CamReadCmosSensor(0x03) << 8)|CamReadCmosSensor(0x04);
}

/*************************************************************************
* FUNCTION
*    SP0718AeEnable
*
* DESCRIPTION
*    disable/enable AE
*
* PARAMETERS
*    Enable
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718AeEnable(kal_bool Enable)
{
  const kal_bool AeEnable = (SP0718Sensor.AecMode3&0x05) ? KAL_TRUE : KAL_FALSE;
  
  if (SP0718Sensor.BypassAe)
  {
    Enable = KAL_FALSE;
  }
  if (AeEnable == Enable)
  {
    return;
  }
  if (Enable)
  {
    SP0718Sensor.AecMode3 |= 0x05;
  }
  else
  {
    SP0718Sensor.AecMode3 &= 0x10;
  }
  SP0718_SET_PAGE1;
  CamWriteCmosSensor(0x32, SP0718Sensor.AecMode3);
}

/*************************************************************************
* FUNCTION
*    SP0718AwbEnable
*
* DESCRIPTION
*    disable/enable awb
*
* PARAMETERS
*    Enable
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718AwbEnable(kal_bool Enable)
{
  const kal_bool AwbEnable = (SP0718Sensor.Ctrl3A&0x10) ? KAL_TRUE : KAL_FALSE;
  
  if (SP0718Sensor.BypassAwb)
  {
    Enable = KAL_FALSE;
  }
  if (AwbEnable == Enable)
  {
    return;
  }
  if (Enable)
  {
    SP0718Sensor.Ctrl3A |= 0x10;
  }
  else
  {
    SP0718Sensor.Ctrl3A &= 0x05;
  }
    SP0718_SET_PAGE1;
  CamWriteCmosSensor(0x32, SP0718Sensor.Ctrl3A);
}

/*************************************************************************
* FUNCTION
*    SP0718SetBaseShutter
*
* DESCRIPTION
*    This function set base shutter to the CMOS sensor
*
* PARAMETERS
*    Banding: 50Hz or 60Hz
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718SetBaseShutter(kal_uint8 Banding)
{
}

/*************************************************************************
* FUNCTION
*    SP0718SetDummy
*
* DESCRIPTION
*    This function add DummyPixel and DummyLine.
*
* PARAMETERS
*    DummyPixel
*    DummyLine
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718SetDummy(kal_uint16 DummyPixel, kal_uint16 DummyLine)
{
}

/*************************************************************************
* FUNCTION
*    SP0718SetVideoFps
*
* DESCRIPTION
*    This function calculate & fix frame rate in video mode
*
* PARAMETERS
*    Fps: video frame rate, base on SP0718_FPS(x)
*
* RETURNS
*    Fps: real frame rate set by sensor
*
* LOCAL AFFECTED
*
*************************************************************************/
static kal_uint16 SP0718SetVideoFps(kal_uint16 Fps)
{
}

/*************************************************************************
* FUNCTION
*    SP0718SetMinFps
*
* DESCRIPTION
*    This function calculate & set min frame rate in camera preview mode
*
* PARAMETERS
*    Fps: min frame rate, base on SP0718_FPS(x)
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718SetMinFps(kal_uint16 Fps)
{
}

/*************************************************************************
* FUNCTION
*    SP0718Ev
*
* DESCRIPTION
*    Exposure setting.
*
* PARAMETERS
*    In, Out
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718Ev(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    const P_CAL_FEATURE_TYPE_ENUM_STRUCT Exposure = &Out->FeatureInfo.FeatureEnum;
    
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_VIDEO_SUPPORT;
    Exposure->IsSupport = KAL_TRUE;
    Exposure->ItemCount = 10;
    Exposure->SupportItem[0] = CAM_EV_NEG_4_3;
    Exposure->SupportItem[1] = CAM_EV_NEG_3_3;
    Exposure->SupportItem[2] = CAM_EV_NEG_2_3;
    Exposure->SupportItem[3] = CAM_EV_NEG_1_3;
    Exposure->SupportItem[4] = CAM_EV_ZERO;
    Exposure->SupportItem[5] = CAM_EV_POS_1_3;
    Exposure->SupportItem[6] = CAM_EV_POS_2_3;
    Exposure->SupportItem[7] = CAM_EV_POS_3_3;
    Exposure->SupportItem[8] = CAM_EV_POS_4_3;
    Exposure->SupportItem[9] = CAM_EV_NIGHT_SHOT;
  }
  else if (CAL_FEATURE_SET_OPERATION == In->FeatureCtrlCode)
  {
    static const kal_uint8 Data[9][1]=
    {
      {0xc0}, /* EV -4 */
      {0xd0}, /* EV -3 */
      {0xe0}, /* EV -2 */
      {0xf0}, /* EV -1 */
      {0x00}, /* EV 0 */
      {0x10}, /* EV +1 */
      {0x20}, /* EV +2 */
      {0x30}, /* EV +3 */
      {0x40}, /* EV +4 */
    };
    kal_uint8 i;
    
    switch (In->FeatureSetValue)
    {
    case CAM_EV_NEG_4_3: i = 0; break;
    case CAM_EV_NEG_3_3: i = 1; break;
    case CAM_EV_NEG_2_3: i = 2; break;
    case CAM_EV_NEG_1_3: i = 3; break;
    case CAM_EV_ZERO:    i = 4; break;
    case CAM_EV_POS_1_3: i = 5; break;
    case CAM_EV_POS_2_3: i = 6; break;
    case CAM_EV_POS_3_3: i = 7; break;
    case CAM_EV_POS_4_3: i = 8; break;
    default:
      return MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
    }
    SP0718_SET_PAGE1;
    CamWriteCmosSensor(0xdb, Data[i][0]); /* YBright */
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718Wb
*
* DESCRIPTION
*    Wb setting.
*
* PARAMETERS
*    In, Out
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718Wb(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    const P_CAL_FEATURE_TYPE_ENUM_STRUCT Wb = &Out->FeatureInfo.FeatureEnum;
    
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_VIDEO_SUPPORT;
    Wb->IsSupport = KAL_TRUE;
    Wb->ItemCount = 6;
    Wb->SupportItem[0] = CAM_WB_AUTO;
    Wb->SupportItem[1] = CAM_WB_CLOUD;
    Wb->SupportItem[2] = CAM_WB_DAYLIGHT;
    Wb->SupportItem[3] = CAM_WB_INCANDESCENCE;
    Wb->SupportItem[4] = CAM_WB_FLUORESCENT;
    Wb->SupportItem[5] = CAM_WB_TUNGSTEN;
  }
  else if (CAL_FEATURE_SET_OPERATION == In->FeatureCtrlCode)
  {
    static const kal_uint8 AwbGain[5][2]=
    { /* R gain, B gain. base: 2.6bits, 0x40=1x,0x60=1.5x */
     {0xdc,0x75}, /* cloud */
     {0xc8,0x89}, /* daylight */
     {0xaa,0xce}, /* INCANDESCENCE *///ȹ 
     {0x91,0xc8}, /* FLUORESCENT */ //ӫƺչѶдĲͬ4200-5000
     {0x75,0xe2}, /* TUNGSTEN */   //˿//2800K
    };
    static kal_uint8 BakRGain, BakBGain;
    kal_uint8 RGain, BGain;
    kal_int8 i = sizeof(AwbGain) / sizeof(AwbGain[0]);
    SP0718_SET_PAGE2;
    RGain = CamReadCmosSensor(0x26);
    BGain = CamReadCmosSensor(0x27);
    while (--i >= 0)
    {
      if (RGain == AwbGain[i][0] && BGain == AwbGain[i][1])
      {
        break;
      }
    }
    if (i < 0)
    {
      BakRGain = RGain;
      BakBGain = BGain;
    }
    switch (In->FeatureSetValue)
    {
    case CAM_WB_AUTO:
      SP0718_SET_PAGE2;	
      CamWriteCmosSensor(0x26, BakRGain); /* for AWB can adjust back */
      CamWriteCmosSensor(0x27, BakBGain);
      SP0718_SET_PAGE1;
      SP0718AwbEnable(KAL_TRUE);
      return MM_ERROR_NONE;
    case CAM_WB_CLOUD:         i = 0; break;
    case CAM_WB_DAYLIGHT:      i = 1; break;
    case CAM_WB_INCANDESCENCE: i = 2; break;
    case CAM_WB_FLUORESCENT:   i = 3; break;
    case CAM_WB_TUNGSTEN:      i = 4; break;
    default:
      return MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
    }
    SP0718AwbEnable(KAL_FALSE);
    SP0718_SET_PAGE2;	
    CamWriteCmosSensor(0x26, AwbGain[i][0]); /* awb R gain */
    CamWriteCmosSensor(0x27, AwbGain[i][1]); /* awb B gain */
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718Effect
*
* DESCRIPTION
*    Effect setting.
*
* PARAMETERS
*    In, Out
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718Effect(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    const P_CAL_FEATURE_TYPE_ENUM_STRUCT Effect = &Out->FeatureInfo.FeatureEnum;
    
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_VIDEO_SUPPORT;
    Effect->IsSupport = KAL_TRUE;
    Effect->ItemCount = 7;
    Effect->SupportItem[0] = CAM_EFFECT_ENC_NORMAL;
    Effect->SupportItem[1] = CAM_EFFECT_ENC_GRAYSCALE;
    Effect->SupportItem[2] = CAM_EFFECT_ENC_SEPIA;
    Effect->SupportItem[3] = CAM_EFFECT_ENC_SEPIAGREEN;
    Effect->SupportItem[4] = CAM_EFFECT_ENC_SEPIABLUE;
    Effect->SupportItem[5] = CAM_EFFECT_ENC_COLORINV;
    Effect->SupportItem[6] = CAM_EFFECT_ENC_GRAYINV;
  }
  else if (CAL_FEATURE_SET_OPERATION == In->FeatureCtrlCode)
  {
    static const kal_uint8 Data[7][3]=
    {
      {0x00,0x80,0x80}, /* NORMAL */
      {0x20,0x80,0x80}, /* GRAYSCALE */
      {0x10,0xc0,0x20}, /* SEPIA */
      {0x10,0x20,0x20}, /* SEPIAGREEN */
      {0x10,0x20,0xf0}, /* SEPIABLUE */
      {0x04,0x80,0x80}, /* COLORINV */
      {0x08,0x80,0x80}, /* GRAYINV */
    };
    kal_uint8 i;
    
    switch (In->FeatureSetValue)
    {
    case CAM_EFFECT_ENC_NORMAL:     i = 0; break;
    case CAM_EFFECT_ENC_GRAYSCALE:  i = 1; break;
    case CAM_EFFECT_ENC_SEPIA:      i = 2; break;
    case CAM_EFFECT_ENC_SEPIAGREEN: i = 3; break;
    case CAM_EFFECT_ENC_SEPIABLUE:  i = 4; break;
    case CAM_EFFECT_ENC_COLORINV:   i = 5; break;
    case CAM_EFFECT_ENC_GRAYINV:    i = 6; break;
    default:
      return MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
    }
    SP0718_SET_PAGE1;
    CamWriteCmosSensor(0x66, Data[i][0]); /* special effect */
    CamWriteCmosSensor(0x67, Data[i][1]); /* fixed Cb */
    CamWriteCmosSensor(0x68, Data[i][2]); /* fixed Cr */
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718StillCaptureSize
*
* DESCRIPTION
*    This function get capture size
*
* PARAMETERS
*    In, Out
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
MM_ERROR_CODE_ENUM SP0718StillCaptureSize(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  const P_CAL_FEATURE_TYPE_ENUM_STRUCT CapSize = &Out->FeatureInfo.FeatureEnum;
  
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_SUPPORT;
    CapSize->IsSupport = KAL_TRUE;
    CapSize->ItemCount = 2;
    CapSize->SupportItem[0] = CAM_IMAGE_SIZE_WALLPAPER;
    CapSize->SupportItem[1] = CAM_IMAGE_SIZE_VGA;
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718Banding
*
* DESCRIPTION
*    This function get banding setting
*
* PARAMETERS
*    In, Out
*
* RETURNS
*    error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718Banding(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  const P_CAL_FEATURE_TYPE_ENUM_STRUCT Banding = &Out->FeatureInfo.FeatureEnum;
  
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_VIDEO_SUPPORT;
    Banding->IsSupport = KAL_TRUE;
    Banding->ItemCount = 2;
    Banding->SupportItem[0] = CAM_BANDING_50HZ;
    Banding->SupportItem[1] = CAM_BANDING_60HZ;
  }
  else if (CAL_FEATURE_SET_OPERATION == In->FeatureCtrlCode)
  {
	switch (In->FeatureSetValue)
	{
		case CAM_BANDING_50HZ:
		SP0718_CAM_BANDING_50HZ = KAL_TRUE;
		break;
		
	case CAM_BANDING_60HZ:
		SP0718_CAM_BANDING_50HZ = KAL_FALSE;
		break;

	default:
       	return MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
			
	}

	if(SP0718_video_mode == KAL_FALSE) //CAM 
	{
		if(SP0718_NIGHT_ENABLE == KAL_FALSE)  //DALLIGHT
		{
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xcd,SP0718_NORMAL_Y0ffset);
			CamWriteCmosSensor(0xce,0x1f);

			if(SP0718_CAM_BANDING_50HZ == KAL_TRUE)  //50HZ
			{

			//capture preview daylight 24M 50hz 15-8FPS maxgain:  
			  
			CamWriteCmosSensor(0xfd,0x00);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x04,0xce);
			CamWriteCmosSensor(0x06,0x00);
			CamWriteCmosSensor(0x09,0x02);
			CamWriteCmosSensor(0x0a,0xc4);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xef,0x4d);
			CamWriteCmosSensor(0xf0,0x00);
			CamWriteCmosSensor(0x02,0x0c);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x06,0x47);
			CamWriteCmosSensor(0x07,0x00);
			CamWriteCmosSensor(0x08,0x01);
			CamWriteCmosSensor(0x09,0x00);
		//Status                            
			CamWriteCmosSensor(0xfd,0x02);
			CamWriteCmosSensor(0xbe,0x9c);
			CamWriteCmosSensor(0xbf,0x03);
			CamWriteCmosSensor(0xd0,0x9c);
			CamWriteCmosSensor(0xd1,0x03);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x5b,0x03);
			CamWriteCmosSensor(0x5c,0x9c);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x32,0x15);
 
					dbg_print("SP0718_CAM_DAY_50HZ  \r\n");
					}
			else  //60HZ
			{
			//capture preview daylight 24M 60hz 15-8FPS maxgain: 
			CamWriteCmosSensor(0xfd,0x00);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x04,0x80);
			CamWriteCmosSensor(0x06,0x00);
			CamWriteCmosSensor(0x09,0x02);
			CamWriteCmosSensor(0x0a,0xc9);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xef,0x40);
			CamWriteCmosSensor(0xf0,0x00);
			CamWriteCmosSensor(0x02,0x0f);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x06,0x3a);
			CamWriteCmosSensor(0x07,0x00);
			CamWriteCmosSensor(0x08,0x01);
			CamWriteCmosSensor(0x09,0x00);
		//Status                            
			CamWriteCmosSensor(0xfd,0x02);
			CamWriteCmosSensor(0xbe,0xc0);
			CamWriteCmosSensor(0xbf,0x03);
			CamWriteCmosSensor(0xd0,0xc0);
			CamWriteCmosSensor(0xd1,0x03);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x5b,0x03);
			CamWriteCmosSensor(0x5c,0xc0);	
			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);
						
			dbg_print("SP0718_CAM_DAY_60HZ  \r\n");

			}
		}
		else                        // night
		{

			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xcd,SP0718_LOWLIGHT_Y0ffset);
			CamWriteCmosSensor(0xce,0x1f);

		
			if(SP0718_CAM_BANDING_50HZ == KAL_TRUE)  //50HZ
			{

			//capture preview night 24M 50hz 15-6FPS maxgain:
		        CamWriteCmosSensor(0xfd,0x00);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x04,0xce);
			CamWriteCmosSensor(0x06,0x00);
			CamWriteCmosSensor(0x09,0x02);
			CamWriteCmosSensor(0x0a,0xc4);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xef,0x4d);
			CamWriteCmosSensor(0xf0,0x00);
			CamWriteCmosSensor(0x02,0x10);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x06,0x47);
			CamWriteCmosSensor(0x07,0x00);
			CamWriteCmosSensor(0x08,0x01);
			CamWriteCmosSensor(0x09,0x00);
		//Status                            
			CamWriteCmosSensor(0xfd,0x02);
			CamWriteCmosSensor(0xbe,0xd0);
			CamWriteCmosSensor(0xbf,0x04);
			CamWriteCmosSensor(0xd0,0xd0);
			CamWriteCmosSensor(0xd1,0x04);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x5b,0x04);
			CamWriteCmosSensor(0x5c,0xd0);
			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);
			  dbg_print("SP0718_CAM_NIGHT_50HZ  \r\n");
			}
			else  //60HZ
			{
			//capture preview night 24M 60hz 15-6FPS maxgain:
		        CamWriteCmosSensor(0xfd,0x00);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x04,0x80);
			CamWriteCmosSensor(0x06,0x00);
			CamWriteCmosSensor(0x09,0x02);
			CamWriteCmosSensor(0x0a,0xc9);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xef,0x40);
			CamWriteCmosSensor(0xf0,0x00);
			CamWriteCmosSensor(0x02,0x14);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x06,0x3a);
			CamWriteCmosSensor(0x07,0x00);
			CamWriteCmosSensor(0x08,0x01);
			CamWriteCmosSensor(0x09,0x00);
		//Status                            
			CamWriteCmosSensor(0xfd,0x02);
			CamWriteCmosSensor(0xbe,0x00);
			CamWriteCmosSensor(0xbf,0x05);
			CamWriteCmosSensor(0xd0,0x00);
			CamWriteCmosSensor(0xd1,0x05);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x5b,0x05);
			CamWriteCmosSensor(0x5c,0x00);
			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);
			  dbg_print("SP0718_CAM_NIGHT_60HZ  \r\n");
			}
		}
	}
	else          //video 
	{
		if(SP0718_NIGHT_ENABLE == KAL_FALSE)  //DALLIGHT
		{

			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xcd,SP0718_NORMAL_Y0ffset);
			CamWriteCmosSensor(0xce,0x1f);

		
			if(SP0718_CAM_BANDING_50HZ == KAL_TRUE)  //50HZ
			{
						//Video record daylight 24M 50hz 14-14FPS maxgain
						
			CamWriteCmosSensor(0xfd,0x00);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x04,0xb0);
			CamWriteCmosSensor(0x06,0x00);
			CamWriteCmosSensor(0x09,0x03);
			CamWriteCmosSensor(0x0a,0x31);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xef,0x48);
			CamWriteCmosSensor(0xf0,0x00);
			CamWriteCmosSensor(0x02,0x07);
			CamWriteCmosSensor(0x03,0x01);
			CamWriteCmosSensor(0x06,0x42);
			CamWriteCmosSensor(0x07,0x00);
			CamWriteCmosSensor(0x08,0x01);
			CamWriteCmosSensor(0x09,0x00);
			//Status                            
			CamWriteCmosSensor(0xfd,0x02);
			CamWriteCmosSensor(0xbe,0xf8);
			CamWriteCmosSensor(0xbf,0x01);
			CamWriteCmosSensor(0xd0,0xf8);
			CamWriteCmosSensor(0xd1,0x01);
			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0x5b,0x01);
			CamWriteCmosSensor(0x5c,0xf8);            
			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);        
					dbg_print("SP0718_VID_DAY_50HZ  \r\n");
				}
				else  //60HZ
				{
					//Video record daylight 24M 60Hz 14-14FPS maxgain:
			CamWriteCmosSensor(0xfd,0x00);                              
			CamWriteCmosSensor(0x03,0x01);      
			CamWriteCmosSensor(0x04,0x68);      
			CamWriteCmosSensor(0x06,0x00);      
			CamWriteCmosSensor(0x09,0x03);      
			CamWriteCmosSensor(0x0a,0x31);      
			CamWriteCmosSensor(0xfd,0x01);      
			CamWriteCmosSensor(0xef,0x3c);      
			CamWriteCmosSensor(0xf0,0x00);      
			CamWriteCmosSensor(0x02,0x08);      
			CamWriteCmosSensor(0x03,0x01);      
			CamWriteCmosSensor(0x06,0x36);      
			CamWriteCmosSensor(0x07,0x00);      
			CamWriteCmosSensor(0x08,0x01);      
			CamWriteCmosSensor(0x09,0x00);      
			//Status                            
			CamWriteCmosSensor(0xfd,0x02);      
			CamWriteCmosSensor(0xbe,0xe0);      
			CamWriteCmosSensor(0xbf,0x01);      
			CamWriteCmosSensor(0xd0,0xe0);      
			CamWriteCmosSensor(0xd1,0x01);      
			CamWriteCmosSensor(0xfd,0x01);      
			CamWriteCmosSensor(0x5b,0x01);      
			CamWriteCmosSensor(0x5c,0xe0); 
			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);
					  dbg_print("SP0718_VID_DAY_60HZ  \r\n");
					}

		}
		else                        // night
		{

			CamWriteCmosSensor(0xfd,0x01);
			CamWriteCmosSensor(0xcd,SP0718_LOWLIGHT_Y0ffset);
			CamWriteCmosSensor(0xce,0x1f);

			if(SP0718_CAM_BANDING_50HZ == KAL_TRUE)  //50HZ
			{
				//SI13_SP0718 24M 1Ƶ 50Hz 8.0392-8fps AE_Parameters_ maxgain
			dbg_print("SP0718_VID_NIGTHT_50HZ  \r\n");
			CamWriteCmosSensor(0xfd,0x00); 
  			CamWriteCmosSensor(0x03,0x00); 
  			CamWriteCmosSensor(0x04,0xf6); 
  			CamWriteCmosSensor(0x06,0x00); 
  			CamWriteCmosSensor(0x09,0x08); 
  			CamWriteCmosSensor(0x0a,0x1d); 
  			CamWriteCmosSensor(0xfd,0x01); 
  			CamWriteCmosSensor(0xef,0x29); 
  			CamWriteCmosSensor(0xf0,0x00); 
  			CamWriteCmosSensor(0x02,0x0c); 
  			CamWriteCmosSensor(0x03,0x01); 
  			CamWriteCmosSensor(0x06,0x23); 
  			CamWriteCmosSensor(0x07,0x00); 
  			CamWriteCmosSensor(0x08,0x01); 
  			CamWriteCmosSensor(0x09,0x00); 
  			//Status                       
  			CamWriteCmosSensor(0xfd,0x02); 
  			CamWriteCmosSensor(0xbe,0xec); 
  			CamWriteCmosSensor(0xbf,0x01); 
  			CamWriteCmosSensor(0xd0,0xec); 
  			CamWriteCmosSensor(0xd1,0x01); 
  			CamWriteCmosSensor(0xfd,0x01); 
  			CamWriteCmosSensor(0x5b,0x01); 
  			CamWriteCmosSensor(0x5c,0xec); 
  			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);
					
			}
			else  //60HZ
			{
				//SI13_SP0718 24M 1Ƶ 60Hz 8-8fps AE_Parameters_ maxgain
				dbg_print("SP0718_VID_NIGTHT_60HZ  \r\n");
		    CamWriteCmosSensor(0xfd,0x00); 
  			CamWriteCmosSensor(0x03,0x00); 
  			CamWriteCmosSensor(0x04,0xcc); 
  			CamWriteCmosSensor(0x06,0x00); 
  			CamWriteCmosSensor(0x09,0x08); 
  			CamWriteCmosSensor(0x0a,0x2b); 
  			CamWriteCmosSensor(0xfd,0x01); 
  			CamWriteCmosSensor(0xef,0x22); 
  			CamWriteCmosSensor(0xf0,0x00); 
  			CamWriteCmosSensor(0x02,0x0f); 
  			CamWriteCmosSensor(0x03,0x01); 
  			CamWriteCmosSensor(0x06,0x1c); 
  			CamWriteCmosSensor(0x07,0x00); 
  			CamWriteCmosSensor(0x08,0x01); 
  			CamWriteCmosSensor(0x09,0x00); 
  			//Status                       
  			CamWriteCmosSensor(0xfd,0x02); 
  			CamWriteCmosSensor(0xbe,0xfe); 
  			CamWriteCmosSensor(0xbf,0x01); 
  			CamWriteCmosSensor(0xd0,0xfe); 
  			CamWriteCmosSensor(0xd1,0x01); 
  			CamWriteCmosSensor(0xfd,0x01); 
  			CamWriteCmosSensor(0x5b,0x01); 
  			CamWriteCmosSensor(0x5c,0xfe); 
  			CamWriteCmosSensor(0xfd,0x01);
                        CamWriteCmosSensor(0x32,0x15);

			}
		}
	}
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718SceneMode
*
* DESCRIPTION
*    This function get/set scene mode of sensor.
*
* PARAMETERS
*    IsCam: camera scene mode or not
*    In, Out
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718SceneMode(kal_bool IsCam, P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
  if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
  {
    const P_CAL_FEATURE_TYPE_ENUM_STRUCT SceneMode = &Out->FeatureInfo.FeatureEnum;
    
    Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
    Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
    Out->FeatureModeSupport = CAL_FEATURE_CAMERA_SUPPORT;
    SceneMode->IsSupport = KAL_TRUE;
    SceneMode->ItemCount = 2;
    if (IsCam)
    {
      SceneMode->SupportItem[0] = CAM_AUTO_DSC;
      SceneMode->SupportItem[1] = CAM_NIGHTSCENE;
    }
    else
    {
      SceneMode->SupportItem[0] = CAM_VIDEO_AUTO;
      SceneMode->SupportItem[1] = CAM_VIDEO_NIGHT;
    }
  }
 else if (CAL_FEATURE_SET_OPERATION == In->FeatureCtrlCode)
 
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718NightMode
*
* DESCRIPTION
*    This function set night mode to sensor.
*
* PARAMETERS
*    Enable
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718NightMode(kal_uint32 banding_para, kal_bool night_mode)
{
	if(banding_para == 1)
		SP0718_CAM_BANDING_50HZ = 0;
	else
		SP0718_CAM_BANDING_50HZ=1;
	if(night_mode == 1)
		SP0718_NIGHT_ENABLE = 1;
	else
		SP0718_NIGHT_ENABLE = 0 ;
}

/*************************************************************************
* FUNCTION
*  SP0718GammaSelect
*
* DESCRIPTION
*  This function config the gamma curve fitting for  
*  different LCD performances.
*
* PARAMETERS
*  level: different gamma curves.
*         Typical level value is from 1 to 5 which 
*         means small level gamma to high level gamma.
*
* RETURNS
*  NONE
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718GammaSelect(kal_uint8 level)
{
}

/*************************************************************************
* FUNCTION
*    SP0718InitialSetting
*
* DESCRIPTION
*    This function initialize the registers of CMOS sensor
*
* PARAMETERS
*    None
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718InitialSetting(void)
{
  CamWriteCmosSensor(0xfd,0x00);
  CamWriteCmosSensor(0x1C,0x00);
  CamWriteCmosSensor(0x31,0x10);
  CamWriteCmosSensor(0x27,0xb3);//0xb3	//2x gain
  CamWriteCmosSensor(0x1b,0x17);
  CamWriteCmosSensor(0x26,0xaa);
  CamWriteCmosSensor(0x37,0x02);
  CamWriteCmosSensor(0x28,0x8f);
  CamWriteCmosSensor(0x1a,0x73);
  CamWriteCmosSensor(0x1e,0x1b);
  CamWriteCmosSensor(0x21,0x06);  //blackout voltage
  CamWriteCmosSensor(0x22,0x2a);  //colbias
  CamWriteCmosSensor(0x0f,0x3f);
  CamWriteCmosSensor(0x10,0x3e);
  CamWriteCmosSensor(0x11,0x00);
  CamWriteCmosSensor(0x12,0x01);
  CamWriteCmosSensor(0x13,0x3f);
  CamWriteCmosSensor(0x14,0x04);
  CamWriteCmosSensor(0x15,0x30);
  CamWriteCmosSensor(0x16,0x31);
  CamWriteCmosSensor(0x17,0x01);
  CamWriteCmosSensor(0x69,0x31);
  CamWriteCmosSensor(0x6a,0x2a);
  CamWriteCmosSensor(0x6b,0x33);
  CamWriteCmosSensor(0x6c,0x1a);
  CamWriteCmosSensor(0x6d,0x32);
  CamWriteCmosSensor(0x6e,0x28);
  CamWriteCmosSensor(0x6f,0x29);
  CamWriteCmosSensor(0x70,0x34);
  CamWriteCmosSensor(0x71,0x18);
CamWriteCmosSensor(0x36,0x00);//02 delete badframe
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x5d,0x51);//position
CamWriteCmosSensor(0xf2,0x19);

//Blacklevel
CamWriteCmosSensor(0x1f,0x10);
CamWriteCmosSensor(0x20,0x1f);
//pregain 
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0x00,0x88);
CamWriteCmosSensor(0x01,0x88);
//SI15_SP0718 24M 50Hz 15-8fps 
//ae setting
CamWriteCmosSensor(0xfd,0x00);
CamWriteCmosSensor(0x03,0x01);
CamWriteCmosSensor(0x04,0xce);
CamWriteCmosSensor(0x06,0x00);
CamWriteCmosSensor(0x09,0x02);
CamWriteCmosSensor(0x0a,0xc4);
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xef,0x4d);
CamWriteCmosSensor(0xf0,0x00);
CamWriteCmosSensor(0x02,0x0c);
CamWriteCmosSensor(0x03,0x01);
CamWriteCmosSensor(0x06,0x47);
CamWriteCmosSensor(0x07,0x00);
CamWriteCmosSensor(0x08,0x01);
CamWriteCmosSensor(0x09,0x00);
//Status   
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0xbe,0x9c);
CamWriteCmosSensor(0xbf,0x03);
CamWriteCmosSensor(0xd0,0x9c);
CamWriteCmosSensor(0xd1,0x03);
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x5b,0x03);
CamWriteCmosSensor(0x5c,0x9c);

//rpc
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xe0,0x40);////24//4c//48//4c//44//4c//3e//3c//3a//38//rpc_1base_max
CamWriteCmosSensor(0xe1,0x30);////24//3c//38//3c//36//3c//30//2e//2c//2a//rpc_2base_max
CamWriteCmosSensor(0xe2,0x2e);////24//34//30//34//2e//34//2a//28//26//26//rpc_3base_max
CamWriteCmosSensor(0xe3,0x2a);////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_4base_max
CamWriteCmosSensor(0xe4,0x2a);////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_5base_max
CamWriteCmosSensor(0xe5,0x28);////24//2c//2a//2c//28//2c//24//22//20//rpc_6base_max
CamWriteCmosSensor(0xe6,0x28);////24//2c//2a//2c//28//2c//24//22//20//rpc_7base_max
CamWriteCmosSensor(0xe7,0x26);////24//2a//28//2a//26//2a//22//20//20//1e//rpc_8base_max
CamWriteCmosSensor(0xe8,0x26);////24//2a//28//2a//26//2a//22//20//20//1e//rpc_9base_max
CamWriteCmosSensor(0xe9,0x26);////24//2a//28//2a//26//2a//22//20//20//1e//rpc_10base_max
CamWriteCmosSensor(0xea,0x26);////24//28//26//28//24//28//20//1f//1e//1d//rpc_11base_max
CamWriteCmosSensor(0xf3,0x26);////24//28//26//28//24//28//20//1f//1e//1d//rpc_12base_max
CamWriteCmosSensor(0xf4,0x26);////24//28//26//28//24//28//20//1f//1e//1d//rpc_13base_max
//ae gain &status
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x04,0xe0);//rpc_max_indr
CamWriteCmosSensor(0x05,0x26);//1e//rpc_min_indr 
CamWriteCmosSensor(0x0a,0xa0);//rpc_max_outdr
CamWriteCmosSensor(0x0b,0x26);//rpc_min_outdr
CamWriteCmosSensor(0x5a,0x40);//dp rpc   
CamWriteCmosSensor(0xfd,0x02); 
CamWriteCmosSensor(0xbc,0xa0);//rpc_heq_low
CamWriteCmosSensor(0xbd,0x80);//rpc_heq_dummy
CamWriteCmosSensor(0xb8,0x80);//mean_normal_dummy
CamWriteCmosSensor(0xb9,0x90);//mean_dummy_normal

//ae target
CamWriteCmosSensor(0xfd,0x01); 
CamWriteCmosSensor(0xeb,SP0718_P1_0xeb);//78 
CamWriteCmosSensor(0xec,SP0718_P1_0xec);//78
CamWriteCmosSensor(0xed,0x0a);	
CamWriteCmosSensor(0xee,0x10);

//lsc       
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x26,0x30);
CamWriteCmosSensor(0x27,0x2c);
CamWriteCmosSensor(0x28,0x07);
CamWriteCmosSensor(0x29,0x08);
CamWriteCmosSensor(0x2a,0x40);
CamWriteCmosSensor(0x2b,0x03);
CamWriteCmosSensor(0x2c,0x00);
CamWriteCmosSensor(0x2d,0x00);
      
CamWriteCmosSensor(0xa1,0x24);
CamWriteCmosSensor(0xa2,0x27);
CamWriteCmosSensor(0xa3,0x27);
CamWriteCmosSensor(0xa4,0x2b);
CamWriteCmosSensor(0xa5,0x1c);
CamWriteCmosSensor(0xa6,0x1a);
CamWriteCmosSensor(0xa7,0x1a);
CamWriteCmosSensor(0xa8,0x1a);
CamWriteCmosSensor(0xa9,0x18);
CamWriteCmosSensor(0xaa,0x1c);
CamWriteCmosSensor(0xab,0x17);
CamWriteCmosSensor(0xac,0x17);
CamWriteCmosSensor(0xad,0x08);
CamWriteCmosSensor(0xae,0x08);
CamWriteCmosSensor(0xaf,0x08);
CamWriteCmosSensor(0xb0,0x00);
CamWriteCmosSensor(0xb1,0x00);
CamWriteCmosSensor(0xb2,0x00);
CamWriteCmosSensor(0xb3,0x00);
CamWriteCmosSensor(0xb4,0x00);
CamWriteCmosSensor(0xb5,0x02);
CamWriteCmosSensor(0xb6,0x06);
CamWriteCmosSensor(0xb7,0x00);
CamWriteCmosSensor(0xb8,0x00);
       
       
//DP       
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x48,0x09);
CamWriteCmosSensor(0x49,0x99);
        
//awb       
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x32,0x05);
CamWriteCmosSensor(0xfd,0x00);
CamWriteCmosSensor(0xe7,0x03);
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0x26,0xc8);
CamWriteCmosSensor(0x27,0xb6);
CamWriteCmosSensor(0xfd,0x00);
CamWriteCmosSensor(0xe7,0x00);
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0x1b,0x80);
CamWriteCmosSensor(0x1a,0x80);
CamWriteCmosSensor(0x18,0x26);
CamWriteCmosSensor(0x19,0x28);
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0x2a,0x00);
CamWriteCmosSensor(0x2b,0x08);
CamWriteCmosSensor(0x28,0xef);//0xa0//f8
CamWriteCmosSensor(0x29,0x20);

//d65 90  e2 93
CamWriteCmosSensor(0x66,0x42);//0x59//0x60////0x58//4e//0x48
CamWriteCmosSensor(0x67,0x62);//0x74//0x70//0x78//6b//0x69
CamWriteCmosSensor(0x68,0xee);//0xd6//0xe3//0xd5//cb//0xaa
CamWriteCmosSensor(0x69,0x18);//0xf4//0xf3//0xf8//ed
CamWriteCmosSensor(0x6a,0xa6);//0xa5
//indoor 91
CamWriteCmosSensor(0x7c,0x3b);//0x45//30//41//0x2f//0x44
CamWriteCmosSensor(0x7d,0x5b);//0x70//60//55//0x4b//0x6f
CamWriteCmosSensor(0x7e,0x15);//0a//0xed
CamWriteCmosSensor(0x7f,0x39);//23//0x28
CamWriteCmosSensor(0x80,0xaa);//0xa6
//cwf   92 
CamWriteCmosSensor(0x70,0x3e);//0x38//41//0x3b
CamWriteCmosSensor(0x71,0x59);//0x5b//5f//0x55
CamWriteCmosSensor(0x72,0x31);//0x30//22//0x28
CamWriteCmosSensor(0x73,0x4f);//0x54//44//0x45
CamWriteCmosSensor(0x74,0xaa);
//tl84  93 
CamWriteCmosSensor(0x6b,0x1b);//0x18//11
CamWriteCmosSensor(0x6c,0x3a);//0x3c//25//0x2f
CamWriteCmosSensor(0x6d,0x3e);//0x3a//35
CamWriteCmosSensor(0x6e,0x59);//0x5c//46//0x52
CamWriteCmosSensor(0x6f,0xaa);
//f    94
CamWriteCmosSensor(0x61,0xea);//0x03//0x00//f4//0xed
CamWriteCmosSensor(0x62,0x03);//0x1a//0x25//0f//0f
CamWriteCmosSensor(0x63,0x6a);//0x62//0x60//52//0x5d
CamWriteCmosSensor(0x64,0x8a);//0x7d//0x85//70//0x75//0x8f
CamWriteCmosSensor(0x65,0x6a);//0xaa//6a
      
CamWriteCmosSensor(0x75,0x80);
CamWriteCmosSensor(0x76,0x20);
CamWriteCmosSensor(0x77,0x00);
CamWriteCmosSensor(0x24,0x25);

//ڵƫʹ//ƫ
CamWriteCmosSensor(0x20,0xd8);
CamWriteCmosSensor(0x21,0xa3);//82//a8ƫնȻеƫ
CamWriteCmosSensor(0x22,0xd0);//e3//bc
CamWriteCmosSensor(0x23,0x86);

//outdoor r\b range
CamWriteCmosSensor(0x78,0xc3);//d8
CamWriteCmosSensor(0x79,0xba);//82
CamWriteCmosSensor(0x7a,0xa6);//e3
CamWriteCmosSensor(0x7b,0x99);//86


//skin 
CamWriteCmosSensor(0x08,0x15);//
CamWriteCmosSensor(0x09,0x04);//
CamWriteCmosSensor(0x0a,0x20);//
CamWriteCmosSensor(0x0b,0x12);//
CamWriteCmosSensor(0x0c,0x27);//
CamWriteCmosSensor(0x0d,0x06);//
CamWriteCmosSensor(0x0e,0x63);//

//wt th
CamWriteCmosSensor(0x3b,0x10);
//gw
CamWriteCmosSensor(0x31,0x60);
CamWriteCmosSensor(0x32,0x60);
CamWriteCmosSensor(0x33,0xc0);
CamWriteCmosSensor(0x35,0x6f);

// sharp
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0xde,0x0f);
CamWriteCmosSensor(0xd2,0x02);//6//ƺڰױߣ0-ߴ֣f-ϸ
CamWriteCmosSensor(0xd3,0x06);
CamWriteCmosSensor(0xd4,0x06);
CamWriteCmosSensor(0xd5,0x06);
CamWriteCmosSensor(0xd7,0x20);//10//2xжֵ
CamWriteCmosSensor(0xd8,0x30);//24//1A//4x
CamWriteCmosSensor(0xd9,0x38);//28//8x
CamWriteCmosSensor(0xda,0x38);//16x
CamWriteCmosSensor(0xdb,0x08);//
CamWriteCmosSensor(0xe8,0x58);//48//ǿ
CamWriteCmosSensor(0xe9,0x48);
CamWriteCmosSensor(0xea,0x30);
CamWriteCmosSensor(0xeb,0x20);
CamWriteCmosSensor(0xec,0x48);//60//80
CamWriteCmosSensor(0xed,0x48);//50//60
CamWriteCmosSensor(0xee,0x30);
CamWriteCmosSensor(0xef,0x20);
//ƽ̹
CamWriteCmosSensor(0xf3,0x50);
CamWriteCmosSensor(0xf4,0x10);
CamWriteCmosSensor(0xf5,0x10);
CamWriteCmosSensor(0xf6,0x10);
//dns       
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x64,0x44);//طԵƽ  //0-ǿ8-
CamWriteCmosSensor(0x65,0x22);
CamWriteCmosSensor(0x6d,0x04);//8//ǿƽƽ̹ƽֵ
CamWriteCmosSensor(0x6e,0x06);//8
CamWriteCmosSensor(0x6f,0x10);
CamWriteCmosSensor(0x70,0x10);
CamWriteCmosSensor(0x71,0x08);//0d//ƽƽ̹ƽֵ	
CamWriteCmosSensor(0x72,0x12);//1b
CamWriteCmosSensor(0x73,0x1c);//20
CamWriteCmosSensor(0x74,0x24);
CamWriteCmosSensor(0x75,0x44);//[7:4]ƽ̹ǿȣ[3:0]ƽ̹ǿȣ0-ǿ8-
CamWriteCmosSensor(0x76,0x02);//46
CamWriteCmosSensor(0x77,0x02);//33
CamWriteCmosSensor(0x78,0x02);
CamWriteCmosSensor(0x81,0x10);//18//2x//жֵǿƽƽ
CamWriteCmosSensor(0x82,0x20);//30//4x
CamWriteCmosSensor(0x83,0x30);//40//8x
CamWriteCmosSensor(0x84,0x48);//50//16x
CamWriteCmosSensor(0x85,0x0c);//12/8+reg0x81 ڶֵƽ̹ͷƽ̹
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0xdc,0x0f);
       
//gamma    
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x8b,0x00);//00//00     
CamWriteCmosSensor(0x8c,0x0a);//0c//09     
CamWriteCmosSensor(0x8d,0x16);//19//17     
CamWriteCmosSensor(0x8e,0x1f);//25//24     
CamWriteCmosSensor(0x8f,0x2a);//30//33     
CamWriteCmosSensor(0x90,0x3c);//44//47     
CamWriteCmosSensor(0x91,0x4e);//54//58     
CamWriteCmosSensor(0x92,0x5f);//61//64     
CamWriteCmosSensor(0x93,0x6c);//6d//70     
CamWriteCmosSensor(0x94,0x82);//80//81     
CamWriteCmosSensor(0x95,0x94);//92//8f     
CamWriteCmosSensor(0x96,0xa6);//a1//9b     
CamWriteCmosSensor(0x97,0xb2);//ad//a5     
CamWriteCmosSensor(0x98,0xbf);//ba//b0     
CamWriteCmosSensor(0x99,0xc9);//c4//ba     
CamWriteCmosSensor(0x9a,0xd1);//cf//c4     
CamWriteCmosSensor(0x9b,0xd8);//d7//ce     
CamWriteCmosSensor(0x9c,0xe0);//e0//d7     
CamWriteCmosSensor(0x9d,0xe8);//e8//e1     
CamWriteCmosSensor(0x9e,0xef);//ef//ea     
CamWriteCmosSensor(0x9f,0xf8);//f7//f5     
CamWriteCmosSensor(0xa0,0xff);//ff//ff     
//CCM      
CamWriteCmosSensor(0xfd,0x02);
CamWriteCmosSensor(0x15,0xd0);//b>th
CamWriteCmosSensor(0x16,0x95);//r<th
//gcͷƫ
//!F        
CamWriteCmosSensor(0xa0,0x80);//80
CamWriteCmosSensor(0xa1,0x00);//00
CamWriteCmosSensor(0xa2,0x00);//00
CamWriteCmosSensor(0xa3,0x00);//06
CamWriteCmosSensor(0xa4,0x8c);//8c
CamWriteCmosSensor(0xa5,0xf4);//ed
CamWriteCmosSensor(0xa6,0x0c);//0c
CamWriteCmosSensor(0xa7,0xf4);//f4
CamWriteCmosSensor(0xa8,0x80);//80
CamWriteCmosSensor(0xa9,0x00);//00
CamWriteCmosSensor(0xaa,0x30);//30
CamWriteCmosSensor(0xab,0x0c);//0c 
//F        
CamWriteCmosSensor(0xac,0x8c);
CamWriteCmosSensor(0xad,0xf4);
CamWriteCmosSensor(0xae,0x00);
CamWriteCmosSensor(0xaf,0xed);
CamWriteCmosSensor(0xb0,0x8c);
CamWriteCmosSensor(0xb1,0x06);
CamWriteCmosSensor(0xb2,0xf4);
CamWriteCmosSensor(0xb3,0xf4);
CamWriteCmosSensor(0xb4,0x99);
CamWriteCmosSensor(0xb5,0x0c);
CamWriteCmosSensor(0xb6,0x03);
CamWriteCmosSensor(0xb7,0x0f);
        
//sat u     
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xd3,0x9c);//0x88//50
CamWriteCmosSensor(0xd4,0x98);//0x88//50
CamWriteCmosSensor(0xd5,0x8c);//50
CamWriteCmosSensor(0xd6,0x84);//50
//sat v   
CamWriteCmosSensor(0xd7,0x9c);//0x88//50
CamWriteCmosSensor(0xd8,0x98);//0x88//50
CamWriteCmosSensor(0xd9,0x8c);//50
CamWriteCmosSensor(0xda,0x84);//50
//auto_sat  
CamWriteCmosSensor(0xdd,0x30);
CamWriteCmosSensor(0xde,0x10);
CamWriteCmosSensor(0xd2,0x01);//autosa_en
CamWriteCmosSensor(0xdf,0xff);//a0//y_mean_th
        
//uv_th     
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xc2,0xaa);
CamWriteCmosSensor(0xc3,0xaa);
CamWriteCmosSensor(0xc4,0x66);
CamWriteCmosSensor(0xc5,0x66); 

//heq
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0x0f,0xff);
CamWriteCmosSensor(0x10,SP0718_P1_0x10); //out
CamWriteCmosSensor(0x14,SP0718_P1_0x14); 
CamWriteCmosSensor(0x11,SP0718_P1_0x11); //nr
CamWriteCmosSensor(0x15,SP0718_P1_0x15);  
CamWriteCmosSensor(0x12,SP0718_P1_0x12); //dummy
CamWriteCmosSensor(0x16,SP0718_P1_0x16); 
CamWriteCmosSensor(0x13,SP0718_P1_0x13); //low 	
CamWriteCmosSensor(0x17,SP0718_P1_0x17);   	

CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xcd,0x20);
CamWriteCmosSensor(0xce,0x1f);
CamWriteCmosSensor(0xcf,0x20);
CamWriteCmosSensor(0xd0,0x55);  
//auto 
CamWriteCmosSensor(0xfd,0x01);
CamWriteCmosSensor(0xfb,0x33);
//CamWriteCmosSensor(0x32,0x15);
CamWriteCmosSensor(0x33,0xff);
CamWriteCmosSensor(0x34,0xe7);
  
}

/*************************************************************************
* FUNCTION
*	SP0718_Write_More_Registers
*
* DESCRIPTION
*	This function is served for FAE to modify the necessary Init Regs. Do not modify the regs
*     in init_SP0718() directly.
*
* PARAMETERS
*	None
*
* RETURNS
*	None
*
* GLOBALS AFFECTED
*
*************************************************************************/
void SP0718MoreSetting(void)
{
	//  TODO: FAE Modify the Init Regs here!!! 
}
/*************************************************************************
* FUNCTION
*    SP0718GetSensorInfo
*
* DESCRIPTION
*    This function set sensor infomation
*
* PARAMETERS
*    Info
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718GetSensorInfo(P_IMAGE_SENSOR_GET_SENSOR_INFO_STRUCT Info)
{
  Info->SensorId = SP0718_SENSOR_ID;
  Info->SensorIf = IMAGE_SENSOR_IF_PARALLEL;
  Info->PreviewNormalDataFormat = IMAGE_SENSOR_DATA_OUT_FORMAT_UYVY;
  Info->PreviewHMirrorDataFormat = IMAGE_SENSOR_DATA_OUT_FORMAT_UYVY;
  Info->PreviewVMirrorDataFormat = IMAGE_SENSOR_DATA_OUT_FORMAT_UYVY;
  Info->PreviewHVMirrorDataFormat = IMAGE_SENSOR_DATA_OUT_FORMAT_UYVY;
  Info->CaptureDataFormat = IMAGE_SENSOR_DATA_OUT_FORMAT_UYVY;
  Info->PixelClkInv = KAL_FALSE;
  Info->PixelClkPolarity = POLARITY_LOW;
  Info->HsyncPolarity = POLARITY_LOW;
  Info->VsyncPolarity = POLARITY_HIGH;
  Info->PreviewMclkFreq = SP0718_MCLK;
  Info->CaptureMclkFreq = SP0718_MCLK;
  Info->VideoMclkFreq = SP0718_MCLK;
  Info->PreviewWidth = SP0718_IMAGE_SENSOR_WIDTH_DRV;
  Info->PreviewHeight = SP0718_IMAGE_SENSOR_HEIGHT_DRV;
  Info->FullWidth = SP0718_IMAGE_SENSOR_WIDTH_DRV;
  Info->FullHeight = SP0718_IMAGE_SENSOR_HEIGHT_DRV;
  Info->SensorAfSupport = KAL_FALSE;
  Info->SensorFlashSupport = KAL_FALSE;
}

/*************************************************************************
* FUNCTION
*    SP0718PowerOn
*
* DESCRIPTION
*    This function reset sensor and read id
*
* PARAMETERS
*    None
*
* RETURNS
*    sensor id
*
* LOCAL AFFECTED
*
*************************************************************************/
static kal_uint16 SP0718PowerOn(void)
{
  kal_uint16 SensorId;
  
  CisModulePowerOn(SP0718Sensor.SensorIdx, KAL_TRUE);
  kal_sleep_task(2);  /* delay for power stable */
  CameraSccbOpen(CAMERA_SCCB_SENSOR, SP0718_WRITE_ID, CAMERA_SCCB_8BIT, CAMERA_SCCB_8BIT, 100);
  CamPdnPinCtrl(SP0718Sensor.SensorIdx, 0);
  
  CamRstPinCtrl(SP0718Sensor.SensorIdx, 1);
  kal_sleep_task(1);
  CamRstPinCtrl(SP0718Sensor.SensorIdx, 0);
  kal_sleep_task(1);
  CamRstPinCtrl(SP0718Sensor.SensorIdx, 1);
  kal_sleep_task(3);
  
  SensorId = CamReadCmosSensor(0x02);
  if (SensorId != SP0718_SENSOR_ID) /* READ SENSOR ID */
  {
    //dbg_print("SP0718PowerOn!the SensorId = %x , SP0718Sensor.SensorIdx = %d\r\n",SensorId,SP0718Sensor.SensorIdx);
#if defined(__SP0718_DEBUG_TRACE__)
    kal_print_string_trace(MOD_ENG, TRACE_ERROR, "SENSOR ID: %x", SensorId);
#endif
  }
  return SensorId;
}

/*************************************************************************
* FUNCTION
*    SP0718DetectSensorId
*
* DESCRIPTION
*    This function detect sensor id
*
* PARAMETERS
*    None
*
* RETURNS
*    sensor id
*
* LOCAL AFFECTED
*
*************************************************************************/
static kal_uint32 SP0718DetectSensorId(void)
{
  MM_ERROR_CODE_ENUM SP0718SensorClose(void);
  kal_uint8 SensorId;
  IMAGE_SENSOR_INDEX_ENUM InvIdx;
  kal_int8 i, j;
  
  switch (SP0718Sensor.SensorIdx)
  {
  case IMAGE_SENSOR_MAIN:
  case IMAGE_SENSOR_BAK1:
    InvIdx = IMAGE_SENSOR_SUB;
    break;
  default:
    InvIdx = IMAGE_SENSOR_MAIN;
    break;
  }
  for (i = 1; i >= 0; i--)
  {
    for (j = 1; j >= 0; j--)
    {
      CamRstPinCtrl(InvIdx, i);
      CamPdnPinCtrl(InvIdx, j);
      SensorId = SP0718PowerOn();
      SP0718SensorClose();
      //dbg_print("SP0718DetectSensorId!SP0718_SENSOR_ID = %x,i=%d,j=%d\r\n",SP0718_SENSOR_ID,i,j);
      if (SP0718_SENSOR_ID == SensorId)
      {
        return SP0718_SENSOR_ID;
      }
    }
  }
  return NULL_SENSOR_ID;
}

/*************************************************************************
* FUNCTION
*    SP0718InitOperationPara
*
* DESCRIPTION
*    This function set sensor operation para
*
* PARAMETERS
*    Para
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718InitOperationPara(P_IMAGE_SENSOR_GET_INIT_OPERATION_PARA_STRUCT Para)
{
  Para->CaptureDelayFrame = 2;
  Para->PreviewDelayFrame = 4;
  Para->PreviewDisplayWaitFrame = 8;
}

/*************************************************************************
* FUNCTION
*    SP0718Preview
*
* DESCRIPTION
*    This function set to preview mode
*
* PARAMETERS
*    Id, In, Out
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718Preview(CAL_SCENARIO_ENUM Id, P_IMAGE_SENSOR_SCENARIO_PARA_IN_STRUCT In, P_IMAGE_SENSOR_SCENARIO_PARA_OUT_STRUCT Out)
{
  kal_uint16 DummyPixel;
  
  SP0718SetMirror(In->ImageMirror);
  //SP0718SetClock(SP0718_PREVIEW_CLK);
    switch (Id)
  {
    case CAL_SCENARIO_VIDEO:
        //GC0309SetVideoFps(In->MaxVideoFrameRate);
        SP0718_video_mode = KAL_TRUE;
	 SP0718NightMode(In->BandingFreq, In->NightMode);
	 break;
    default:
        SP0718_video_mode = KAL_FALSE;
	 SP0718NightMode(In->BandingFreq, In->NightMode);
        /* roll back shutter&gain from capture state */
	 break;
    }
  //SP0718SetBaseShutter(CAM_BANDING_50HZ == In->BandingFreq ? SP0718_NUM_50HZ : SP0718_NUM_60HZ);
  //SP0718NightMode(In->NightMode);
  SP0718AeEnable(KAL_TRUE);
  SP0718AwbEnable(KAL_TRUE);
  
  Out->WaitStableFrameNum = 0;
  Out->GrabStartX = SP0718_X_START;
  Out->GrabStartY = SP0718_Y_START;
  Out->ExposureWindowWidth = SP0718_IMAGE_SENSOR_WIDTH;
  Out->ExposureWindowHeight = SP0718_IMAGE_SENSOR_HEIGHT;
}

/*************************************************************************
* FUNCTION
*    SP0718Capture
*
* DESCRIPTION
*    This function set to capture mode
*
* PARAMETERS
*    Id, In, Out
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718Capture(CAL_SCENARIO_ENUM Id, P_IMAGE_SENSOR_SCENARIO_PARA_IN_STRUCT In, P_IMAGE_SENSOR_SCENARIO_PARA_OUT_STRUCT Out)
{
  Out->WaitStableFrameNum = 0;
  Out->GrabStartX = SP0718_X_START;
  Out->GrabStartY = SP0718_Y_START;
  Out->ExposureWindowWidth = SP0718_IMAGE_SENSOR_WIDTH;
  Out->ExposureWindowHeight = SP0718_IMAGE_SENSOR_HEIGHT;
}

/*************************************************************************
* FUNCTION
*    SP0718SensorOpen
*
* DESCRIPTION
*    This function read sensor id and init sensor
*
* PARAMETERS
*    None
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718SensorOpen(void)
{
#if 0
  if (SP0718PowerOn() != SP0718_SENSOR_ID)
  {
    return MM_ERROR_SENSOR_READ_ID_FAIL;
  }
#else
    MM_ERROR_CODE_ENUM SP0718SensorClose(void);
    kal_uint8 SensorId;
    IMAGE_SENSOR_INDEX_ENUM InvIdx;
    kal_int8 i, j;

    switch (SP0718Sensor.SensorIdx)
    {
        case IMAGE_SENSOR_MAIN:
        case IMAGE_SENSOR_BAK1:
            InvIdx = IMAGE_SENSOR_SUB;
            break;
        default:
            InvIdx = IMAGE_SENSOR_MAIN;
            break;
    }
    CamRstPinCtrl(InvIdx, 1);
    CamPdnPinCtrl(InvIdx,1);
    SensorId = SP0718PowerOn();
    if (SP0718_SENSOR_ID != SensorId)
    {
        //dbg_print("SP0718SensorOpen!the line = %d\r\n",__LINE__);
        kal_prompt_trace(MOD_MED,"Telacom SP0718SensorOpen!the line = %d\r\n",__LINE__);
        SP0718SensorClose();
        return MM_ERROR_SENSOR_READ_ID_FAIL;
    }
#endif

  SP0718InitialSetting();
 // SP0718MoreSetting();
  /* default setting */
  SP0718Sensor.BypassAe = KAL_FALSE;
  SP0718Sensor.BypassAwb = KAL_FALSE;
  SP0718Sensor.Ctrl3A = CamReadCmosSensor(0x32);
  SP0718Sensor.AecMode3 = CamReadCmosSensor(0x32);
  SP0718Sensor.VideoFps = 0;
  SP0718Sensor.IntClk = 0;
  SP0718Sensor.FrameHeight = 0;
  SP0718Sensor.LineLength = 0;
  
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718SensorFeatureCtrl
*
* DESCRIPTION
*    This function set sensor feature mode
*
* PARAMETERS
*    Id: scenario id
*    In: input parameter
*    Out: output parameter
*    OutLen: output parameter length
*    RealOutLen: real output parameter length
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718FlashMode(P_CAL_FEATURE_CTRL_STRUCT In, P_CAL_FEATURE_CTRL_STRUCT Out)
{
    if (CAL_FEATURE_QUERY_OPERATION == In->FeatureCtrlCode)
    {
        const P_CAL_FEATURE_TYPE_ENUM_STRUCT FlashMode = &Out->FeatureInfo.FeatureEnum;
 
        Out->FeatureType = CAL_FEATURE_TYPE_ENUMERATE;
        Out->FeatureOperationSupport = CAL_FEATURE_SET_OPERATION|CAL_FEATURE_QUERY_OPERATION;
        Out->FeatureModeSupport = CAL_FEATURE_CAMERA_SUPPORT;
        // TODO: add supported flash mode enum here
#ifndef LED_FLASHLIGHT_SUPPORT
        FlashMode->IsSupport = KAL_FALSE;
#else
        FlashMode->IsSupport = KAL_TRUE;
        FlashMode->ItemCount = 2;
        FlashMode->SupportItem[0] = CAM_FLASH_OFF;
        FlashMode->SupportItem[1] = CAM_FLASH_ON;
#endif
    }
    return MM_ERROR_NONE;
}
static MM_ERROR_CODE_ENUM SP0718SensorFeatureCtrl(kal_uint32 Id, void *In, void *Out, kal_uint32 OutLen, kal_uint32 *RealOutLen)
{
  MM_ERROR_CODE_ENUM ErrCode = MM_ERROR_NONE;
  
  if (Id >= CAL_FEATURE_WEBCAM_BEGIN && Id <= CAL_FEATURE_WEBCAM_END)
  {
    return SP0718WebcamFeatureCtrl(Id, In, Out, OutLen, RealOutLen);
  }
  switch (Id)
  {
  /* query and set series */
  case CAL_FEATURE_CAMERA_BRIGHTNESS:
  case CAL_FEATURE_CAMERA_EV_VALUE:
    ErrCode = SP0718Ev(In, Out);
    break;
  case CAL_FEATURE_CAMERA_CONTRAST:
  case CAL_FEATURE_CAMERA_HUE:
  case CAL_FEATURE_CAMERA_SATURATION:
  case CAL_FEATURE_CAMERA_SHARPNESS:
  case CAL_FEATURE_CAMERA_GAMMA:
    ErrCode = MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
    break;
  case CAL_FEATURE_CAMERA_WB:
    ErrCode = SP0718Wb(In, Out);
    break;
  case CAL_FEATURE_CAMERA_IMAGE_EFFECT:
    ErrCode = SP0718Effect(In, Out);
    break;
  case CAL_FEATURE_CAMERA_STILL_CAPTURE_SIZE:
    ErrCode = SP0718StillCaptureSize(In, Out);
    break;
  case CAL_FEATURE_CAMERA_BANDING_FREQ:
    ErrCode = SP0718Banding(In, Out);
    break;
  case CAL_FEATURE_CAMERA_SCENE_MODE:
    ErrCode = SP0718SceneMode(KAL_TRUE, In, Out);
    break;
  case CAL_FEATURE_CAMERA_VIDEO_SCENE_MODE:
    ErrCode = SP0718SceneMode(KAL_FALSE, In, Out);
    break;
    case CAL_FEATURE_CAMERA_FLASH_MODE:
      ErrCode = SP0718FlashMode(In, Out);
      break;
    case IMAGE_SENSOR_FEATURE_QUERY_YUV_CTRL_FLASHLIGHT_TRIGGER_MODE:
        ((P_IMAGE_SENSOR_COMM_DATA_STRUCT)Out)->FeatureValue = FLASHLIGHT_CONTINUOUS_MODE;
        break;
    case IMAGE_SENSOR_FEATURE_CTRL_FLASHLIGHT:
        FlashlightPowerOn(SP0718Sensor.SensorIdx, ((P_IMAGE_SENSOR_COMM_CTRL_STRUCT)In)->FeatureEnable);
        break;
  
  /* get info series */
  case IMAGE_SENSOR_FEATURE_GET_SENSOR_INFO:
    SP0718GetSensorInfo(Out);
    break;
  case IMAGE_SENSOR_FEATURE_GET_SENSOR_ID:
    ((P_IMAGE_SENSOR_COMM_DATA_STRUCT)Out)->FeatureValue = SP0718DetectSensorId();
    break;
  case IMAGE_SENSOR_FEATURE_GET_SENSOR_FRAME_RATE:
    ((P_IMAGE_SENSOR_COMM_DATA_STRUCT)Out)->FeatureValue = SP0718Sensor.VideoFps;
    break;
  case IMAGE_SENSOR_FEATURE_GET_INIT_OPERATION_PARA:
    SP0718InitOperationPara(Out);
    break;
  case IMAGE_SENSOR_FEATURE_GET_HW_DEPENDENT_SETTING:
    CamGetHWInfo(SP0718Sensor.SensorIdx, Out);
    break;
  
  /* set para series */
  case IMAGE_SENSOR_FEATURE_SET_SENSOR_POWER_DOWN:
    CamPdnPinCtrl(SP0718Sensor.SensorIdx, 1);
    CamRstPinCtrl(SP0718Sensor.SensorIdx, 0);
    break;
  case IMAGE_SENSOR_FEATURE_SET_NVRAM_SENSOR_INFO:
    SP0718Sensor.NvramData = &(((nvram_camera_para_struct *)In)->SENSOR);
    break;
  case IMAGE_SENSOR_FEATURE_SET_CAMERA_SOURCE:
    SP0718Sensor.SensorIdx = ((P_IMAGE_SENSOR_SET_CAMERA_SOURCE_STRUCT)In)->ImageSensorIdx;
    break;
  case IMAGE_SENSOR_FEATURE_CTRL_NIGHTMODE: /* move to preview */
    break;
  case IMAGE_SENSOR_FEATURE_SET_AE_BYPASS:
    SP0718Sensor.BypassAe = *(kal_bool *)In;
    SP0718AeEnable(SP0718Sensor.BypassAe ? KAL_FALSE : KAL_TRUE);
    break;
  case IMAGE_SENSOR_FEATURE_SET_AWB_BYPASS:
    SP0718Sensor.BypassAwb = *(kal_bool *)In;
    SP0718AwbEnable(SP0718Sensor.BypassAwb ? KAL_FALSE : KAL_TRUE);
    break;
  default:
    ErrCode = MM_ERROR_SENSOR_FEATURE_NOT_SUPPORT;
    break;
  }
  return ErrCode;
}

/*************************************************************************
* FUNCTION
*    SP0718SensorCtrl
*
* DESCRIPTION
*    This function set sensor scenario mode
*
* PARAMETERS
*    Id: scenario id
*    In/Out: parameter
*    OutLen: Out parameter length
*    RealOutLen: Real out parameter length
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718SensorCtrl(CAL_SCENARIO_ENUM Id, void *In, void *Out, kal_uint32 OutLen, kal_uint32 *RealOutLen)
{
  switch (Id)
  {
  case CAL_SCENARIO_CAMERA_PREVIEW:
  case CAL_SCENARIO_VIDEO:
  case CAL_SCENARIO_WEBCAM_PREVIEW:
  case CAL_SCENARIO_WEBCAM_CAPTURE:
    SP0718Preview(Id, In, Out);
    break;
  case CAL_SCENARIO_CAMERA_STILL_CAPTURE:
    SP0718Capture(Id, In, Out);
    break;
  default:
    break;
  }
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718SensorClose
*
* DESCRIPTION
*    This function is to turn off sensor module power.
*
* PARAMETERS
*    None
*
* RETURNS
*    Error code
*
* LOCAL AFFECTED
*
*************************************************************************/
static MM_ERROR_CODE_ENUM SP0718SensorClose(void)
{
    kal_prompt_trace(MOD_MED,"Telacom SP0718SensorClose!");
  CamPdnPinCtrl(SP0718Sensor.SensorIdx, 1);
  CamRstPinCtrl(SP0718Sensor.SensorIdx, 0);
  CisModulePowerOn(SP0718Sensor.SensorIdx, KAL_FALSE);
  CameraSccbClose(CAMERA_SCCB_SENSOR);
  
  return MM_ERROR_NONE;
}

/*************************************************************************
* FUNCTION
*    SP0718SensorFunc
*
* DESCRIPTION
*    This function get sensor function pointer
*
* PARAMETERS
*    sensor function pointer
*
* RETURNS
*    Error code
*
* GLOBALS AFFECTED
*
*************************************************************************/
MM_ERROR_CODE_ENUM SP0718SensorFunc(P_IMAGE_SENSOR_FUNCTION_STRUCT *pfSensorFunc)
{
  static IMAGE_SENSOR_FUNCTION_STRUCT ImageSensorFuncSP0718 =
  {
    SP0718SensorOpen,
    SP0718SensorFeatureCtrl,
    SP0718SensorCtrl,
    SP0718SensorClose,
  };
  *pfSensorFunc = &ImageSensorFuncSP0718;
  
  return MM_ERROR_NONE;
}

