/*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*

               CAMERA    API    M O D U L E

                        EDIT HISTORY FOR MODULE

when        who       what, where, why
--------    ---       -------------------------------------------------------
2013.3.16  semisky   Created file.
*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*/

/*===========================================================================

                           INCLUDE FILES FOR MODULE

===========================================================================*/
#include <linux/delay.h>
#include <asm/system.h>
#include <mach/hardware.h>
#include <asm/io.h>
#include "sensor_if.h"
#include "cam.h"
#include "cam_reg.h"
#include "tcc_cam_i2c.h"

#ifdef CONFIG_VIDEO_CAMERA_SENSOR_SP0718


#if defined(CONFIG_VIDEO_DUAL_CAMERA_SUPPORT)
extern int CameraID;
#endif
#define SP0718_NORMAL_Y0ffset  0x20
#define SP0718_LOWLIGHT_Y0ffset  0x25
//AE target
#define  SP0718_P1_0xeb  0x78
#define  SP0718_P1_0xec  0x6c
//HEQ
#define  SP0718_P1_0x10  0x00//outdoor
#define  SP0718_P1_0x14  0x20
#define  SP0718_P1_0x11  0x00//nr
#define  SP0718_P1_0x15  0x18
#define  SP0718_P1_0x12  0x00//dummy
#define  SP0718_P1_0x16  0x10
#define  SP0718_P1_0x13  0x00//low
#define  SP0718_P1_0x17  0x00

/* Array of image sizes supported by SP0718.  These must be ordered from 
 * smallest image size to largest.
 */
#if defined(CONFIG_VIDEO_DUAL_CAMERA_SUPPORT)
struct capture_size sensor_sizes_SP0718[] = {
	{  640,  480 },	/* VGA */
	{  352,  288 },	/* CIF */
	{  320,  240 },	/* QVGA */
	{  176,  144 },	/* QCIF */
};
#else
struct capture_size sensor_sizes[] = {
	{  640,  480 },	/* VGA */
	{  352,  288 },	/* CIF */
	{  320,  240 },	/* QVGA */
	{  176,  144 },	/* QCIF */
};
#endif
typedef enum {
   CAM_WB_AUTO=0,
   CAM_WB_CLOUD,
   CAM_WB_DAYLIGHT,
   CAM_WB_INCANDESCENCE,
   CAM_WB_TUNGSTEN,
   CAM_WB_FLUORESCENT,
   CAM_WB_MANUAL
} WB_GROUP_ENUM;

/* register initialization tables for sensor */
/* common sensor register initialization for all image sizes, pixel formats, 
 * and frame rates
 */

static struct sensor_reg sensor_initialize_SP0718[] = 
{		//saturday param test
		{0xfd , 0x00},
		{0x1C , 0x3c},
		{0x31 , 0x10},
		{0x30 , 0x01},
		{0x27 , 0xb3},///0xb3	///2x gain
		{0x1b , 0x17},
		{0x26 , 0xaa},
		{0x37 , 0x02},
		{0x28 , 0x8f},
		{0x1a , 0x73},
		{0x1e , 0x1b},
		{0x21 , 0x06},///blackout voltage
		{0x22 , 0x2a},///colbias
		{0x0f , 0x3f},
		{0x10 , 0x3e},
		{0x11 , 0x00},
		{0x12 , 0x01},
		{0x13 , 0x3f},
		{0x14 , 0x04},
		{0x15 , 0x30},
		{0x16 , 0x31},
		{0x17 , 0x01},
		{0x69 , 0x31},
		{0x6a , 0x2a},
		{0x6b , 0x33},
		{0x6c , 0x1a},
		{0x6d , 0x32},
		{0x6e , 0x28},
		{0x6f , 0x29},
		{0x70 , 0x34},
		{0x71 , 0x18},
		{0x36,0x00},//02 delete badframe
		{0xfd,0x01},
		{0x5d,0x51},//position
		{0xf2,0x19},

		//Blacklevel
		{0x1f,0x10},
		{0x20,0x1f},
		//pregain 
		{0xfd,0x02},
		{0x00,0x88},
		{0x01,0x88},
		//SI15_SP0718 12M 50Hz 12-8fps 
		//ae setting
		{0xfd,0x00},
		{0x03,0x01},
		{0x04,0x74},
		{0x06,0x00},
		{0x09,0x00},
		{0x0a,0x76},
		{0xfd,0x01},
		{0xef,0x3e},
		{0xf0,0x00},
		{0x02,0x0c},
		{0x03,0x01},
		{0x06,0x38},
		{0x07,0x00},
		{0x08,0x01},
		{0x09,0x00},
		//Status   
		{0xfd,0x02},
		{0xbe,0xe8},
		{0xbf,0x02},
		{0xd0,0xe8},
		{0xd1,0x02},
		{0xfd,0x01},
		{0x5b,0x02},
		{0x5c,0xe8},

		///rpc
		{0xfd , 0x01},
		{0xe0 , 0x40},//////24///4c///48///4c///44///4c///3e///3c///3a///38///rpc_1base_max
		{0xe1 , 0x30},//////24///3c///38///3c///36///3c///30///2e///2c///2a///rpc_2base_max
		{0xe2 , 0x2e},//////24///34///30///34///2e///34///2a///28///26///26///rpc_3base_max
		{0xe3 , 0x2a},//////24///2a///2e///2c///2e///2a///2e///26///24///22///rpc_4base_max
		{0xe4 , 0x2a},//////24///2a///2e///2c///2e///2a///2e///26///24///22///rpc_5base_max
		{0xe5 , 0x28},//////24///2c///2a///2c///28///2c///24///22///20///rpc_6base_max
		{0xe6 , 0x28},//////24///2c///2a///2c///28///2c///24///22///20///rpc_7base_max
		{0xe7 , 0x26},//////24///2a///28///2a///26///2a///22///20///20///1e///rpc_8base_max
		{0xe8 , 0x26},//////24///2a///28///2a///26///2a///22///20///20///1e///rpc_9base_max
		{0xe9 , 0x26},//////24///2a///28///2a///26///2a///22///20///20///1e///rpc_10base_max
		{0xea , 0x26},//////24///28///26///28///24///28///20///1f///1e///1d///rpc_11base_max
		{0xf3 , 0x26},//////24///28///26///28///24///28///20///1f///1e///1d///rpc_12base_max
		{0xf4 , 0x26},//////24///28///26///28///24///28///20///1f///1e///1d///rpc_13base_max
		///ae gain &status
		{0xfd , 0x01},
		{0x04 , 0xe0},///rpc_max_indr
		{0x05 , 0x26},///1e///rpc_min_indr 
		{0x0a , 0xa0},///rpc_max_outdr
		{0x0b,0x26},//rpc_min_outdr
		{0x5a,0x40},//dp rpc   
		{0xfd,0x02}, 
		{0xbc,0xa0},//rpc_heq_low
		{0xbd,0x80},//rpc_heq_dummy
		{0xb8,0x80},//mean_normal_dummy
		{0xb9,0x90},//mean_dummy_normal

		//ae target
		{0xfd,0x01}, 
		{0xeb,SP0718_P1_0xeb},//78 
		{0xec,SP0718_P1_0xec},//78
		{0xed,0x0a},	
		{0xee,0x10},

		//lsc       
		{0xfd,0x01},
		{0x26,0x30},
		{0x27,0x2c},
		{0x28,0x07},
		{0x29,0x08},
		{0x2a,0x40},
		{0x2b,0x03},
		{0x2c,0x00},
		{0x2d,0x00},

		{0xa1,0x24},
		{0xa2,0x27},
		{0xa3,0x27},
		{0xa4,0x2b},
		{0xa5,0x1c},
		{0xa6,0x1a},
		{0xa7,0x1a},
		{0xa8,0x1a},
		{0xa9,0x18},
		{0xaa,0x1c},
		{0xab,0x17},
		{0xac,0x17},
		{0xad,0x08},
		{0xae,0x08},
		{0xaf,0x08},
		{0xb0,0x00},
		{0xb1,0x00},
		{0xb2,0x00},
		{0xb3,0x00},
		{0xb4,0x00},
		{0xb5,0x02},
		{0xb6,0x06},
		{0xb7,0x00},
		{0xb8,0x00},


		//DP       
		{0xfd,0x01},
		{0x48,0x09},
		{0x49,0x99},

		//awb       
		{0xfd,0x01},
		{0x32,0x05},
		{0xfd,0x00},
		{0xe7,0x03},
		{0xfd,0x02},
		{0x26,0xc8},
		{0x27,0xb6},
		{0xfd,0x00},
		{0xe7,0x00},
		{0xfd,0x02},
		{0x1b,0x80},
		{0x1a,0x80},
		{0x18,0x26},
		{0x19,0x28},
		{0xfd,0x02},
		{0x2a,0x00},
		{0x2b,0x08},
		{0x28,0xef},//0xa0//f8
		{0x29,0x20},

		//d65 90  e2 93
		{0x66,0x42},//0x59//0x60////0x58//4e//0x48
		{0x67,0x62},//0x74//0x70//0x78//6b//0x69
		{0x68,0xee},//0xd6//0xe3//0xd5//cb//0xaa
		{0x69,0x18},//0xf4//0xf3//0xf8//ed
		{0x6a,0xa6},//0xa5
		//indoor 91
		{0x7c,0x3b},//0x45//30//41//0x2f//0x44
		{0x7d,0x5b},//0x70//60//55//0x4b//0x6f
		{0x7e,0x15},//0a//0xed
		{0x7f,0x39},//23//0x28
		{0x80,0xaa},//0xa6
		//cwf   92 
		{0x70,0x3e},//0x38//41//0x3b
		{0x71,0x59},//0x5b//5f//0x55
		{0x72,0x31},//0x30//22//0x28
		{0x73,0x4f},//0x54//44//0x45
		{0x74,0xaa},
		//tl84  93 
		{0x6b,0x1b},//0x18//11
		{0x6c,0x3a},//0x3c//25//0x2f
		{0x6d,0x3e},//0x3a//35
		{0x6e,0x59},//0x5c//46//0x52
		{0x6f,0xaa},
		//f    94
		{0x61,0xea},//0x03//0x00//f4//0xed
		{0x62,0x03},//0x1a//0x25//0f//0f
		{0x63,0x6a},//0x62//0x60//52//0x5d
		{0x64,0x8a},//0x7d//0x85//70//0x75//0x8f
		{0x65,0x6a},//0xaa//6a

		{0x75,0x80},
		{0x76,0x20},
		{0x77,0x00},
		{0x24,0x25},

		//ڵƫʹ//ƫ
		{0x20,0xd8},
		{0x21,0xa3},//82//a8ƫնȻеƫ
		{0x22,0xd0},//e3//bc
		{0x23,0x86},

		//outdoor r\b range
		{0x78,0xc3},//d8
		{0x79,0xba},//82
		{0x7a,0xa6},//e3
		{0x7b,0x99},//86


		//skin 
		{0x08,0x15},//
		{0x09,0x04},//
		{0x0a,0x20},//
		{0x0b,0x12},//
		{0x0c,0x27},//
		{0x0d,0x06},//
		{0x0e,0x63},//

		//wt th
		{0x3b,0x10},
		//gw
		{0x31,0x60},
		{0x32,0x60},
		{0x33,0xc0},
		{0x35,0x6f},

		// sharp
		{0xfd,0x02},
		{0xde,0x0f},
		{0xd2,0x02},//6//ƺڰױߣ0-ߴ֣f-ϸ
		{0xd3,0x06},
		{0xd4,0x06},
		{0xd5,0x06},
		{0xd7,0x20},//10//2xжֵ
		{0xd8,0x30},//24//1A//4x
		{0xd9,0x38},//28//8x
		{0xda,0x38},//16x
		{0xdb,0x08},//
		{0xe8,0x58},//48//ǿ
		{0xe9,0x48},
		{0xea,0x30},
		{0xeb,0x20},
		{0xec,0x48},//60//80
		{0xed,0x48},//50//60
		{0xee,0x30},
		{0xef,0x20},
		//ƽ̹
		{0xf3,0x50},
		{0xf4,0x10},
		{0xf5,0x10},
		{0xf6,0x10},
		//dns       
		{0xfd,0x01},
		{0x64,0x44},//طԵƽ  //0-ǿ8-
		{0x65,0x22},
		{0x6d,0x04},//8//ǿƽƽ̹ƽֵ
		{0x6e,0x06},//8
		{0x6f,0x10},
		{0x70,0x10},
		{0x71,0x08},//0d//ƽƽ̹ƽֵ	
		{0x72,0x12},//1b
		{0x73,0x1c},//20
		{0x74,0x24},
		{0x75,0x44},//[7:4]ƽ̹ǿȣ[3:0]ƽ̹ǿȣ0-ǿ8-
		{0x76,0x02},//46
		{0x77,0x02},//33
		{0x78,0x02},
		{0x81,0x10},//18//2x//жֵǿƽƽ
		{0x82,0x20},//30//4x
		{0x83,0x30},//40//8x
		{0x84,0x48},//50//16x
		{0x85,0x0c},//12/8+reg0x81 ڶֵƽ̹ͷƽ̹
		{0xfd,0x02},
		{0xdc,0x0f},

		//gamma    
		{0xfd,0x01},
		{0x8b,0x00},//00//00     
		{0x8c,0x0a},//0c//09     
		{0x8d,0x16},//19//17     
		{0x8e,0x1f},//25//24     
		{0x8f,0x2a},//30//33     
		{0x90,0x3c},//44//47     
		{0x91,0x4e},//54//58     
		{0x92,0x5f},//61//64     
		{0x93,0x6c},//6d//70     
		{0x94,0x82},//80//81     
		{0x95,0x94},//92//8f     
		{0x96,0xa6},//a1//9b     
		{0x97,0xb2},//ad//a5     
		{0x98,0xbf},//ba//b0     
		{0x99,0xc9},//c4//ba     
		{0x9a,0xd1},//cf//c4     
		{0x9b,0xd8},//d7//ce     
		{0x9c,0xe0},//e0//d7     
		{0x9d,0xe8},//e8//e1     
		{0x9e,0xef},//ef//ea     
		{0x9f,0xf8},//f7//f5     
		{0xa0,0xff},//ff//ff     
		//CCM      
		{0xfd,0x02},
		{0x15,0xd0},//b>th
		{0x16,0x95},//r<th
		//gcͷƫ
		//!F        
		{0xa0,0x80},//80
		{0xa1,0x00},//00
		{0xa2,0x00},//00
		{0xa3,0x00},//06
		{0xa4,0x8c},//8c
		{0xa5,0xf4},//ed
		{0xa6,0x0c},//0c
		{0xa7,0xf4},//f4
		{0xa8,0x80},//80
		{0xa9,0x00},//00
		{0xaa,0x30},//30
		{0xab,0x0c},//0c 
		//F        
		{0xac,0x8c},
		{0xad,0xf4},
		{0xae,0x00},
		{0xaf,0xed},
		{0xb0,0x8c},
		{0xb1,0x06},
		{0xb2,0xf4},
		{0xb3,0xf4},
		{0xb4,0x99},
		{0xb5,0x0c},
		{0xb6,0x03},
		{0xb7,0x0f},

		//sat u     
		{0xfd,0x01},
		{0xd3,0x9c},//0x88//50
		{0xd4,0x98},//0x88//50
		{0xd5,0x8c},//50
		{0xd6,0x84},//50
		//sat v   
		{0xd7,0x9c},//0x88//50
		{0xd8,0x98},//0x88//50
		{0xd9,0x8c},//50
		{0xda,0x84},//50
		//auto_sat  
		{0xdd,0x30},
		{0xde,0x10},
		{0xd2,0x01},//autosa_en
		{0xdf,0xff},//a0//y_mean_th

		//uv_th     
		{0xfd,0x01},
		{0xc2,0xaa},
		{0xc3,0xaa},
		{0xc4,0x66},
		{0xc5,0x66}, 

		//heq
		{0xfd,0x01},
		{0x0f,0xff},
		{0x10,SP0718_P1_0x10}, //out
		{0x14,SP0718_P1_0x14}, 
		{0x11,SP0718_P1_0x11}, //nr
		{0x15,SP0718_P1_0x15},  
		{0x12,SP0718_P1_0x12}, //dummy
		{0x16,SP0718_P1_0x16}, 
		{0x13,SP0718_P1_0x13}, //low 	
		{0x17,SP0718_P1_0x17},   	

		{0xfd,0x01},
		{0xcd,0x20},
		{0xce,0x1f},
		{0xcf,0x20},
		{0xd0,0x55},  
		//auto 
		{0xfd,0x01},
		{0xfb,0x33},
		{0x32,0x15},
		{0x33,0xff},
		{0x34,0xe7},
		{0x35,0x00},                                  

		{0xff , 0xff}                                 
};


static struct sensor_reg sensor_set_preview_SP0718[] = {
	{0xfd},{0x00},
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_set_capture_SP0718[] = {
	{0xfd},{0x00},
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

 struct sensor_reg* sensor_reg_common_SP0718[3] =
{
	sensor_initialize_SP0718,
	sensor_set_preview_SP0718,
	sensor_set_capture_SP0718	
};

static struct sensor_reg sensor_brightness_0_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_brightness_1_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_brightness_2_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_brightness_3_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_brightness_4_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_brightness_SP0718[5] =
{
	sensor_brightness_0_SP0718,
	sensor_brightness_1_SP0718,
	sensor_brightness_2_SP0718,
	sensor_brightness_3_SP0718,
	sensor_brightness_4_SP0718
};


static struct sensor_reg sensor_awb_auto_SP0718[] = {
	{0xfd,0x02},                      
	{0x26,0xc8},		                  
	{0x27,0xb6},                      
	{0xfd,0x01}, 		
	{0x32,0x15},   
	{0xfd,0x00},  
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_awb_daylight_SP0718[] = {
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x7b},		             
	{0x27,0xd3},		             
	{0xfd,0x00},   	
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};


//  WB-CWF = incandescent
static struct sensor_reg sensor_awb_incandescent_SP0718[] = {
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x75},		             
	{0x27,0xe2},		             
	{0xfd,0x00}, 
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

//  WB-A28 = fluorescent
static struct sensor_reg sensor_awb_fluorescent_SP0718[] = {
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x91},		             
	{0x27,0xc8},		             
	{0xfd,0x00}, 
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_awb_cloudy_SP0718[] = {
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xdc},		             
	{0x27,0xe5},		             
	{0xfd,0x00}, 		
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_awb_sunset_SP0718[] = {
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0xc8},		             
	{0x27,0x89},		             
	{0xfd,0x00}, 	
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_awb_SP0718[6] =
{
	sensor_awb_auto_SP0718,
	sensor_awb_daylight_SP0718,
	sensor_awb_incandescent_SP0718,
	sensor_awb_fluorescent_SP0718,
	sensor_awb_cloudy_SP0718,
	sensor_awb_sunset_SP0718
	
};


static struct sensor_reg sensor_iso_auto_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_iso_100_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_iso_200_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_iso_400_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_iso_SP0718[4] =
{
	sensor_iso_auto_SP0718,
	sensor_iso_100_SP0718,
	sensor_iso_200_SP0718,
	sensor_iso_400_SP0718
};


static struct sensor_reg sensor_effect_normal_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_effect_gray_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_effect_negative_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_effect_sepia_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_effect_sharpness_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_effect_sketch_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_effect_SP0718[6] =
{
	sensor_effect_normal_SP0718,
	sensor_effect_gray_SP0718,
	sensor_effect_negative_SP0718,
	sensor_effect_sepia_SP0718,
	sensor_effect_sharpness_SP0718,
	sensor_effect_sketch_SP0718,
};


static struct sensor_reg sensor_reg_flipnone_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_reg_hflip_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_reg_vflip_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_reg_hvflip_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_flip_SP0718[4] =
{
	sensor_reg_flipnone_SP0718,
	sensor_reg_hflip_SP0718,
	sensor_reg_vflip_SP0718,
	sensor_reg_hvflip_SP0718,
};


static struct sensor_reg sensor_secne_auto_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_secne_night_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_secne_landscape_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_secne_portrait_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_secne_sport_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_scene_SP0718[5] =
{
	sensor_secne_auto_SP0718,
	sensor_secne_night_SP0718,
	sensor_secne_landscape_SP0718,
	sensor_secne_portrait_SP0718,
	sensor_secne_sport_SP0718
};

static struct sensor_reg sensor_me_mtrix_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_me_center_weighted_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_me_spot_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_metering_exposure_SP0718[3] =
{
	sensor_me_mtrix_SP0718,
	sensor_me_center_weighted_SP0718,
	sensor_me_spot_SP0718
};

static struct sensor_reg sensor_af_single_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_af_closeness_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_af_manual_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_af_SP0718[2] =
{
	sensor_af_single_SP0718,
    sensor_af_closeness_SP0718
//	sensor_af_manual_SP0718
};

static struct sensor_reg sensor_sw_standby_SP0718[] = {
    {SP0718_REG_TERM, SP0718_VAL_TERM}
};


static struct sensor_reg sensor_exposure_L2_SP0718[] = {
	{0xfd,0x01},
	{0xdb,0x20},//level 0
	//	EV +5
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_exposure_L1_SP0718[] = {
	{0xfd,0x01},
	{0xdb,0x10},//level 0
	//	EV +3
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

//default
static struct sensor_reg sensor_exposure_L0_SP0718[] = {
	{0xfd,0x01},
	{0xdb,0x00},//level 0
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_exposure_LN1_SP0718[] = {
	{0xfd,0x01},
	{0xdb,0xf0},//level 0
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

static struct sensor_reg sensor_exposure_LN2_SP0718[] = {
	{0xfd,0x01},
	{0xdb,0xe0},//level 0
	{SP0718_REG_TERM, SP0718_VAL_TERM}
};

struct sensor_reg* sensor_reg_exposure_SP0718[5] =
{	
	sensor_exposure_L2_SP0718,
	sensor_exposure_L1_SP0718,
	sensor_exposure_L0_SP0718,
	sensor_exposure_LN1_SP0718,
	sensor_exposure_LN2_SP0718	
};
static int write_regs_SP0718(const struct sensor_reg reglist[])
{
	int err;
	int err_cnt = 0;	
	unsigned char data[132];
	unsigned char bytes;
	struct sensor_reg *next = reglist;
	
	while (!((next->reg == SP0718_REG_TERM) && (next->val == SP0718_VAL_TERM)))
	{
		bytes = 0;
		data[bytes++] = (unsigned char)(next->reg) & 0xff; 	
		data[bytes]   = (unsigned char)(next->val) & 0xff;
		
        err = DDI_I2C_Write(data, 1, 1);

        if (err)
        {
			err_cnt++;
			if(err_cnt >= 3)
			{
				printk("ERROR: Sensor I2C !!!! \n"); 
				return err;
			}
        }
        else
        {
        	err_cnt = 0;
            next++;
        }
	}

	return 0;
}
static char SP0718ReadReg(char reg)
{
    char val;
    int ret;
	
	ret = DDI_I2C_Read(reg,1,&val,1);
	if(ret != 0)
		 printk("OV7675ReadReg error:reg = 0x%x\n",reg);

    return val;
}
static void SP0718WriteReg(unsigned short reg,int val)
{
	unsigned char data[2];
    int ret;
	
	data[0]= (u8)reg&0xff;
	data[1]= (u8)val&0xff;

    ret = DDI_I2C_Write(data, 1, 1);
	if(ret != 0)
		 printk("OV7675WriteReg error:reg = 0x%x\n",reg);
}
static void sensor_SP0718_detect(void)
{
	unsigned char pid=0;

	SP0718WriteReg(PAGE_REG,PAGE0); //[2:0]page select
	
	DDI_I2C_Read(SP0718_REG_PID,1,&pid,1); //PAGE 23
	printk("-->%s chip_id(0x%x) must is 0x%x\n",__func__,pid,SP0718_CHIP_ID);
}

static int sensor_open_SP0718(bool bChangeCamera)
{
printk("-->%s\n",__FUNCTION__);
	//sensor_powerdown_disable();
	//reset on io
	//sensor_reset_low();
	//active mclk before power on
	
	mdelay(10);
	//power supply
	sensor_power_enable();
	mdelay(10);
	//standby off io
	CIF_Open();
	mdelay(1);
	sensor_powerdown_disable();
	mdelay(10);
	sensor_powerdown_enable();
	mdelay(10);
	sensor_powerdown_disable();
	mdelay(100);
	//reset after power on
	//sensor_reset_high();
	//mdelay(10);
	//sensor_reset_low();
	//mdelay(100);
	//sensor_reset_high();
	//mdelay(10);

	sensor_SP0718_detect();


	return write_regs_SP0718(sensor_reg_common_SP0718[0]);
}
                               

/*************************************************************************
* FUNCTION
*    SP0718PreviewSetting
*
* DESCRIPTION
*    This function config Preview setting related registers of CMOS sensor.
*
* PARAMETERS
*    None
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void SP0718PreviewSetting(void)
{
	write_regs_SP0718(sensor_reg_common_SP0718[1]);
}
static void SP0718CaptureSetting(void)
{
	write_regs_SP0718(sensor_reg_common_SP0718[2]);
}

static int sensor_close_SP0718(void)
{
	printk("sensor_close_SP0718\n");
	msleep(10);	
	
	CIF_ONOFF(OFF);

	sensor_reset_low();
	sensor_powerdown_disable();
	CIF_Close();
	sensor_power_disable();

	msleep(5);

	return 0;
}

static int sensor_powerdown_SP0718(bool bChangeCamera)
{
	if(bChangeCamera){

		sensor_powerdown_disable();
		sensor_delay(10);

		sensor_reset_low();
		sensor_delay(10);

		CIF_Open();
		sensor_delay(40);

		sensor_reset_high();

		msleep(10);	

		sensor_powerdown_enable();
		msleep(10);

	
		printk("SP0718 change camera powerdown! \n");
	}
	else{		
		sensor_reset_low();
		sensor_power_disable();
		sensor_powerdown_enable();
			
		printk("SP0718 first/last camera powerdown! \n");
	}

	return 0;
}

static int sensor_preview_SP0718(void)
{
printk("sensor_preview_SP0718\n");
	sensor_SP0718_detect();
	SP0718PreviewSetting();

	return 0;//write_regs_SP0718(sensor_reg_common_SP0718[1]);
}

static int sensor_capture_SP0718(void)  
{
	SP0718CaptureSetting();

	return 0;//write_regs_SP0718(sensor_reg_common_SP0718[2]);
//	return write_regs_SP0718(sensor_reg_af_SP0718[0]);
}

static int sensor_capturecfg_SP0718(int width, int height)
{
	return 0;
}

static int sensor_zoom_SP0718(int val)
{
	return 0;
}

// val == 0 : normal focus
// val == 1 : closeness focus
static int sensor_autofocus_SP0718(int val)  
{
	printk("sensor_autofocus_SP0718\n");

	return 0;
}

static int sensor_effect_SP0718(int val)
{		

	return 0;
}

static int sensor_flip_SP0718(int val)
{
	return write_regs_SP0718(sensor_reg_flip_SP0718[val]);
}

static int sensor_iso_SP0718(int val)
{
	return write_regs_SP0718(sensor_reg_iso_SP0718[val]);
}

static int sensor_me_SP0718(int val)
{
	return write_regs_SP0718(sensor_reg_metering_exposure_SP0718[val]);
}
static int SP0718AwbEnable(bool Enable)
{
  return 0;
}
static int sensor_wb_SP0718(int val)
{
printk("-->%s val = 0x%x",__func__,val);
	return write_regs_SP0718(sensor_reg_awb_SP0718[val]);
}

static int sensor_bright_SP0718(int val)
{
	return write_regs_SP0718(sensor_reg_brightness_SP0718[val]);
}

static int sensor_scene_SP0718(int val)
{
	return write_regs_SP0718(sensor_reg_scene_SP0718[val]);
}

static int sensor_focus_mode_SP0718(int val)
{
	printk("sensor_focus_mode_SP0718[%d] \n", val);

	return 0;
}
static int sensor_check_esd_SP0718(int val)
{
	return 0;
}

static int sensor_exposure_SP0718(int val)
{
	printk("sensor exposure value[%d] \n", val);

	switch(val){
		case 20: val = 0; break;
		case 10: val = 1; break;
		case 0: val = 2; break;
		case -10: val = 3; break;
		case -20: val = 4; break;
		default: val = 2; break;
	}
	
	return write_regs_SP0718(sensor_reg_exposure_SP0718[val]);
}

static int sensor_check_luma_SP0718(int val)
{
	return 0;
}

/**********************************************************
*    Function  
**********************************************************/
#if defined(CONFIG_VIDEO_DUAL_CAMERA_SUPPORT)
void sensor_info_init_SP0718(TCC_SENSOR_INFO_TYPE *sensor_info)
{
	sensor_info->i2c_addr 				= 0x42;
	sensor_info->reg_term 				= 0x0000;
	sensor_info->val_term 				= 0x0000;
	#if defined(CONFIG_USE_ISP)
		sensor_info->m_clock 					= 247500;
		sensor_info->m_clock_source 			= PCDIRECTPLL2;
		sensor_info->s_clock 					= 1440000;
		sensor_info->s_clock_source 			= PCDIRECTPLL2;
	#else  // CONFIG_USE_ISP
		sensor_info->m_clock 					= 247500;
		sensor_info->m_clock_source 			= PCDIRECTPLL2;
		sensor_info->s_clock 					= 1440000;
		sensor_info->s_clock_source 			= PCDIRECTPLL2;
	#endif  // CONFIG_USE_ISP

	sensor_info->preview_w 				= 640;
	sensor_info->preview_h 				= 480;
	sensor_info->preview_zoom_offset_x 	= 8;
	sensor_info->preview_zoom_offset_y 	= 6;
	sensor_info->capture_w 				= 640;
	sensor_info->capture_h 				= 480;
	sensor_info->capture_zoom_offset_x 	= 8;
	sensor_info->capture_zoom_offset_y 	= 6;
	sensor_info->max_zoom_step 			= 15;
	sensor_info->cam_capchg_width 		= 640;
	sensor_info->framerate				= 10;
	sensor_info->p_clock_pol 				= POSITIVE_EDGE;
	sensor_info->v_sync_pol 				= ACT_HIGH;//ACT_LOW;
	sensor_info->h_sync_pol                 = ACT_HIGH;
	#if defined(CONFIG_ARCH_TCC892X)
		sensor_info->de_pol                 = ACT_LOW;
	#endif	
	sensor_info->format 					= M420_ZERO;
	sensor_info->capture_skip_frame 		= 1;
	sensor_info->sensor_sizes 			= sensor_sizes_SP0718;
}

void sensor_init_fnc_SP0718(SENSOR_FUNC_TYPE *sensor_func)
{
	sensor_func->Open 					= sensor_open_SP0718;
	sensor_func->Close 					= sensor_close_SP0718;
	sensor_func->PowerDown				= sensor_powerdown_SP0718;

	sensor_func->Set_Preview 			= sensor_preview_SP0718;
	sensor_func->Set_Capture 			= sensor_capture_SP0718;
	sensor_func->Set_CaptureCfg 			= sensor_capturecfg_SP0718;

	sensor_func->Set_Zoom 				= sensor_zoom_SP0718;
	sensor_func->Set_AF 				= sensor_autofocus_SP0718;
	sensor_func->Set_Effect 				= sensor_effect_SP0718;
	sensor_func->Set_Flip 				= sensor_flip_SP0718;
	sensor_func->Set_ISO 				= sensor_iso_SP0718;
	sensor_func->Set_ME 				= sensor_me_SP0718;
	sensor_func->Set_WB 				= sensor_wb_SP0718;
	sensor_func->Set_Bright 				= sensor_bright_SP0718;
	sensor_func->Set_Scene 				= sensor_scene_SP0718;
	sensor_func->Set_Exposure			= sensor_exposure_SP0718;
	sensor_func->Set_FocusMode			= sensor_focus_mode_SP0718;

	sensor_func->Check_ESD 				= sensor_check_esd_SP0718;
	sensor_func->Check_Luma 			= sensor_check_luma_SP0718;
}
#else
void sensor_init_fnc(SENSOR_FUNC_TYPE *sensor_func)
{
	sensor_func->Open 					= sensor_open_SP0718;
	sensor_func->Close 					= sensor_close_SP0718;
	sensor_func->PowerDown				= sensor_powerdown_SP0718;

	sensor_func->Set_Preview 			= sensor_preview_SP0718;
	sensor_func->Set_Capture 			= sensor_capture_SP0718;
	sensor_func->Set_CaptureCfg 			= sensor_capturecfg_SP0718;

	sensor_func->Set_Zoom 				= sensor_zoom_SP0718;
	sensor_func->Set_AF 				= sensor_autofocus_SP0718;
	sensor_func->Set_Effect 				= sensor_effect_SP0718;
	sensor_func->Set_Flip 				= sensor_flip_SP0718;
	sensor_func->Set_ISO 				= sensor_iso_SP0718;
	sensor_func->Set_ME 				= sensor_me_SP0718;
	sensor_func->Set_WB 				= sensor_wb_SP0718;
	sensor_func->Set_Bright 				= sensor_bright_SP0718;
	sensor_func->Set_Scene 				= sensor_scene_SP0718;
	sensor_func->Set_Exposure			= sensor_exposure_SP0718;
	sensor_func->Set_FocusMode			= sensor_focus_mode_SP0718;

	sensor_func->Check_ESD 				= sensor_check_esd_SP0718;
	sensor_func->Check_Luma 			= sensor_check_luma_SP0718;
}
#endif
#endif


