/******************************************************************************
 ** Copyright (c) 
 ** File Name:		sensor_SP0718.c 										  *
 ** Author: 													  *
 ** DATE:															  *
 ** Description:   This file contains driver for sensor SP0718. 
 ** 														 
 ******************************************************************************

 ******************************************************************************
 ** 					   Edit History 									  *
 ** ------------------------------------------------------------------------- *
 ** DATE		   NAME 			DESCRIPTION 							  *
 ** 	  
 ******************************************************************************/

/**---------------------------------------------------------------------------*
 ** 						Dependencies									  *
 **---------------------------------------------------------------------------*/
#include "ms_customize_trc.h"
#include "sensor_cfg.h"
#include "sensor_drv.h"
#include "os_api.h"
#include "chip.h"
#include "dal_dcamera.h"

/**---------------------------------------------------------------------------*
 ** 						Compiler Flag									  *
 **---------------------------------------------------------------------------*/
#ifdef	 __cplusplus
	extern	 "C" 
	{
#endif
/**---------------------------------------------------------------------------*
 ** 					Extern Function Declaration 						  *
 **---------------------------------------------------------------------------*/
extern uint32 OS_TickDelay(uint32 ticks);
/*
extern void Sensor_SetMCLK(uint32 mclk);
*/
//extern ERR_I2C_E I2C_WriteCmdArrNoStop(uint8 addr, uint8 *pCmd, uint32 len, BOOLEAN ack_en);




 /**---------------------------------------------------------------------------*
 **                         Const variables                                   *
 **---------------------------------------------------------------------------*/
#define SP0718_I2C_ADDR_W    			0x42    //  lxy test
#define SP0718_I2C_ADDR_R    			0x43
#define SP0718_I2C_ACK				       0x0
#define SENSOR_GAIN_SCALE		              16    //  lxy test
LOCAL uint32			s_sensor_mclk		 = 0;
LOCAL uint32			Antiflicker		 = DCAMERA_FLICKER_50HZ;

#define SP0718_NORMAL_Y0ffset  0x20
#define SP0718_LOWLIGHT_Y0ffset  0x25
//AE target
#define  SP0718_P1_0xeb  0x78
#define  SP0718_P1_0xec  0x6c
//HEQ
#define  SP0718_P1_0x10  0x00//outdoor
#define  SP0718_P1_0x14  0x20
#define  SP0718_P1_0x11  0x00//nr
#define  SP0718_P1_0x15  0x18
#define  SP0718_P1_0x12  0x00//dummy
#define  SP0718_P1_0x16  0x10
#define  SP0718_P1_0x13  0x00//low
#define  SP0718_P1_0x17  0x00


/**---------------------------------------------------------------------------*
 ** 					Local Function Prototypes							  *
 **---------------------------------------------------------------------------*/
LOCAL uint32 set_SP0718_ae_enable(uint32 enable);
LOCAL uint32 set_hmirror_enable(uint32 enable);
LOCAL uint32 set_vmirror_enable(uint32 enable);
LOCAL uint32 set_preview_mode(uint32 preview_mode);
LOCAL uint32 SP0718_Identify(uint32 param);
LOCAL uint32 SP0718_Power_On(uint32 power_on);
LOCAL uint32 set_brightness(uint32 level);
LOCAL uint32 set_contrast(uint32 level);
LOCAL uint32 set_sharpness(uint32 level);
LOCAL uint32 set_saturation(uint32 level);
LOCAL uint32 set_image_effect(uint32 effect_type);

LOCAL uint32 read_ev_value(uint32 value);
LOCAL uint32 write_ev_value(uint32 exposure_value);
LOCAL uint32 read_gain_value(uint32 value);
LOCAL uint32 write_gain_value(uint32 gain_value);
LOCAL uint32 read_gain_scale(uint32 value);
LOCAL uint32 set_frame_rate(uint32 param);
LOCAL uint32 SP0718_set_work_mode(uint32 mode);

LOCAL uint32 set_SP0718_ev(uint32 level);
LOCAL uint32 set_SP0718_awb(uint32 mode);
LOCAL uint32 set_SP0718_anti_flicker(uint32 mode);
LOCAL uint32 set_SP0718_video_mode(uint32 mode);

/**---------------------------------------------------------------------------*
 ** 						Local Variables 								 *
 **---------------------------------------------------------------------------*/
LOCAL void SP0718_Delayms (uint32 ms)	
{
	uint32 t1, t2;
	
	t1 = t2 = SCI_GetTickCount ();
	
	do{
		t2 = SCI_GetTickCount ();
	}while (t2 < (t1+ms));
}
__align(4) const SENSOR_REG_T SP0718_YUV_640X480[]=

{		//saturday param test
		{0xfd , 0x00},
		{0x1C , 0x00},   //0x00  lxy  
		{0x31 , 0x10},   //0x00  lxy  
		{0x27 , 0xb3},///0xb3	///2x gain
		{0x1b , 0x17},
		{0x26 , 0xaa},
		{0x37 , 0x02},
		{0x28 , 0x8f},
		{0x1a , 0x73},
		{0x1e , 0x1b},
		{0x21 , 0x06},///blackout voltage
		{0x22,0x2a},  //colbias
		{0x0f,0x3f},
		{0x10,0x3e},
		{0x11,0x00},
		{0x12,0x01},
		{0x13,0x3f},
		{0x14,0x04},
		{0x15,0x30},
		{0x16,0x31},
		{0x17,0x01},
		{0x69,0x31},
		{0x6a,0x2a},
		{0x6b,0x33},
		{0x6c,0x1a},
		{0x6d,0x32},
		{0x6e,0x28},
		{0x6f,0x29},
		{0x70,0x34},
		{0x71,0x18},
		{0x36,0x00},//02 delete badframe
		{0xfd,0x01},
		{0x5d,0x51},//position
		{0xf2,0x19},

		//Blacklevel
		{0x1f,0x10},
		{0x20,0x1f},
		//pregain 
		{0xfd,0x02},
		{0x00,0x88},
		{0x01,0x88},
		//SI15_SP0718 24M 50Hz 15-8fps 
		//ae setting
		{0xfd,0x00},
		{0x03,0x01},
		{0x04,0xce},
		{0x06,0x00},
		{0x09,0x02},
		{0x0a,0xc4},
		{0xfd,0x01},
		{0xef,0x4d},
		{0xf0,0x00},
		{0x02,0x0c},
		{0x03,0x01},
		{0x06,0x47},
		{0x07,0x00},
		{0x08,0x01},
		{0x09,0x00},
		//Status   
		{0xfd,0x02},
		{0xbe,0x9c},
		{0xbf,0x03},
		{0xd0,0x9c},
		{0xd1,0x03},
		{0xfd,0x01},
		{0x5b,0x03},
		{0x5c,0x9c},

		//rpc
		{0xfd,0x01},
		{0xe0,0x40},////24//4c//48//4c//44//4c//3e//3c//3a//38//rpc_1base_max
		{0xe1,0x30},////24//3c//38//3c//36//3c//30//2e//2c//2a//rpc_2base_max
		{0xe2,0x2e},////24//34//30//34//2e//34//2a//28//26//26//rpc_3base_max
		{0xe3,0x2a},////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_4base_max
		{0xe4,0x2a},////24//2a//2e//2c//2e//2a//2e//26//24//22//rpc_5base_max
		{0xe5,0x28},////24//2c//2a//2c//28//2c//24//22//20//rpc_6base_max
		{0xe6,0x28},////24//2c//2a//2c//28//2c//24//22//20//rpc_7base_max
		{0xe7,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_8base_max
		{0xe8,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_9base_max
		{0xe9,0x26},////24//2a//28//2a//26//2a//22//20//20//1e//rpc_10base_max
		{0xea,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_11base_max
		{0xf3,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_12base_max
		{0xf4,0x26},////24//28//26//28//24//28//20//1f//1e//1d//rpc_13base_max
		//ae gain &status
		{0xfd,0x01},
		{0x04,0xe0},//rpc_max_indr
		{0x05,0x26},//1e//rpc_min_indr 
		{0x0a,0xa0},//rpc_max_outdr
		{0x0b,0x26},//rpc_min_outdr
		{0x5a,0x40},//dp rpc   
		{0xfd,0x02}, 
		{0xbc,0xa0},//rpc_heq_low
		{0xbd,0x80},//rpc_heq_dummy
		{0xb8,0x80},//mean_normal_dummy
		{0xb9,0x90},//mean_dummy_normal

		//ae target
		{0xfd,0x01}, 
		{0xeb,SP0718_P1_0xeb},//78 
		{0xec,SP0718_P1_0xec},//78
		{0xed,0x0a},	
		{0xee,0x10},

		//lsc       
		{0xfd,0x01},
		{0x26,0x30},
		{0x27,0x2c},
		{0x28,0x07},
		{0x29,0x08},
		{0x2a,0x40},
		{0x2b,0x03},
		{0x2c,0x00},
		{0x2d,0x00},
		  
		{0xa1,0x24},
		{0xa2,0x27},
		{0xa3,0x27},
		{0xa4,0x2b},
		{0xa5,0x1c},
		{0xa6,0x1a},
		{0xa7,0x1a},
		{0xa8,0x1a},
		{0xa9,0x18},
		{0xaa,0x1c},
		{0xab,0x17},
		{0xac,0x17},
		{0xad,0x08},
		{0xae,0x08},
		{0xaf,0x08},
		{0xb0,0x00},
		{0xb1,0x00},
		{0xb2,0x00},
		{0xb3,0x00},
		{0xb4,0x00},
		{0xb5,0x02},
		{0xb6,0x06},
		{0xb7,0x00},
		{0xb8,0x00},
		   
		   
		//DP       
		{0xfd,0x01},
		{0x48,0x09},
		{0x49,0x99},
		    
		//awb       
		{0xfd,0x01},
		{0x32,0x05},
		{0xfd,0x00},
		{0xe7,0x03},
		{0xfd,0x02},
		{0x26,0xc8},
		{0x27,0xb6},
		{0xfd,0x00},
		{0xe7,0x00},
		{0xfd,0x02},
		{0x1b,0x80},
		{0x1a,0x80},
		{0x18,0x26},
		{0x19,0x28},
		{0xfd,0x02},
		{0x2a,0x00},
		{0x2b,0x08},
		{0x28,0xef},//0xa0//f8
		{0x29,0x20},

		//d65 90  e2 93
		{0x66,0x42},//0x59//0x60////0x58//4e//0x48
		{0x67,0x62},//0x74//0x70//0x78//6b//0x69
		{0x68,0xee},//0xd6//0xe3//0xd5//cb//0xaa
		{0x69,0x18},//0xf4//0xf3//0xf8//ed
		{0x6a,0xa6},//0xa5
		//indoor 91
		{0x7c,0x3b},//0x45//30//41//0x2f//0x44
		{0x7d,0x5b},//0x70//60//55//0x4b//0x6f
		{0x7e,0x15},//0a//0xed
		{0x7f,0x39},//23//0x28
		{0x80,0xaa},//0xa6
		//cwf   92 
		{0x70,0x3e},//0x38//41//0x3b
		{0x71,0x59},//0x5b//5f//0x55
		{0x72,0x31},//0x30//22//0x28
		{0x73,0x4f},//0x54//44//0x45
		{0x74,0xaa},
		//tl84  93 
		{0x6b,0x1b},//0x18//11
		{0x6c,0x3a},//0x3c//25//0x2f
		{0x6d,0x3e},//0x3a//35
		{0x6e,0x59},//0x5c//46//0x52
		{0x6f,0xaa},
		//f    94
		{0x61,0xea},//0x03//0x00//f4//0xed
		{0x62,0x03},//0x1a//0x25//0f//0f
		{0x63,0x6a},//0x62//0x60//52//0x5d
		{0x64,0x8a},//0x7d//0x85//70//0x75//0x8f
		{0x65,0x6a},//0xaa//6a
		  
		{0x75,0x80},
		{0x76,0x20},
		{0x77,0x00},
		{0x24,0x25},

		//ڵƫʹ//ƫ
		{0x20,0xd8},
		{0x21,0xa3},//82//a8ƫնȻеƫ
		{0x22,0xd0},//e3//bc
		{0x23,0x86},

		//outdoor r\b range
		{0x78,0xc3},//d8
		{0x79,0xba},//82
		{0x7a,0xa6},//e3
		{0x7b,0x99},//86


		//skin 
		{0x08,0x15},//
		{0x09,0x04},//
		{0x0a,0x20},//
		{0x0b,0x12},//
		{0x0c,0x27},//
		{0x0d,0x06},//
		{0x0e,0x63},//

		//wt th
		{0x3b,0x10},
		//gw
		{0x31,0x60},
		{0x32,0x60},
		{0x33,0xc0},
		{0x35,0x6f},

		// sharp
		{0xfd,0x02},
		{0xde,0x0f},
		{0xd2,0x02},//6//ƺڰױߣ0-ߴ֣f-ϸ
		{0xd3,0x06},
		{0xd4,0x06},
		{0xd5,0x06},
		{0xd7,0x20},//10//2xжֵ
		{0xd8,0x30},//24//1A//4x
		{0xd9,0x38},//28//8x
		{0xda,0x38},//16x
		{0xdb,0x08},//
		{0xe8,0x58},//48//ǿ
		{0xe9,0x48},
		{0xea,0x30},
		{0xeb,0x20},
		{0xec,0x48},//60//80
		{0xed,0x48},//50//60
		{0xee,0x30},
		{0xef,0x20},
		//ƽ̹
		{0xf3,0x50},
		{0xf4,0x10},
		{0xf5,0x10},
		{0xf6,0x10},
		//dns       
		{0xfd,0x01},
		{0x64,0x44},//طԵƽ  //0-ǿ8-
		{0x65,0x22},
		{0x6d,0x04},//8//ǿƽƽ̹ƽֵ
		{0x6e,0x06},//8
		{0x6f,0x10},
		{0x70,0x10},
		{0x71,0x08},//0d//ƽƽ̹ƽֵ	
		{0x72,0x12},//1b
		{0x73,0x1c},//20
		{0x74,0x24},
		{0x75,0x44},//[7:4]ƽ̹ǿȣ[3:0]ƽ̹ǿȣ0-ǿ8-
		{0x76,0x02},//46
		{0x77,0x02},//33
		{0x78,0x02},
		{0x81,0x10},//18//2x//жֵǿƽƽ
		{0x82,0x20},//30//4x
		{0x83,0x30},//40//8x
		{0x84,0x48},//50//16x
		{0x85,0x0c},//12/8+reg0x81 ڶֵƽ̹ͷƽ̹
		{0xfd,0x02},
		{0xdc,0x0f},
		   
		//gamma    
		{0xfd,0x01},
		{0x8b,0x00},//00//00     
		{0x8c,0x0a},//0c//09     
		{0x8d,0x16},//19//17     
		{0x8e,0x1f},//25//24     
		{0x8f,0x2a},//30//33     
		{0x90,0x3c},//44//47     
		{0x91,0x4e},//54//58     
		{0x92,0x5f},//61//64     
		{0x93,0x6c},//6d//70     
		{0x94,0x82},//80//81     
		{0x95,0x94},//92//8f     
		{0x96,0xa6},//a1//9b     
		{0x97,0xb2},//ad//a5     
		{0x98,0xbf},//ba//b0     
		{0x99,0xc9},//c4//ba     
		{0x9a,0xd1},//cf//c4     
		{0x9b,0xd8},//d7//ce     
		{0x9c,0xe0},//e0//d7     
		{0x9d,0xe8},//e8//e1     
		{0x9e,0xef},//ef//ea     
		{0x9f,0xf8},//f7//f5     
		{0xa0,0xff},//ff//ff     
		//CCM      
		{0xfd,0x02},
		{0x15,0xd0},//b>th
		{0x16,0x95},//r<th
		//gcͷƫ
		//!F        
		{0xa0,0x80},//80
		{0xa1,0x00},//00
		{0xa2,0x00},//00
		{0xa3,0x00},//06
		{0xa4,0x8c},//8c
		{0xa5,0xf4},//ed
		{0xa6,0x0c},//0c
		{0xa7,0xf4},//f4
		{0xa8,0x80},//80
		{0xa9,0x00},//00
		{0xaa,0x30},//30
		{0xab,0x0c},//0c 
		//F        
		{0xac,0x8c},
		{0xad,0xf4},
		{0xae,0x00},
		{0xaf,0xed},
		{0xb0,0x8c},
		{0xb1,0x06},
		{0xb2,0xf4},
		{0xb3,0xf4},
		{0xb4,0x99},
		{0xb5,0x0c},
		{0xb6,0x03},
		{0xb7,0x0f},
		    
		//sat u     
		{0xfd,0x01},
		{0xd3,0x9c},//0x88//50
		{0xd4,0x98},//0x88//50
		{0xd5,0x8c},//50
		{0xd6,0x84},//50
		//sat v   
		{0xd7,0x9c},//0x88//50
		{0xd8,0x98},//0x88//50
		{0xd9,0x8c},//50
		{0xda,0x84},//50
		//auto_sat  
		{0xdd,0x30},
		{0xde,0x10},
		{0xd2,0x01},//autosa_en
		{0xdf,0xff},//a0//y_mean_th
		    
		//uv_th     
		{0xfd,0x01},
		{0xc2,0xaa},
		{0xc3,0xaa},
		{0xc4,0x66},
		{0xc5,0x66}, 

		//heq
		{0xfd,0x01},
		{0x0f,0xff},
		{0x10,SP0718_P1_0x10}, //out
		{0x14,SP0718_P1_0x14}, 
		{0x11,SP0718_P1_0x11}, //nr
		{0x15,SP0718_P1_0x15},  
		{0x12,SP0718_P1_0x12}, //dummy
		{0x16,SP0718_P1_0x16}, 
		{0x13,SP0718_P1_0x13}, //low 	
		{0x17,SP0718_P1_0x17},   	

		{0xfd,0x01},
		{0xcd,0x20},
		{0xce,0x1f},
		{0xcf,0x20},
		{0xd0,0x55},  
		//auto 
		{0xfd,0x01},
		{0xfb,0x33},
		//{0x32,0x15},
		{0x33,0xff},
		{0x34,0xe7},
		{0x35,0x40},	  	

		{0xff , 0xff}                                 
};



LOCAL SENSOR_REG_TAB_INFO_T s_SP0718_resolution_Tab_YUV[]=
{
	// COMMON INIT
	{ADDR_AND_LEN_OF_ARRAY(SP0718_YUV_640X480), 640, 480, 12, SENSOR_IMAGE_FORMAT_YUV422},
	
	// YUV422 PREVIEW 1	
	{PNULL, 0, 640, 480,12, SENSOR_IMAGE_FORMAT_YUV422},
	{PNULL, 0, 0, 0, 0, 0},
	{PNULL, 0, 0, 0, 0, 0},
	{PNULL, 0, 0, 0, 0, 0},
	
	// YUV422 PREVIEW 2 
	{PNULL, 0, 0, 0, 0, 0},
	{PNULL, 0, 0, 0, 0, 0},
	{PNULL, 0, 0, 0, 0, 0},
	{PNULL, 0, 0, 0, 0, 0}

};

LOCAL SENSOR_IOCTL_FUNC_TAB_T s_SP0718_ioctl_func_tab = 
{
    // Internal 
    PNULL,
    PNULL,
    PNULL,
    SP0718_Identify,

    PNULL,			// write register
    PNULL,			// read  register	
    PNULL,
    PNULL,

    // External
    set_SP0718_ae_enable,
    set_hmirror_enable,
    set_vmirror_enable,

    set_brightness,
    set_contrast,
    set_sharpness,
    set_saturation,

    set_preview_mode,	
    set_image_effect,

	PNULL,	//	SP0718_BeforeSnapshot,
    PNULL,		//SP0718_After_Snapshot,

    PNULL,

    read_ev_value,
    write_ev_value,
    read_gain_value,
    write_gain_value,
    read_gain_scale,
    set_frame_rate,	
    PNULL,
    PNULL,
    set_SP0718_awb,
    PNULL,
    PNULL,
    set_SP0718_ev,
    PNULL,
    PNULL,
    PNULL,
    PNULL,
    PNULL,
    set_SP0718_anti_flicker,
    set_SP0718_video_mode,
    PNULL,
    PNULL
};

/**---------------------------------------------------------------------------*
 ** 						Global Variables								  *
 **---------------------------------------------------------------------------*/
 PUBLIC SENSOR_INFO_T g_SP0718_yuv_info =
{
	SP0718_I2C_ADDR_W,				// salve i2c write address
	SP0718_I2C_ADDR_R, 				// salve i2c read address
	
	0,								// bit0: 0: i2c register value is 8 bit, 1: i2c register value is 16 bit
									// bit2: 0: i2c register addr  is 8 bit, 1: i2c register addr  is 16 bit
									// other bit: reseved
	SENSOR_HW_SIGNAL_PCLK_P|\
	SENSOR_HW_SIGNAL_VSYNC_P|\
	SENSOR_HW_SIGNAL_HSYNC_P,		// bit0: 0:negative; 1:positive -> polarily of pixel clock
									// bit2: 0:negative; 1:positive -> polarily of horizontal synchronization signal
									// bit4: 0:negative; 1:positive -> polarily of vertical synchronization signal
									// other bit: reseved											
											
	// preview mode
	SENSOR_ENVIROMENT_NORMAL|\
	SENSOR_ENVIROMENT_NIGHT|\
	SENSOR_ENVIROMENT_SUNNY,		

	// image effect
	SENSOR_IMAGE_EFFECT_NORMAL|\
	SENSOR_IMAGE_EFFECT_BLACKWHITE|\
	SENSOR_IMAGE_EFFECT_RED|\
	SENSOR_IMAGE_EFFECT_GREEN|\
	SENSOR_IMAGE_EFFECT_BLUE|\
	SENSOR_IMAGE_EFFECT_YELLOW|\
	SENSOR_IMAGE_EFFECT_NEGATIVE|\
	SENSOR_IMAGE_EFFECT_CANVAS,
	// while balance mode
	0,
		
	7,								// bit[0:7]: count of step in brightness, contrast, sharpness, saturation
									// bit[8:31] reseved
	
	SENSOR_LOW_PULSE_RESET,			// reset pulse level
	100,								// reset pulse width(ms)
	
	SENSOR_HIGH_LEVEL_PWDN,			// 1: high level valid; 0: low level valid
	
	2,								// count of identify code
	0x02, 0x71,						// supply two code to identify sensor.
	0x02, 0x71,						// for Example: index = 0-> Device id, index = 1 -> version id	
									
	SENSOR_AVDD_2800MV,				// voltage of avdd	

	640,							// max width of source image
	480,							// max height of source image
	"SP0718",						// name of sensor												

	SENSOR_IMAGE_FORMAT_YUV422,		// define in SENSOR_IMAGE_FORMAT_E enum,
									// if set to SENSOR_IMAGE_FORMAT_MAX here, image format depent on SENSOR_REG_TAB_INFO_T
	SENSOR_IMAGE_PATTERN_YUV422_YUYV,	// pattern of input image form sensor;			

	s_SP0718_resolution_Tab_YUV,	// point to resolution table information structure
	&s_SP0718_ioctl_func_tab,		// point to ioctl function table
			
	PNULL,							// information and table about Rawrgb sensor
	PNULL,							// extend information about sensor	
       SENSOR_AVDD_2800MV,                     // iovdd
	SENSOR_AVDD_1800MV,                      // dvdd
	3,
       3,
	0,
	2,
	0,
	0,
	0,
	0,
       0,
};
/**---------------------------------------------------------------------------*
 ** 							Function  Definitions
 **---------------------------------------------------------------------------*/
LOCAL void SP0718_WriteReg( uint8  subaddr, uint8 data )
{
	
	#ifndef	_USE_DSP_I2C_
		//uint8 cmd[2];
		//cmd[0]	=	subaddr;
		//cmd[1]	=	data;		
		//I2C_WriteCmdArr(SP0718_I2C_ADDR_W, cmd, 2, SCI_TRUE);
		Sensor_WriteReg(subaddr, data);
	#else
		DSENSOR_IICWrite((uint16)subaddr, (uint16)data);
	#endif

	//SCI_TRACE_LOW:"SENSOR: SP0718_WriteReg reg/value(%x,%x) !!"
	SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_469_112_2_18_0_32_40_1375,(uint8*)"dd", subaddr, data);

}

LOCAL uint8 SP0718_ReadReg( uint8  subaddr)
{
	uint8 value = 0;
	
	#ifndef	_USE_DSP_I2C_
	//I2C_WriteCmdArrNoStop(SP0718_I2C_ADDR_W, &subaddr, 1,SCI_TRUE);
	//I2C_ReadCmd(SP0718_I2C_ADDR_R, &value, SCI_TRUE);
	value =Sensor_ReadReg( subaddr);
	#else
		value = (uint16)DSENSOR_IICRead((uint16)subaddr);
	#endif

    //SCI_TRACE_LOW:"SENSOR: SP0718_ReadReg reg/value(%x,%x) !!"
    SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_485_112_2_18_0_32_40_1376,(uint8*)"dd", subaddr, value);
    
	return value;
}


#if 0
LOCAL uint32 SP0718_Power_On(uint32 power_on)
{
	SENSOR_AVDD_VAL_E		dvdd_val=g_SP0718_yuv_info.dvdd_val;
	SENSOR_AVDD_VAL_E		avdd_val=g_SP0718_yuv_info.avdd_val;
	SENSOR_AVDD_VAL_E		iovdd_val=g_SP0718_yuv_info.iovdd_val;  
	BOOLEAN 				power_down=g_SP0718_yuv_info.power_down_level;	    
	BOOLEAN 				reset_level=g_SP0718_yuv_info.reset_pulse_level;
	uint32 				reset_width=g_SP0718_yuv_info.reset_pulse_width;	    

	if(SCI_TRUE==power_on)
	{
		Sensor_SetVoltage(dvdd_val, avdd_val, iovdd_val);
		SCI_Sleep(10);        
#if 1//def DOUBLE_CAMERA_SUPPORT
		if(SENSOR_MAIN == Sensor_GetCurId()) //(Sensor_GetFrontCameraState())
		{
			GPIO_SetFrontSensorPwdn(power_down);
			GPIO_SetSensorPwdn(!power_down);						
		}
		else
		{
			GPIO_SetSensorPwdn(power_down);
			GPIO_SetFrontSensorPwdn(!power_down);			
		}
#else      
		GPIO_SetSensorPwdn(!power_down);
#endif

		//Open Mclk in default frequency
		Sensor_SetMCLK(24);   
		SCI_Sleep(50);  

		// reset pin on SP0718
		Sensor_SetResetLevel(!reset_level);
		SCI_Sleep(50);
		Sensor_SetResetLevel(reset_level);
		SCI_Sleep(reset_width);
		Sensor_SetResetLevel(!reset_level);
		SCI_Sleep(50);

	}
	else
	{
#if 1//def DOUBLE_CAMERA_SUPPORT
		GPIO_SetFrontSensorPwdn(!power_down);
		GPIO_SetSensorPwdn(!power_down);
#else	
		GPIO_SetSensorPwdn(!power_down);
#endif

		Sensor_SetMCLK(SENSOR_DISABLE_MCLK);           

		Sensor_SetVoltage(SENSOR_AVDD_CLOSED, SENSOR_AVDD_CLOSED, SENSOR_AVDD_CLOSED);   
		SCI_Sleep(20);
	}


	return SCI_SUCCESS;
}
#endif


LOCAL uint32 SP0718_Identify(uint32 param)
{
    uint16  iden_reg_val = 0;
    uint32  ret = SENSOR_OP_ERR;

    iden_reg_val = SP0718_ReadReg(0x02);
    if(iden_reg_val == 0x71)
    {
            ret = SENSOR_OP_SUCCESS;
    }
    
    //SCI_TRACE_LOW:"SP0718_Identify: ret = %d"
    SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_544_112_2_18_0_32_40_1380,(uint8*)"d", ret);
    return ret;

}

LOCAL uint32 set_SP0718_ae_enable(uint32 enable)
{
   #define AE_ENABLE  (0x32)
    unsigned char ae_value;
		SP0718_WriteReg(0xfd,0x01);
        ae_value=SP0718_ReadReg(AE_ENABLE);

        if(0x00==enable)
        {
            ae_value &= 0xf8;
            SP0718_WriteReg(AE_ENABLE,ae_value);
        }
        else if(0x01==enable)
        {
            ae_value|=0x05;
            SP0718_WriteReg(AE_ENABLE,ae_value);
        }

	return 0;
}


LOCAL uint32 set_hmirror_enable(uint32 enable)
{
#ifdef DOUBLE_CAMERA_SUPPORT
uint32 i;
BOOLEAN is_page0 = FALSE;
for(i = 0; i < s_SP0718_resolution_Tab_YUV->reg_count; i++)
{
if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_addr == 0xfd)
{
if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value == 0x00)
is_page0 = TRUE;
else if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value == 0x01)
is_page0 = FALSE;
}
if ((s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_addr == 0x31) && is_page0)
{
if(enable)
{
s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value |= 0x20;
}
else
{
s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value &= ~0x20;
}
break;
}
}
return SCI_SUCCESS; 
#endif
}


LOCAL uint32 set_vmirror_enable(uint32 enable)
{
#ifdef DOUBLE_CAMERA_SUPPORT
uint32 i;
BOOLEAN is_page0 = FALSE;
for(i = 0; i < s_SP0718_resolution_Tab_YUV->reg_count; i++)
{
if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_addr == 0xfd)
{
if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value == 0x00)
is_page0 = TRUE;
else if (s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value == 0x01)
is_page0 = FALSE;
}
if ((s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_addr == 0x31) && is_page0)
{
if(enable)
{
s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value |= 0x40;
}
else
{
s_SP0718_resolution_Tab_YUV->sensor_reg_tab_ptr[i].reg_value &= ~0x40;
}
break;
}
}
return SCI_SUCCESS; 
#endif
}
/******************************************************************************/
// Description: set brightness 
// Global resource dependence: 
// Author:
// Note:
//		level  must smaller than 8
/******************************************************************************/
__align(4) const SENSOR_REG_T SP0718_brightness_tab[][3]=
{
	{
	    // Brightness -3
	{0xfd,0x01},
	{0xdb,0xd0},//level -2
	{0xff,0xff},
		},

	{
		    // Brightness -2
	{0xfd,0x01},
	{0xdb,0xe0},//level -2
	{0xff,0xff},
	},

	{
	    // Brightness -1
	{0xfd,0x01},
	{0xdb,0xf0},//level -1
	{0xff,0xff},
	},

	{
	    //  Brightness 0
	{0xfd,0x01},
	{0xdb,0x00},//level 0
	{0xff,0xff},
	
	},

	{
	// Brightness +1
	{0xfd,0x01},
	{0xdb,0x10},//level +1
	{0xff,0xff},
	},

	{
	//  Brightness +2
	{0xfd,0x01},
	{0xdb,0x20},//level +2
	{0xff,0xff},
	},
    {		
     //  Brightness +3
	{0xfd,0x01},
	{0xdb,0x30},//level +3
	{0xff,0xff},
	},

};


LOCAL uint32 set_brightness(uint32 level)
{
	uint16 i;
	SENSOR_REG_T* sensor_reg_ptr = (SENSOR_REG_T*)SP0718_brightness_tab[level];

	SCI_ASSERT(level < 7);/*assert verified*/
	SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/
	
	for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) && (0xFF != sensor_reg_ptr[i].reg_value); i++)
	{
		SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
	}
	OS_TickDelay(100); 
	//SCI_TRACE_LOW:"set_brightness: level = %d"
	SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_644_112_2_18_0_32_41_1384,(uint8*)"d", level);
	
	return 0;
}

__align(4) const SENSOR_REG_T SP0718_ev_tab[][5]=
{   
    {
		
	//level -3   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb-0x18},
	{0xec,SP0718_P1_0xec-0x18},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
	//level -2   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb-0x10},
	{0xec,SP0718_P1_0xec-0x10},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
	//level -1   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb-0x08},
	{0xec,SP0718_P1_0xec-0x08},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
	//level 0   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb},
	{0xec,SP0718_P1_0xec},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
	//level +1   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb+0x08},
	{0xec,SP0718_P1_0xec+0x08},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
	//level +2   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb+0x10},
	{0xec,SP0718_P1_0xec+0x10},	
	{0xfd, 0x00},
	{0xff,0xff}
	},
    {
    	//level +3   
	{0xfd,0x01},   
	{0xeb,SP0718_P1_0xeb+0x18},
	{0xec,SP0718_P1_0xec+0x18},	
	{0xfd, 0x00},
	{0xff,0xff}

	},
};


LOCAL uint32 set_SP0718_ev(uint32 level)
{
    uint16 i; 
    SENSOR_REG_T* sensor_reg_ptr = (SENSOR_REG_T*)SP0718_ev_tab[level];

    SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/
    SCI_ASSERT(level < 6);/*assert verified*/
 
    for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) ||(0xFF != sensor_reg_ptr[i].reg_value) ; i++)
    {
        SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
    }

    
    //SCI_TRACE_LOW:"SENSOR: set_ev: level = %d"
    SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_724_112_2_18_0_32_41_1385,(uint8*)"d", level);

    return 0;
}

/******************************************************************************/
// Description: anti 50/60 hz banding flicker
// Global resource dependence: 
// Author:
// Note:
//		level  must smaller than 8
/******************************************************************************/
LOCAL uint32 set_SP0718_anti_flicker(uint32 param )
{ 
 #if 1
     switch(param)
    {
        case DCAMERA_FLICKER_50HZ:			
		Antiflicker = DCAMERA_FLICKER_50HZ;
		SCI_TRACE_LOW( " set_SP0718_anti_flicker  50hz \r\n" );
		break;
            
        case DCAMERA_FLICKER_60HZ:
		Antiflicker = DCAMERA_FLICKER_60HZ;
		SCI_TRACE_LOW( " set_SP0718_anti_flicker  60hz \r\n" );
		break;
            
        default:
            break;
    }
    
    SP0718_Delayms(100); 
#endif
    return 0;
}

/******************************************************************************/
// Description: set video mode
// Global resource dependence: 
// Author:
// Note:
//		 
/******************************************************************************/
__align(4) const SENSOR_REG_T SP0718_video_mode_nand_tab[][9]=
{
    {		
		{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xfd, 0x00},{0xff,0xff},
	},

};
__align(4) const SENSOR_REG_T SP0718_video_mode_nor_tab_50hz[][40]=
{
    {
//SI15_SP0718 12M 50Hz 12.1569-12fps AE_Parameters_20130107172021.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x74},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x76},
  {0xfd,0x01},
  {0xef,0x3e},
  {0xf0,0x00},
  {0x02,0x08},
  {0x03,0x01},
  {0x06,0x38},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xf0},
  {0xbf,0x01},
  {0xd0,0xf0},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xf0},
  {0xfd,0x01},
  {0x32,0x15},

	},    
    //Video record daylight 
{
//SI15_SP0718 12M 50Hz 12.1569-12fps AE_Parameters_20130107172021.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x74},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x76},
  {0xfd,0x01},
  {0xef,0x3e},
  {0xf0,0x00},
  {0x02,0x08},
  {0x03,0x01},
  {0x06,0x38},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xf0},
  {0xbf,0x01},
  {0xd0,0xf0},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xf0},
  {0xfd,0x01},
  {0x32,0x15},
	},
   // UPCC  
    {
//SI15_SP0718 12M 50Hz 12.1569-12fps AE_Parameters_20130107172021.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x74},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x76},
  {0xfd,0x01},
  {0xef,0x3e},
  {0xf0,0x00},
  {0x02,0x08},
  {0x03,0x01},
  {0x06,0x38},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xf0},
  {0xbf,0x01},
  {0xd0,0xf0},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xf0},
  {0xfd,0x01},
  {0x32,0x15},

	},  
};
__align(4) const SENSOR_REG_T SP0718_video_mode_nor_tab_60hz[][40]=
{
    {
	//SI15_SP0718 12M 60Hz 12-12fps AE_Parameters_20130107172015.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x32},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x82},
  {0xfd,0x01},
  {0xef,0x33},
  {0xf0,0x00},
  {0x02,0x0a},
  {0x03,0x01},
  {0x06,0x2d},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xfe},
  {0xbf,0x01},
  {0xd0,0xfe},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xfe},
  {0xfd,0x01},
  {0x32,0x15},
	},    
    //Video record daylight 
    {
	//SI15_SP0718 12M 60Hz 12-12fps AE_Parameters_20130107172015.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x32},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x82},
  {0xfd,0x01},
  {0xef,0x33},
  {0xf0,0x00},
  {0x02,0x0a},
  {0x03,0x01},
  {0x06,0x2d},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xfe},
  {0xbf,0x01},
  {0xd0,0xfe},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xfe},
  {0xfd,0x01},
  {0x32,0x15},
	},
   // UPCC  
    {
	//SI15_SP0718 12M 60Hz 12-12fps AE_Parameters_20130107172015.txt
//ae setting  
  {0xfd,0x00},
  {0x03,0x01},
  {0x04,0x32},
  {0x06,0x00},
  {0x09,0x00},
  {0x0a,0x82},
  {0xfd,0x01},
  {0xef,0x33},
  {0xf0,0x00},
  {0x02,0x0a},
  {0x03,0x01},
  {0x06,0x2d},
  {0x07,0x00},
  {0x08,0x01},
  {0x09,0x00},
//Status      
  {0xfd,0x02},
  {0xbe,0xfe},
  {0xbf,0x01},
  {0xd0,0xfe},
  {0xd1,0x01},
  {0xfd,0x01},
  {0x5b,0x01},
  {0x5c,0xfe},
  {0xfd,0x01},
  {0x32,0x15},
	}  
};

LOCAL uint32 set_SP0718_video_mode(uint32 mode)
{
#if 1
    uint8 data=0x00;
    uint16 i;
    SENSOR_REG_T* sensor_reg_ptr = PNULL;
    uint8 tempregval = 0;
    SCI_TRACE_LOW(" xg:set_SP0718_video_mode ,%d,%d\r\n",mode,Antiflicker);
    SCI_ASSERT(mode <=DCAMERA_MODE_MAX);

	if(Antiflicker ==DCAMERA_FLICKER_50HZ )
	{
	    if(CHIP_DetectMemType())
	       
	            sensor_reg_ptr = (SENSOR_REG_T*)SP0718_video_mode_nor_tab_50hz[mode];

	    else
	       
	            sensor_reg_ptr = (SENSOR_REG_T*)SP0718_video_mode_nor_tab_50hz[mode];
	}
	else
	{
	    if(CHIP_DetectMemType())
	       
	            sensor_reg_ptr = (SENSOR_REG_T*)SP0718_video_mode_nor_tab_60hz[mode];

	    else
	       
	            sensor_reg_ptr = (SENSOR_REG_T*)SP0718_video_mode_nor_tab_60hz[mode];

	}

    SCI_ASSERT(PNULL != sensor_reg_ptr);

    for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) || (0xFF != sensor_reg_ptr[i].reg_value); i++)
    {
    	SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
    }
#endif	
    return 0;
}


/******************************************************************************/
// Description: set wb mode 
// Global resource dependence: 
// Author:
// Note:
//		
/******************************************************************************/

/*
// enum: preview while balance mode
typedef enum
{
	DCAMERA_WB_MODE_AUTO = 0x00,			//Զ
	DCAMERA_WB_MODE_INCANDESCENCE,		//׳
	DCAMERA_WB_MODE_U30,				//ùԴ
	DCAMERA_WB_MODE_CWF,				//ӫ
	DCAMERA_WB_MODE_FLUORESCENT,		//չ
	DCAMERA_WB_MODE_SUN,				//
	DCAMERA_WB_MODE_CLOUD,				//
	DCAMERA_WB_MODE_MAX
}DCAMERA_PARAM_WB_MODE_E;
*/

__align(4) const SENSOR_REG_T SP0718_awb_tab[][7]=
	{
	   	//AUTO  // AUTO 3000K~7000K 
	   	{
		{0xfd,0x02},                      
		{0x26,0xc8},		                  
		{0x27,0xb6},                      
		{0xfd,0x01}, 		
		{0x32,0x15},   //awb & ae  opened
		{0xfd,0x00},  
		{0xff,0xff} 

		},    
		//INCANDESCENCE:  //2800K~3000K  
		{
	{0xfd,0x01},  
	{0x32,0x05},                  
	{0xfd,0x02},                  
	{0x26,0x75},		              
	{0x27,0xe2},		              
	{0xfd,0x00},		
      {0xff,0xff} 
		},
	 	//U30 //3000K   //  lxy  
		{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x75},		             
	{0x27,0xe2},		             
	{0xfd,0x00}, 
	{0xff,0xff} 
		},  
	 	//CWF  //4150K  
		{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x75},		             
	{0x27,0xe2},		             
	{0xfd,0x00}, 
	{0xff,0xff} 
		},   
		//FLUORESCENT://4200K~5000K   
		{
	{0xfd,0x01}, 
	{0x32,0x05},                 
	{0xfd,0x02},                 
	{0x26,0x91},		             
	{0x27,0xc8},		             
	{0xfd,0x00}, 
	{0xff,0xff} 
		},
		//SUN: //DAYLIGHT//6500K    
		{
			{0xfd,0x01}, 
			{0x32,0x05},           
			{0xfd,0x02},           
			{0x26,0xba},	         
			{0x27,0x95},	         
			{0xfd,0x00}, 		
			{0xff,0xff} 
		},  
                //CLOUD: //7000K  
		{
		
			{0xfd,0x01}, 
			{0x32,0x05},          
			{0xfd,0x02},          
			{0x26,0xc8},	        
			{0x27,0x89},	        
			{0xfd,0x00},		
			{0xff,0xff} 
		},
};
	
	LOCAL uint32 set_SP0718_awb(uint32 mode)
	{
		uint16 i;
		
		SENSOR_REG_T* sensor_reg_ptr = (SENSOR_REG_T*)SP0718_awb_tab[mode];
	
		SCI_ASSERT(mode < 7);/*assert verified*/
		SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/
		// hill 0408
		for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) && (0xFF != sensor_reg_ptr[i].reg_value); i++)
		{
			SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
		}
		OS_TickDelay(100); 
		//SCI_TRACE_LOW:"SENSOR: set_awb_mode: mode = %d"
		SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_1119_112_2_18_0_32_42_1386,(uint8*)"d", mode);
		
		return 0;
}

__align(4) const SENSOR_REG_T SP0718_contrast_tab[][12]=
{
	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , 0x00}, 
	{0x11 , 0x00},
	{0x12 , 0x00},
	{0x13 , 0x00},
	{0x14 , 0x00},   
	{0x15 , 0x00},
	{0x16 , 0x00},  
	{0x17 , 0x00},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10-0x08}, 
	{0x11 , SP0718_P1_0x11-0x08},
	{0x12 , SP0718_P1_0x12-0x00},
	{0x13 , SP0718_P1_0x13-0x00},
	{0x14 , SP0718_P1_0x14-0x08},   
	{0x15 , SP0718_P1_0x15-0x08},
	{0x16 , SP0718_P1_0x16-0x00},  
	{0x17 , SP0718_P1_0x17-0x00},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10-0x04}, 
	{0x11 , SP0718_P1_0x11-0x04},
	{0x12 , SP0718_P1_0x12-0x00},
	{0x13 , SP0718_P1_0x13-0x00},
	{0x14 , SP0718_P1_0x14-0x04},   
	{0x15 , SP0718_P1_0x15-0x04},
	{0x16 , SP0718_P1_0x16-0x00},  
	{0x17 , SP0718_P1_0x17-0x00},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10}, 
	{0x11 , SP0718_P1_0x11},
	{0x12 , SP0718_P1_0x12},
	{0x13 , SP0718_P1_0x13},
	{0x14 , SP0718_P1_0x14},   
	{0x15 , SP0718_P1_0x15},
	{0x16 , SP0718_P1_0x16},  
	{0x17 , SP0718_P1_0x17},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10+0x04}, 
	{0x11 , SP0718_P1_0x10+0x04},
	{0x12 , SP0718_P1_0x12+0x04},
	{0x13 , SP0718_P1_0x13+0x04},
	{0x14 , SP0718_P1_0x14+0x04},   
	{0x15 , SP0718_P1_0x15+0x04},
	{0x16 , SP0718_P1_0x16+0x04},  
	{0x17 , SP0718_P1_0x17+0x04},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10+0x08}, 
	{0x11 , SP0718_P1_0x11+0x08},
	{0x12 , SP0718_P1_0x12+0x08},
	{0x13 , SP0718_P1_0x13+0x08},
	{0x14 , SP0718_P1_0x14+0x08},   
	{0x15 , SP0718_P1_0x15+0x08},
	{0x16 , SP0718_P1_0x16+0x08},  
	{0x17 , SP0718_P1_0x17+0x08},  
	{0xff,0xff}
	},

	{
   	{0xfd , 0x01},
	{0x0f , 0xff},
	{0x10 , SP0718_P1_0x10+0x10}, 
	{0x11 , SP0718_P1_0x11+0x10},
	{0x12 , SP0718_P1_0x12+0x10},
	{0x13 , SP0718_P1_0x13+0x10},
	{0x14 , SP0718_P1_0x14+0x10},   
	{0x15 , SP0718_P1_0x15+0x10},
	{0x16 , SP0718_P1_0x16+0x10},  
	{0x17 , SP0718_P1_0x17+0x10},  
	{0xff,0xff}
	}, 
};

LOCAL uint32 set_contrast(uint32 level)
{
    uint16 i;
    SENSOR_REG_T* sensor_reg_ptr;

    sensor_reg_ptr = (SENSOR_REG_T*)SP0718_contrast_tab[level];

    SCI_ASSERT(level < 7 );/*assert verified*/
    SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/

    for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) && (0xFF != sensor_reg_ptr[i].reg_value); i++)
    {
        SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
    }
    OS_TickDelay(20);
    //SCI_TRACE_LOW:"set_contrast: level = %d"
    SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_1193_112_2_18_0_32_42_1387,(uint8*)"d", level);
    return 0;
}


LOCAL uint32 set_sharpness(uint32 level)
{
	
	return 0;
}


LOCAL uint32 set_saturation(uint32 level)
{

	
	return 0;
}

/******************************************************************************/
// Description: set brightness 
// Global resource dependence: 
// Author:
// Note:
//		level  must smaller than 8
/******************************************************************************/

LOCAL uint32 set_preview_mode(uint32 preview_mode)
{
#if 1
	switch (preview_mode)
	{
		case DCAMERA_ENVIRONMENT_NORMAL: 
		case DCAMERA_ENVIRONMENT_SUNNY:
		{
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xcd,SP0718_NORMAL_Y0ffset);
			SP0718_WriteReg(0xce,0x1f);
			if(Antiflicker== DCAMERA_FLICKER_50HZ)
			{
			
			SCI_TRACE_LOW("normal mode 50hz\r\n");
// SP0718  24M  50HZ  8-15FPS 
		  //ae setting
			SP0718_WriteReg(0xfd,0x00);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x04,0xce);
			SP0718_WriteReg(0x06,0x00);
			SP0718_WriteReg(0x09,0x02);
			SP0718_WriteReg(0x0a,0xc4);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xef,0x4d);
			SP0718_WriteReg(0xf0,0x00);
			SP0718_WriteReg(0x02,0x0c);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x06,0x47);
			SP0718_WriteReg(0x07,0x00);
			SP0718_WriteReg(0x08,0x01);
			SP0718_WriteReg(0x09,0x00);
			//Status                 
			SP0718_WriteReg(0xfd,0x02);
			SP0718_WriteReg(0xbe,0x9c);
			SP0718_WriteReg(0xbf,0x03);
			SP0718_WriteReg(0xd0,0x9c);
			SP0718_WriteReg(0xd1,0x03);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x5b,0x03);
			SP0718_WriteReg(0x5c,0x9c);  
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x32,0x15);

			}
			else	
			{
			SCI_TRACE_LOW("normal mode 60hz\r\n");
// SP0718  24M  60HZ  8-15FPS 
		  //ae setting
			SP0718_WriteReg(0xfd,0x00);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x04,0x80);
			SP0718_WriteReg(0x06,0x00);
			SP0718_WriteReg(0x09,0x02);
			SP0718_WriteReg(0x0a,0xc9);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xef,0x40);
			SP0718_WriteReg(0xf0,0x00);
			SP0718_WriteReg(0x02,0x0f);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x06,0x3a);
			SP0718_WriteReg(0x07,0x00);
			SP0718_WriteReg(0x08,0x01);
			SP0718_WriteReg(0x09,0x00);
			//Status                   
			SP0718_WriteReg(0xfd,0x02);
			SP0718_WriteReg(0xbe,0xc0);
			SP0718_WriteReg(0xbf,0x03);
			SP0718_WriteReg(0xd0,0xc0);
			SP0718_WriteReg(0xd1,0x03);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x5b,0x03);
			SP0718_WriteReg(0x5c,0xc0);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x32,0x15);
			}	
			break;
		}
		case DCAMERA_ENVIRONMENT_NIGHT:
		{
                    SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xcd,SP0718_LOWLIGHT_Y0ffset);
			SP0718_WriteReg(0xce,0x1f);
			if(Antiflicker== DCAMERA_FLICKER_50HZ)
			{
			SCI_TRACE_LOW("night mode 50hz\r\n");
//capture preview night 24M 50hz 15-6FPS  
			SP0718_WriteReg(0xfd,0x00);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x04,0xce);
			SP0718_WriteReg(0x06,0x00);
			SP0718_WriteReg(0x09,0x02);
			SP0718_WriteReg(0x0a,0xc4);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xef,0x4d);
			SP0718_WriteReg(0xf0,0x00);
			SP0718_WriteReg(0x02,0x10);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x06,0x47);
			SP0718_WriteReg(0x07,0x00);
			SP0718_WriteReg(0x08,0x01);
			SP0718_WriteReg(0x09,0x00);
			//Statu                  
			SP0718_WriteReg(0xfd,0x02);
			SP0718_WriteReg(0xbe,0xd0);
			SP0718_WriteReg(0xbf,0x04);
			SP0718_WriteReg(0xd0,0xd0);
			SP0718_WriteReg(0xd1,0x04);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x5b,0x04);
			SP0718_WriteReg(0x5c,0xd0);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x32,0x15);
			}
			else
			{
			SCI_TRACE_LOW("night mode 60hz\r\n");
//capture preview night 24M 60hz 15-6FPS 
			SP0718_WriteReg(0xfd,0x00);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x04,0x80);
			SP0718_WriteReg(0x06,0x00);
			SP0718_WriteReg(0x09,0x02);
			SP0718_WriteReg(0x0a,0xc9);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0xef,0x40);
			SP0718_WriteReg(0xf0,0x00);
			SP0718_WriteReg(0x02,0x14);
			SP0718_WriteReg(0x03,0x01);
			SP0718_WriteReg(0x06,0x3a);
			SP0718_WriteReg(0x07,0x00);
			SP0718_WriteReg(0x08,0x01);
			SP0718_WriteReg(0x09,0x00);
			//Status                 
			SP0718_WriteReg(0xfd,0x02);
			SP0718_WriteReg(0xbe,0x00);
			SP0718_WriteReg(0xbf,0x05);
			SP0718_WriteReg(0xd0,0x00);
			SP0718_WriteReg(0xd1,0x05);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x5b,0x05);
			SP0718_WriteReg(0x5c,0x00);
			SP0718_WriteReg(0xfd,0x01);
			SP0718_WriteReg(0x32,0x15);

			}
			break;
		}
		
		default:
		{
			break;
		}

	SCI_Sleep(100);
return 0;
	} 
#endif
}
/*
typedef enum
{
	DCAMERA_EFFECT_NORMAL = 0x00,
	DCAMERA_EFFECT_BLACKWHITE,			// ڰ
	DCAMERA_EFFECT_RED,					// ƫ
	DCAMERA_EFFECT_GREEN,				// ƫ
	DCAMERA_EFFECT_BLUE,				// ƫ
	DCAMERA_EFFECT_YELLOW,				// ƫ
	DCAMERA_EFFECT_NEGATIVE,			// Ƭ
	DCAMERA_EFFECT_CANVAS,				// 
	DCAMERA_EFFECT_RELIEVOS,			// 	
	DCAMERA_EFFECT_MAX
}DCAMERA_PARAM_EFFECT_E;
*/

__align(4) const SENSOR_REG_T SP0718_image_effect_tab[][5]=	
{
    // effect normal
    {
	{0xfd, 0x01},
	{0x66, 0x00},
	{0x67, 0x80},
	{0x68, 0x80},
	{0xff, 0xff}
    },
    //effect BLACKWHITE//GRAYSCALE
    {
	{0xfd, 0x01},
	{0x66, 0x20},
	{0x67, 0x80},
	{0x68, 0x80},
	{0xff, 0xff}
    },
    // effect RED pink      //  lxy 
    {
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0xf0},
	{0x68, 0x20},
	{0xff, 0xff}
    },
    // effect GREEN//SEPIAGREEN   
    {
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0x20},
	{0x68, 0x20},
	{0xff, 0xff}
    },
    // effect  BLUE//SEPIABLUE    
    {
    // Bluish
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0x20},
	{0x68, 0xf0},
	{0xff, 0xff}
    },
    // effect  YELLOW//SEPIAYELLOW    //  lxy  
    {
    //  Greenish
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0xa0},
	{0x68, 0x40},
	{0xff, 0xff}

    },  
  
    // effect NEGATIVE//COLORINV  
    {	     
	{0xfd, 0x01},
	{0x66, 0x04},
	{0x67, 0x80},
	{0x68, 0x80},
	{0xff, 0xff}
    },    
    //effect ANTIQUE//SEPIA   
    {
	{0xfd, 0x01},
	{0x66, 0x10},
	{0x67, 0xc0},
	{0x68, 0x20},
	{0xff, 0xff}
	},
};
LOCAL uint32 set_image_effect(uint32 effect_type)
{
    uint16 i;
    
    SENSOR_REG_T* sensor_reg_ptr = (SENSOR_REG_T*)SP0718_image_effect_tab[effect_type];

    SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/

    for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) || (0xFF != sensor_reg_ptr[i].reg_value) ; i++)
    {
        Sensor_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
    }
    //SCI_TRACE_LOW:"-----------set_image_effect: effect_type = %d------------"
    SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_1477_112_2_18_0_32_42_1389,(uint8*)"d", effect_type);
    
    return 0;
}

LOCAL uint32 read_ev_value(uint32 value)
{
	return 0;
}

LOCAL uint32 write_ev_value(uint32 exposure_value)
{
	
	return 0;	
}

LOCAL uint32 read_gain_value(uint32 value)
{

	
	return 0;
}

LOCAL uint32 write_gain_value(uint32 gain_value)
{

	
	return 0;
}

LOCAL uint32 read_gain_scale(uint32 value)
{
	return SENSOR_GAIN_SCALE;
	
}


LOCAL uint32 set_frame_rate(uint32 param)
    
{
	//SP0718_WriteReg( 0xd8, uint8 data );
	return 0;
}

/******************************************************************************/
// Description:
// Global resource dependence: 
// Author:
// Note:
//		mode 0:normal;	 1:night 
/******************************************************************************/
__align(4) const SENSOR_REG_T SP0718_mode_tab[][5]=
{

	 {    
        {0xff,0xff},
        
    },
	//night
    {    
        {0xff,0xff},
    }

};

LOCAL uint32 SP0718_set_work_mode(uint32 mode)
{
	uint16 i;
	SENSOR_REG_T* sensor_reg_ptr = (SENSOR_REG_T*)SP0718_mode_tab[mode];

	SCI_ASSERT(mode <= 1);/*assert verified*/
	SCI_ASSERT(PNULL != sensor_reg_ptr);/*assert verified*/
	
	for(i = 0; (0xFF != sensor_reg_ptr[i].reg_addr) || (0xFF != sensor_reg_ptr[i].reg_value); i++)
	{
		SP0718_WriteReg(sensor_reg_ptr[i].reg_addr, sensor_reg_ptr[i].reg_value);
	}

	//SCI_TRACE_LOW:"set_work_mode: mode = %d"
	SCI_TRACE_ID(TRACE_TOOL_CONVERT,SENSOR_SP0718_1632_112_2_18_0_32_43_1391,(uint8*)"d", mode);
	return 0;
}
